#ifndef APP_H_
#define APP_H_

#include <include_closed/common/toolkit/list/PointerList.h>
#include <include_closed/common/toolkit/list/PointerListIter.h>
#include <include_closed/common/toolkit/list/StrCpyList.h>
#include <include_closed/common/toolkit/list/StrCpyListIter.h>
#include <include_closed/common/Common.h>
#include <include_open/app/config/Open_MountConfig.h>
#include <include_open/common/threading/Open_Mutex.h>


#ifndef FHGFS_VERSION
#define FHGFS_VERSION   "2009.08-r1"
#endif


// program return codes
#define APPCODE_NO_ERROR               0
#define APPCODE_PROGRAM_ERROR          1
#define APPCODE_INVALID_CONFIG         2
#define APPCODE_INITIALIZATION_ERROR   3
#define APPCODE_RUNTIME_ERROR          4

// forward declarations
struct Config;
struct ExternalHelperd;
struct Logger;
struct Worker;
struct DatagramListener;
struct HeartbeatManager;
struct Node;
struct NodeStoreEx;
struct WorkQueue;
struct CacheStore;
struct NoAllocBufferStore;
struct AcknowledgmentStore;
struct NetFilter;



struct App;
typedef struct App App;


extern void App_init(App* this, MountConfig* mountConfig);
extern App* App_construct(MountConfig* mountConfig);
extern void App_uninit(App* this);
extern void App_destruct(App* this);

extern int App_run(App* this);
extern void App_stop(App* this);

extern fhgfs_bool __App_initDataObjects(App* this, MountConfig* mountConfig);
extern fhgfs_bool __App_initLocalNodeInfo(App* this);
extern fhgfs_bool __App_initStorage(App* this);
extern fhgfs_bool __App_initComponents(App* this);
extern void __App_startComponents(App* this);
extern void __App_stopComponents(App* this);
extern void __App_joinComponents(App* this);

extern void __App_logInfos(App* this);

extern fhgfs_bool __App_workersInit(App* this);
extern void __App_workersStart(App* this);
extern void __App_workersStop(App* this);
extern void __App_workersDelete(App* this);
extern void __App_workersJoin(App* this);


// inliners
static inline struct Logger* App_getLogger(App* this);
static inline struct ExternalHelperd* App_getHelperd(App* this);
static inline struct Config* App_getConfig(App* this);
static inline struct NetFilter* App_getNetFilter(App* this);
static inline StrCpyList* App_getPreferredNodes(App* this);
static inline struct Node* App_getLocalNode(App* this);
static inline struct NodeStoreEx* App_getMgmtNodes(App* this);
static inline struct NodeStoreEx* App_getMetaNodes(App* this);
static inline struct NodeStoreEx* App_getStorageNodes(App* this);
static inline struct WorkQueue* App_getWorkQueue(App* this);
static inline struct WorkQueue* App_getRetryWorkQueue(App* this);
static inline struct NoAllocBufferStore* App_getPagedIOBufStore(App* this);
static inline struct CacheStore* App_getCacheStore(App* this);
static inline struct NoAllocBufferStore* App_getRWStatesStore(App* this);
static inline struct NoAllocBufferStore* App_getPathBufStore(App* this);
static inline struct NoAllocBufferStore* App_getMsgBufStore(App* this);
static inline struct AcknowledgmentStore* App_getAckStore(App* this);
static inline Mutex* App_getFhgfsOpsMutex(App* this);
static inline struct DatagramListener* App_getDatagramListener(App* this);
static inline fhgfs_bool App_getConnRetriesEnabled(App* this);
static inline void App_setConnRetriesEnabled(App* this, fhgfs_bool connRetriesEnabled);


struct App
{
   int appResult;
   MountConfig* mountConfig;

   struct Config*  cfg;
   struct ExternalHelperd* helperd;
   struct Logger*  logger;

   struct NetFilter* netFilter; // empty filter means "all nets allowed"
   StrCpyList* allowedInterfaces; // empty list means "all interfaces accepted"
   StrCpyList* preferredNodes; // empty list means "no preferred nodes => use any"
   struct Node* localNode;
   struct NodeStoreEx* mgmtNodes;
   struct NodeStoreEx* metaNodes;
   struct NodeStoreEx* storageNodes;
   struct WorkQueue* workQueue;
   struct WorkQueue* retryWorkQueue;
   struct NoAllocBufferStore* pagedIOBufStore; // for paged IO read/write
   struct CacheStore* cacheStore; // file caches for "buffered" cache mode
   struct NoAllocBufferStore* rwStatesStore; // for single-threaded read/write remoting
   struct NoAllocBufferStore* pathBufStore; // for dentry path lookups
   struct NoAllocBufferStore* msgBufStore; // for MessagingTk request/response
   struct AcknowledgmentStore* ackStore;

   struct DatagramListener* dgramListener;
   struct HeartbeatManager* heartbeatMgr;
   PointerList* workerList;
   PointerList* retryWorkerList;

   volatile fhgfs_bool connRetriesEnabled; // changed at umount and via procfs

   Mutex* fhgfsOpsMutex; // general lock during fhgfs kernel operations (FhgfsOps_...)
};


struct Logger* App_getLogger(App* this)
{
   return this->logger;
}

struct Config* App_getConfig(App* this)
{
   return this->cfg;
}

struct ExternalHelperd* App_getHelperd(App* this)
{
   return this->helperd;
}

struct NetFilter* App_getNetFilter(App* this)
{
   return this->netFilter;
}

StrCpyList* App_getPreferredNodes(App* this)
{
   return this->preferredNodes;
}

struct Node* App_getLocalNode(App* this)
{
   return this->localNode;
}

struct NodeStoreEx* App_getMgmtNodes(App* this)
{
   return this->mgmtNodes;
}

struct NodeStoreEx* App_getMetaNodes(App* this)
{
   return this->metaNodes;
}

struct NodeStoreEx* App_getStorageNodes(App* this)
{
   return this->storageNodes;
}

struct WorkQueue* App_getWorkQueue(App* this)
{
   return this->workQueue;
}

struct WorkQueue* App_getRetryWorkQueue(App* this)
{
   return this->retryWorkQueue;
}

struct NoAllocBufferStore* App_getPagedIOBufStore(App* this)
{
   return this->pagedIOBufStore;
}

struct CacheStore* App_getCacheStore(App* this)
{
   return this->cacheStore;
}

struct NoAllocBufferStore* App_getRWStatesStore(App* this)
{
   return this->rwStatesStore;
}

struct NoAllocBufferStore* App_getPathBufStore(App* this)
{
   return this->pathBufStore;
}

struct NoAllocBufferStore* App_getMsgBufStore(App* this)
{
   return this->msgBufStore;
}

struct AcknowledgmentStore* App_getAckStore(App* this)
{
   return this->ackStore;
}

Mutex* App_getFhgfsOpsMutex(App* this)
{
   return this->fhgfsOpsMutex;
}

struct DatagramListener* App_getDatagramListener(App* this)
{
   return this->dgramListener;
}

fhgfs_bool App_getConnRetriesEnabled(App* this)
{
   return this->connRetriesEnabled;
}

void App_setConnRetriesEnabled(App* this, fhgfs_bool connRetriesEnabled)
{
   this->connRetriesEnabled = connRetriesEnabled;
}

#endif /*APP_H_*/
