#ifndef NOALLOCBUFFERSTORE_H_
#define NOALLOCBUFFERSTORE_H_

#include <include_open/common/threading/Open_Mutex.h>
#include <include_open/common/threading/Open_Condition.h>
#include <include_open/common/threading/Open_Thread.h>
#include <include_closed/common/Common.h>

/**
 * This BufferStore does not use any kind of memory allocation during its normal
 * operation. However, it does (de)allocate the requested number of buffers
 * during (de)initialization.
 */


struct NoAllocBufferStore;
typedef struct NoAllocBufferStore NoAllocBufferStore;

static inline void NoAllocBufferStore_init(NoAllocBufferStore* this, size_t numBufs,
   size_t bufSize);
static inline NoAllocBufferStore* NoAllocBufferStore_construct(size_t numBufs, size_t bufSize);
static inline void NoAllocBufferStore_uninit(NoAllocBufferStore* this);
static inline void NoAllocBufferStore_destruct(NoAllocBufferStore* this);

static inline char* NoAllocBufferStore_waitForBuf(NoAllocBufferStore* this);
static inline char* NoAllocBufferStore_instantBuf(NoAllocBufferStore* this);
static inline void NoAllocBufferStore_addBuf(NoAllocBufferStore* this, char* buf);

static inline void __NoAllocBufferStore_initBuffers(NoAllocBufferStore* this);

// getters & setters
static inline size_t NoAllocBufferStore_getBufSize(NoAllocBufferStore* this);
static inline fhgfs_bool NoAllocBufferStore_getValid(NoAllocBufferStore* this);


struct NoAllocBufferStore
{
   char** bufArray;

   size_t numBufs;
   size_t bufSize;

   size_t numAvailable; // number of currently available buffers in the store

   fhgfs_bool storeValid;

   Mutex* mutex;
   Condition* newBufCond;
};

/**
 * @param numBufs number of buffers to be allocated
 * @param bufSize size of each allocated buffer
 */
void NoAllocBufferStore_init(NoAllocBufferStore* this, size_t numBufs, size_t bufSize)
{
   this->mutex = Mutex_construct();
   this->newBufCond = Condition_construct();

   this->bufArray = numBufs ? (char**)os_kzalloc(numBufs * sizeof(char*) ) : NULL;

   this->numBufs = numBufs;
   this->bufSize = bufSize;

   this->numAvailable = 0;

   this->storeValid = fhgfs_true;

   __NoAllocBufferStore_initBuffers(this);
}

/**
 * @param numBufs number of buffers to be allocated
 * @param bufSize size of each allocated buffer
 */
struct NoAllocBufferStore* NoAllocBufferStore_construct(size_t numBufs, size_t bufSize)
{
   struct NoAllocBufferStore* this = (NoAllocBufferStore*)os_kmalloc(sizeof(*this) );

   NoAllocBufferStore_init(this, numBufs, bufSize);

   return this;
}

void NoAllocBufferStore_uninit(NoAllocBufferStore* this)
{
   // delete buffers
   size_t i;

   for(i=0; i < this->numAvailable; i++)
   {
      os_vfree( (this->bufArray)[i] );
   }

   // normal clean-up
   Condition_destruct(this->newBufCond);
   Mutex_destruct(this->mutex);

   SAFE_KFREE(this->bufArray);
}

void NoAllocBufferStore_destruct(NoAllocBufferStore* this)
{
   NoAllocBufferStore_uninit(this);

   os_kfree(this);
}

/**
 * Gets a buffer from the store.
 * Waits if no buffer is immediately available.
 *
 * @return a valid buffer pointer
 */
char* NoAllocBufferStore_waitForBuf(NoAllocBufferStore* this)
{
   char* buf;

   Mutex_lock(this->mutex);

   while(!this->numAvailable)
      Condition_wait(this->newBufCond, this->mutex);

   buf = (this->bufArray)[this->numAvailable-1];
   (this->numAvailable)--;

   Mutex_unlock(this->mutex);

   return buf;
}

/**
 * Gets a buffer from the store.
 * Fails if no buffer is immediately available.
 *
 * @return buffer pointer if a buffer was immediately available, NULL otherwise
 */
char* NoAllocBufferStore_instantBuf(NoAllocBufferStore* this)
{
   char* buf;

   Mutex_lock(this->mutex);

   if(!this->numAvailable)
      buf = NULL;
   else
   {
      buf = (this->bufArray)[this->numAvailable-1];
      (this->numAvailable)--;
   }

   Mutex_unlock(this->mutex);

   return buf;
}

void NoAllocBufferStore_addBuf(NoAllocBufferStore* this, char* buf)
{
   Mutex_lock(this->mutex);

   Condition_signal(this->newBufCond);

   (this->bufArray)[this->numAvailable] = buf;
   (this->numAvailable)++;

   Mutex_unlock(this->mutex);
}

void __NoAllocBufferStore_initBuffers(NoAllocBufferStore* this)
{
   size_t i;

   for(i = 0; i < this->numBufs; i++)
   {
      char* buf = os_vmalloc(this->bufSize);
      if(!buf)
      {
         printk_fhgfs(FHGFS_KERN_WARNING, "NoAllocBufferStore_initBuffers: vmalloc failed to alloc "
            "%lld bytes for buffer number %lld\n", (long long)this->bufSize, (long long)i);

         this->storeValid = fhgfs_false;

         break;
      }

      (this->bufArray)[i] = buf;
      (this->numAvailable)++;
   }
}

size_t NoAllocBufferStore_getBufSize(NoAllocBufferStore* this)
{
   return this->bufSize;
}

fhgfs_bool NoAllocBufferStore_getValid(NoAllocBufferStore* this)
{
   return this->storeValid;
}


#endif /*NOALLOCBUFFERSTORE_H_*/
