#include <include_open/app/config/Open_Config.h>
#include <include_open/common/toolkit/Open_StringTk.h>
#include <include_closed/app/config/Config.h>
#include <open/common/Common.h>

#include <linux/fs.h>
#include <asm/uaccess.h>


#define CONFIG_DEFAULT_CFGFILENAME     "/etc/fhgfs_client.conf"
#define CONFIG_FILE_COMMENT_CHAR       '#'
#define CONFIG_FILE_MAX_LINE_LENGTH    1024
#define CONFIG_ERR_BUF_LENGTH          1024


static fhgfs_bool __Config_readLineFromFile(struct Config* this, struct file* cfgFile,
   char* buf, size_t bufLen, fhgfs_bool* outEndOfFile);



fhgfs_bool __Config_loadFromFile(struct Config* this, const char* filename)
{
   fhgfs_bool retVal = fhgfs_true;

   struct file* cfgFile;
   char* line;
   char* trimLine;
   mm_segment_t oldfs;

   //printk_fhgfs(FHGFS_KERN_INFO, "Attempting to read config file: '%s'\n", filename);
   cfgFile = filp_open(filename, (O_RDONLY), 0);
   if(IS_ERR(cfgFile) )
   {
      printk_fhgfs(FHGFS_KERN_WARNING, "Failed to open config file: '%s'\n", filename);
      return fhgfs_false;
   }

   line = os_kmalloc(CONFIG_FILE_MAX_LINE_LENGTH);


   for( ; ; )
   {
      fhgfs_bool endOfFile;
      fhgfs_bool readRes;

      readRes = __Config_readLineFromFile(this, cfgFile, line, CONFIG_FILE_MAX_LINE_LENGTH,
         &endOfFile);

      // stop on end of file
      if(endOfFile)
         break;

      if(!readRes)
      { // error occured
         printk_fhgfs(FHGFS_KERN_WARNING, "Error occured while reading the config file\n");

         retVal = fhgfs_false;

         break;
      }


      // trim the line and add it to the config, ignore comment lines
      trimLine = StringTk_trimCopy(line);
      if(strlen(trimLine) && (trimLine[0] != CONFIG_FILE_COMMENT_CHAR) )
         __Config_addLineToConfigMap(this, trimLine);

      kfree(trimLine);

   }

   // clean up
   kfree(line);

   ACQUIRE_PROCESS_CONTEXT(oldfs);
   filp_close(cfgFile, NULL);
   RELEASE_PROCESS_CONTEXT(oldfs);

   return retVal;
}

fhgfs_bool Config_loadStringListFile(struct Config* this, const char* filename, StrCpyList* outList)
{
   fhgfs_bool retVal = fhgfs_true;

   struct file* listFile;
   char* line;
   char* trimLine;
   mm_segment_t oldfs;

   //printk_fhgfs(FHGFS_KERN_INFO, "Attempting to read configured list file: '%s'\n", filename);
   listFile = filp_open(filename, (O_RDONLY), 0);
   if(IS_ERR(listFile) )
   {
      printk_fhgfs(FHGFS_KERN_WARNING, "Failed to open (config) list file: '%s'\n", filename);
      return fhgfs_false;
   }

   line = os_kmalloc(CONFIG_FILE_MAX_LINE_LENGTH);


   for( ; ; )
   {
      fhgfs_bool endOfFile;
      fhgfs_bool readRes;

      readRes = __Config_readLineFromFile(this, listFile, line, CONFIG_FILE_MAX_LINE_LENGTH,
         &endOfFile);

      // stop on end of file
      if(endOfFile)
         break;

      if(!readRes)
      { // error occured
         printk_fhgfs(FHGFS_KERN_WARNING, "Error occured while reading a (config) list file\n");

         retVal = fhgfs_false;

         break;
      }

      // trim the line and add it to the nodes list, ignore comment lines
      trimLine = StringTk_trimCopy(line);
      if(strlen(trimLine) && (trimLine[0] != CONFIG_FILE_COMMENT_CHAR) )
         StrCpyList_append(outList, trimLine);

      kfree(trimLine);

   }

   // clean up
   kfree(line);

   ACQUIRE_PROCESS_CONTEXT(oldfs);
   filp_close(listFile, NULL);
   RELEASE_PROCESS_CONTEXT(oldfs);

   return retVal;
}

/**
 * @return the string is not alloc'ed and does not need to be freed
 */
const char* __Config_createDefaultCfgFilename()
{
   const char* retVal = "";

   struct file* cfgFile;

   mm_segment_t oldfs;

   // try to open the default config file
   cfgFile = filp_open(CONFIG_DEFAULT_CFGFILENAME, (O_RDONLY), 0);
   if(!IS_ERR(cfgFile) )
   { // this is something that we can open, so it's probably the config file ;)
      retVal = CONFIG_DEFAULT_CFGFILENAME;

      ACQUIRE_PROCESS_CONTEXT(oldfs);
      filp_close(cfgFile, NULL);
      RELEASE_PROCESS_CONTEXT(oldfs);
   }

   return retVal;
}

/**
 * Note: Cuts off lines that are longer than the buffer
 *
 * @param outEndOfFile fhgfs_true if end of file reached
 * @return fhgfs_false on file io error
 */
fhgfs_bool __Config_readLineFromFile(struct Config* this, struct file* cfgFile,
   char* buf, size_t bufLen, fhgfs_bool* outEndOfFile)
{
   mm_segment_t oldfs;
   size_t numRead;
   fhgfs_bool endOfLine;
   fhgfs_bool errorOccured;

   *outEndOfFile = fhgfs_false;
   endOfLine = fhgfs_false;
   errorOccured = fhgfs_false;

   ACQUIRE_PROCESS_CONTEXT(oldfs);

   for(numRead = 0; numRead < (bufLen-1); numRead++)
   {
      char charBuf;

      ssize_t readRes = vfs_read(cfgFile, &charBuf, 1, &cfgFile->f_pos);

      if( (readRes > 0) && (charBuf == '\n') )
      { // end of line
         endOfLine = fhgfs_true;
         break;
      }
      else
      if(readRes > 0)
      { // any normal character
         buf[numRead] = charBuf;
      }
      else
      if(readRes == 0)
      { // end of file
         *outEndOfFile = fhgfs_true;
         break;
      }
      else
      { // error occured
         printk_fhgfs(FHGFS_KERN_WARNING, "Failed to read from file at line position: %lld\n",
            (long long)numRead);

         errorOccured = fhgfs_true;
         break;
      }

   }

   buf[numRead] = 0; // add terminating zero

   // read the rest of the line if it is longer than the buffer
   while(!endOfLine && !(*outEndOfFile) && !errorOccured)
   {
      char charBuf;

      ssize_t readRes = vfs_read(cfgFile, &charBuf, 1, &cfgFile->f_pos);
      if( (readRes > 0) && (charBuf == '\n') )
         endOfLine = fhgfs_true;
      if(readRes == 0)
         *outEndOfFile = fhgfs_true;
      else
      if(readRes < 0)
         errorOccured = fhgfs_true;
   }


   RELEASE_PROCESS_CONTEXT(oldfs);

   return !errorOccured;

}
