#include <include_open/common/toolkit/Open_StringTk.h>
#include "FhgfsOpsSuper.h"
#include "FhgfsInode.h"


/**
 * Called once for initialization of new inodes (and not again if they are re-cycled).
 *
 * Note: Initializes only the FhGFS part of the inode, not the general VFS part.
 */
void FhgfsInode_initOnce(FhgfsInode* fhgfsInode)
{
   spin_lock_init(&fhgfsInode->entryInfoLock);
   init_MUTEX(&fhgfsInode->fileHandlesMutex);
}

/**
 * Called each time an inode is alloced from the mem cache.
 *
 * Note: Initializes only the FhGFS part of the inode, not the general VFS part.
 */
void FhgfsInode_allocInit(FhgfsInode* fhgfsInode)
{
   int i;

   memset(&fhgfsInode->entryInfo, 0, sizeof(fhgfsInode->entryInfo) );

   Time_init(&fhgfsInode->dataCacheTime);

   memset(&fhgfsInode->fileHandles, 0, sizeof(fhgfsInode->fileHandles) );

   for(i=0; i < FHGFS_INODE_FILEHANDLES_NUM; i++)
      fhgfsInode->fileHandles[i].maxUsedNodeIndex = -1;

   fhgfsInode->pattern = NULL;
}

/**
 * Called each time an inode is returned to the mem cache.
 *
 * Note: Destroys only the FhGFS part of the inode, not the general VFS part.
 */
void FhgfsInode_destroyUninit(FhgfsInode* fhgfsInode)
{
   SAFE_KFREE_NOSET(fhgfsInode->entryInfo.ownerNodeID);
   SAFE_KFREE_NOSET(fhgfsInode->entryInfo.entryID);

   Time_uninit(&fhgfsInode->dataCacheTime);

   SAFE_DESTRUCT_NOSET(fhgfsInode->pattern, StripePattern_virtualDestruct);
}

/**
 * @param openFlags OPENFILE_ACCESS_... flags
 * @param allowRWHandle fhgfs_true if you specified read or write flags and would also accept
 * a rw handle (only appropriate for writepage etc., use _handleTypeToOpenFlags() in this
 * case when you get the RemoteIOInfo).
 * @param handleType pass this to releaseHandle()
 * @param outFileHandleID caller may not free or modify (it's not alloced for the caller!)
 */
FhgfsOpsErr FhgfsInode_referenceHandle(FhgfsInode* this, int openFlags, fhgfs_bool allowRWHandle,
   FileHandleType* outHandleType)
{
   App* app = FhgfsOps_getApp(this->vfs_inode.i_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsInode_referenceHandle";

   FhgfsOpsErr retVal = FhgfsOpsErr_SUCCESS;
   FhgfsInodeFileHandle* desiredHandle;
   FhgfsInodeFileHandle* rwHandle = &this->fileHandles[FileHandleType_RW];

   *outHandleType = __FhgfsInode_openFlagsToHandleType(openFlags);
   desiredHandle = &this->fileHandles[*outHandleType];

   if(unlikely(Logger_getLogLevel(log) >= 5) )
      Logger_logFormatted(log, 5, logContext, "called. i_ino: %lld",
         (long long)this->vfs_inode.i_ino);

   down(&this->fileHandlesMutex); // L O C K

   if(desiredHandle->refCount)
   { // desired handle exists => return it
      retVal = __FhgfsInode_referenceTrunc(this, app, openFlags);
      if(retVal != FhgfsOpsErr_SUCCESS)
         goto err_unlock;

      desiredHandle->refCount++;
      //*outFileHandleID = desiredHandle->fileHandleID;
   }
   else
   if(allowRWHandle && rwHandle->refCount)
   { // rw handle allowed and exists => return it
      retVal = __FhgfsInode_referenceTrunc(this, app, openFlags);
      if(retVal != FhgfsOpsErr_SUCCESS)
         goto err_unlock;

      rwHandle->refCount++;
      //*outFileHandleID = rwHandle->fileHandleID;
      *outHandleType = FileHandleType_RW;
   }
   else
   { // no matching handle yet => open file to create a new handle
      OwnerMinInfo entryInfo;
      RemotingIOInfo ioInfo;

      FhgfsInode_getOwnerMinInfoCopy(this, &entryInfo);
      __FhgfsInode_initIOInfo(this, desiredHandle, openFlags, &ioInfo);

      retVal = FhgfsOpsRemoting_openfile(&entryInfo, &ioInfo);

      if(retVal == FhgfsOpsErr_SUCCESS)
      {
         desiredHandle->fileHandleID = ioInfo.fileHandleID;
         this->pattern = ioInfo.pattern;

         desiredHandle->refCount++;
         //*outFileHandleID = ioInfo.fileHandleID;
      }

      // clean up
      FhgfsInode_freeOwnerMinInfoVals(&entryInfo);
   }


err_unlock:

   up(&this->fileHandlesMutex); // U N L O C K

   return retVal;
}

/**
 * Note: even if this results in a (remote) error, the ref count will be decreased and the caller
 * may no longer access the corresponding ressources.
 *
 * @param handleType what you got back from _referenceHandle()
 */
FhgfsOpsErr FhgfsInode_releaseHandle(FhgfsInode* this, FileHandleType handleType,
   int maxUsedNodeIndex)
{
   App* app = FhgfsOps_getApp(this->vfs_inode.i_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsInode_releaseHandle";

   FhgfsOpsErr retVal = FhgfsOpsErr_SUCCESS;
   FhgfsInodeFileHandle* handle = &this->fileHandles[handleType];

   if(unlikely(Logger_getLogLevel(log) >= 5) )
      Logger_logFormatted(log, 5, logContext, "called. i_ino: %lld",
         (long long)this->vfs_inode.i_ino);

   down(&this->fileHandlesMutex); // L O C K

   handle->maxUsedNodeIndex = MAX(handle->maxUsedNodeIndex, maxUsedNodeIndex);

   if(unlikely(!handle->refCount) )
      FHGFS_BUG_ON(fhgfs_true, "refCount already 0"); // refcount already zero => bug
   else
   { // positive refCount
      handle->refCount--;

      if(!handle->refCount)
      { // we dropped the last reference => close remote
         OwnerMinInfo entryInfo;
         int openFlags = FhgfsInode_handleTypeToOpenFlags(handleType);
         RemotingIOInfo ioInfo;

         FhgfsInode_getOwnerMinInfoCopy(this, &entryInfo);
         __FhgfsInode_initIOInfo(this, handle, openFlags, &ioInfo);

         retVal = FhgfsOpsRemoting_closefile(&entryInfo, &ioInfo);

         // clean up
         kfree(handle->fileHandleID);
         FhgfsInode_freeOwnerMinInfoVals(&entryInfo);
      }
   }

   up(&this->fileHandlesMutex); // U N L O C K

   return retVal;
}

/**
 * Converts OPENFILE_ACCESS_... fhgfs flags to FileHandleType
 */
FileHandleType __FhgfsInode_openFlagsToHandleType(int openFlags)
{
   switch(openFlags & OPENFILE_ACCESS_MASK_RW)
   {
      case OPENFILE_ACCESS_WRITE:
      {
         return FileHandleType_WRITE;
      } break;

      case OPENFILE_ACCESS_READWRITE:
      {
         return FileHandleType_RW;
      } break;

      default:
      {
         return FileHandleType_READ;
      } break;
   }
}

/**
 * Truncates the file if OPENFILE_ACCESS_TRUNC has been specified in openFlags.
 */
FhgfsOpsErr __FhgfsInode_referenceTrunc(FhgfsInode* this, App* app, int openFlags)
{
   FhgfsOpsErr retVal = FhgfsOpsErr_SUCCESS;

   if(openFlags & OPENFILE_ACCESS_TRUNC)
   {
      OwnerMinInfo entryInfo;

      FhgfsInode_getOwnerMinInfoCopy(this, &entryInfo);

      retVal = FhgfsOpsRemoting_truncfile(app, &entryInfo, 0);

      // clean up
      FhgfsInode_freeOwnerMinInfoVals(&entryInfo);
   }

   return retVal;
}


/**
 * @param outEntryInfo contained values will be kalloced and need to be kfreed with
 * _freeOwnerMinInfoVals() later.
 */
void FhgfsInode_getOwnerMinInfoCopy(FhgfsInode* fhgfsInode, OwnerMinInfo* outEntryInfo)
{
   if(fhgfsInode->vfs_inode.i_ino == FHGFS_INODE_ROOT_INO)
   { // root node => owner info is not stored in the inode
      App* app = FhgfsOps_getApp(fhgfsInode->vfs_inode.i_sb);

      MetadataTk_getRootMinInfoCopy(app, outEntryInfo);
   }
   else
   { // "normal" inode
      spin_lock(&fhgfsInode->entryInfoLock);

      // note: spinlock => don't sleep
      outEntryInfo->ownerNodeID = kstrdup(fhgfsInode->entryInfo.ownerNodeID, GFP_ATOMIC);
      outEntryInfo->entryID = kstrdup(fhgfsInode->entryInfo.entryID, GFP_ATOMIC);

      spin_unlock(&fhgfsInode->entryInfoLock);
   }
}

/**
 * Note: This may only be used by callers that have aqcuired a reference, either explicitly by
 * calling _referenceHandle() or implicitly via open (=> struct file/FsFileInfo).
 */
void FhgfsInode_getRefIOInfo(FhgfsInode* this, FileHandleType handleType,
   unsigned accessFlags, int* pMaxUsedNodeIndex, RemotingIOInfo* outIOInfo)
{
   FhgfsInodeFileHandle* fileHandle = &this->fileHandles[handleType];

   outIOInfo->app = FhgfsOps_getApp(this->vfs_inode.i_sb);

   outIOInfo->fileHandleID = fileHandle->fileHandleID;
   outIOInfo->pattern = this->pattern;
   outIOInfo->accessFlags = accessFlags;

   outIOInfo->pMaxUsedNodeIndex = pMaxUsedNodeIndex;
}
