#ifndef FHGFSINODE_H_
#define FHGFSINODE_H_

#include <open/common/Common.h>
#include <include_open/common/storage/Open_StorageErrors.h>
#include <include_open/common/toolkit/Open_Time.h>
#include <include_open/filesystem/Open_FhgfsInode.h>
#include <open/common/threading/Mutex.h>
#include "FhgfsOpsSuper.h"

#include <linux/fs.h>
#include <linux/vfs.h>


/*
 * The rest of this is declared in the include_open tree.
 */


#define FHGFS_INODE_ROOT_INO           2 /* traditional root inode number */

#define FHGFS_INODE(inodePointer)      ( (FhgfsInode*)inodePointer)
#define FHGFS_INODE_BLOCKBITS          16 /*PAGE_SHIFT*/ /* debug value */
#define FHGFS_INODE_BLOCKSIZE          (1 << FHGFS_INODE_BLOCKBITS)

#define FHGFS_INODE_FILEHANDLES_NUM    3 /* r, rw, w */


struct StripePattern; // forward declaration



struct FhgfsInodeFileHandle;
typedef struct FhgfsInodeFileHandle FhgfsInodeFileHandle;

struct FhgfsInode;
typedef struct FhgfsInode FhgfsInode;

// mem cache helpers
extern void FhgfsInode_initOnce(FhgfsInode* fhgfsInode);
extern void FhgfsInode_allocInit(FhgfsInode* fhgfsInode);
extern void FhgfsInode_destroyUninit(FhgfsInode* fhgfsInode);


extern FhgfsOpsErr FhgfsInode_referenceHandle(FhgfsInode* this, int openFlags,
   fhgfs_bool allowRWHandle, FileHandleType* outHandleType);
extern FhgfsOpsErr FhgfsInode_releaseHandle(FhgfsInode* this, FileHandleType handleType,
   int maxUsedNodeIndex);

extern FileHandleType __FhgfsInode_openFlagsToHandleType(int openFlags);
extern FhgfsOpsErr __FhgfsInode_referenceTrunc(FhgfsInode* this, App* app, int openFlags);

// getters & setters
extern void FhgfsInode_getOwnerMinInfoCopy(FhgfsInode* this, OwnerMinInfo* outEntryInfo);

// inliners
static inline void FhgfsInode_updateOwnerMinInfoCopy(FhgfsInode* this,
   OwnerMinInfo* entryInfo);
static inline void __FhgfsInode_initIOInfo(FhgfsInode* this, FhgfsInodeFileHandle* fileHandle,
   unsigned accessFlags, RemotingIOInfo* outIOInfo);
static inline int FhgfsInode_handleTypeToOpenFlags(FileHandleType handleType);



struct FhgfsInodeFileHandle
{
   const char* fileHandleID;
   unsigned refCount;
   int maxUsedNodeIndex; // zero-based pattern index (-1 means "none")
};

struct FhgfsInode
{
   struct inode vfs_inode;

   spinlock_t entryInfoLock;
   OwnerMinInfo entryInfo;

   Time dataCacheTime; // last time we updated file contents (monotonic clock) or dir attribs;
                       // protected by i_lock

   struct semaphore fileHandlesMutex;
   FhgfsInodeFileHandle fileHandles[FHGFS_INODE_FILEHANDLES_NUM]; // use FileHandleType as index
   struct StripePattern* pattern; // initialized when file is opened; multiple readers allowed
};

/**
 * Note: calls kfree on the former entry values, so make sure they have been initialized before
 * you call this (e.g. don't call it for new inodes).
 *
 * @param entryInfo values (not the entryInfo struct itself) will be owned by the inode
 */
void FhgfsInode_updateOwnerMinInfoCopy(FhgfsInode* this, OwnerMinInfo* entryInfo)
{
   if(this->vfs_inode.i_ino == FHGFS_INODE_ROOT_INO)
   { // root node => owner info is not stored (see _getOwnerMinInfoCopy() )
      kfree(entryInfo->ownerNodeID);
      kfree(entryInfo->entryID);
   }
   else
   { // "normal" inode
      spin_lock(&this->entryInfoLock);

      kfree(this->entryInfo.ownerNodeID);
      kfree(this->entryInfo.entryID);

      this->entryInfo = *entryInfo;

      spin_unlock(&this->entryInfoLock);
   }
}

/**
 * Note: For internal use only (assigns filehandle->maxUsedNodeIndex pointer to outIOInfo).
 */
void __FhgfsInode_initIOInfo(FhgfsInode* this, FhgfsInodeFileHandle* fileHandle,
   unsigned accessFlags, RemotingIOInfo* outIOInfo)
{
   outIOInfo->app = FhgfsOps_getApp(this->vfs_inode.i_sb);

   outIOInfo->fileHandleID = fileHandle->fileHandleID;
   outIOInfo->pattern = this->pattern;
   outIOInfo->accessFlags = accessFlags;

   outIOInfo->pMaxUsedNodeIndex = &fileHandle->maxUsedNodeIndex;
}


/**
 * Converts FileHandleType to OPENFILE_ACCESS_... fhgfs flags.
 *
 * Note: This is especially required when a handle was referenced with allowRWHandle, because then
 * you don't know the OPENFILE_ACCESS_... flags.
 */
int FhgfsInode_handleTypeToOpenFlags(FileHandleType handleType)
{
   switch(handleType)
   {
      case FileHandleType_WRITE:
      {
         return OPENFILE_ACCESS_WRITE;
      } break;

      case FileHandleType_RW:
      {
         return OPENFILE_ACCESS_READWRITE;
      } break;

      default:
      {
         return OPENFILE_ACCESS_READ;
      } break;
   }
}


#endif /* FHGFSINODE_H_ */
