#include <include_closed/app/App.h>
#include <include_open/common/threading/Open_Mutex.h>
#include <include_open/common/Open_Common.h>
#include <include_open/common/toolkit/Open_Time.h>
#include <include_closed/net/filesystem/FhgfsOpsRemoting.h>
#include <include_closed/filesystem/FsDirInfo.h>
#include <include_closed/filesystem/FsFileInfo.h>
#include <include_closed/toolkit/NoAllocBufferStore.h>
#include <include_closed/common/toolkit/MetadataTk.h>
#include <open/common/Common.h>
#include "FhgfsOps_versions.h"
#include "FhgfsOpsInode.h"
#include "FhgfsOpsDir.h"
#include "FhgfsOpsSuper.h"


struct dentry_operations fhgfs_dentry_ops =
{
   .d_revalidate   = FhgfsOps_revalidate,
   .d_delete       = FhgfsOps_deleteDentry,
};


/**
 * Note: Called when the dcache has a lookup hit (and wants to know whether the cache data
 * is still valid)
 * @return value is quasi-boolean: 0 if entry invalid, 1 if still valid (no other return values
 * allowed).
 */
int FhgfsOps_revalidate(struct dentry* dentry, struct nameidata* nameidata)
{
   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_revalidate";

   int isValid = 0; // quasi-boolean (return value)
   struct dentry* parent = dget_parent(dentry);
   struct inode* parentDir = parent->d_inode;
   FhgfsInode* parentFhgfsInode = FHGFS_INODE(parentDir);
   struct inode* inode = dentry->d_inode;
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);
   const char* entryName = dentry->d_name.name;

   if(unlikely(Logger_getLogLevel(log) >= 5) )
   {
      NoAllocBufferStore* bufStore = App_getPathBufStore(app);
      char* pathStoreBuf;
      char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);

      Logger_logFormatted(log, 5, logContext, "called. Path: %s", path);

      NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);
   }

   if(!inode || !parentDir || is_bad_inode(inode) )
   {
      if(inode && S_ISDIR(inode->i_mode) )
      {
         if(have_submounts(dentry) )
            goto cleanup_put_parent;

         shrink_dcache_parent(dentry);
      }

      goto drop_dentry;
   }

   // active dentry => remote-stat and local-compare

   // note: changed root node will be detected during refresh, either because root node ID is
   // unknown and cannot be referenced as node or because the node sends a not-owner reply

   if(!IS_ROOT(dentry) )
   { // check if this is still a subentry of the given parent
      FhgfsOpsErr subentryRes;
      OwnerMinInfo parentInfo;
      OwnerMinInfo entryInfo;

      FhgfsInode_getOwnerMinInfoCopy(parentFhgfsInode, &parentInfo);
      FhgfsInode_getOwnerMinInfoCopy(fhgfsInode, &entryInfo);

      subentryRes = FhgfsOpsRemoting_validateSubentry(app, &parentInfo, entryName, &entryInfo);

      FhgfsInode_freeOwnerMinInfoVals(&parentInfo);
      FhgfsInode_freeOwnerMinInfoVals(&entryInfo);

      if(subentryRes != FhgfsOpsErr_SUCCESS)
         goto drop_dentry;
   }

   if(S_ISDIR(inode->i_mode) )
      isValid = (__FhgfsOps_revalidateDirInode(app, inode) == 0);
   else
      isValid = (__FhgfsOps_refreshInode(app, inode) == 0);

   if(!isValid)
   {
      drop_dentry:
         d_drop(dentry);
   }


cleanup_put_parent:
   // clean-up
   dput(parent);

   LOG_DEBUG_FORMATTED(log, 5, logContext, "isValid: %s", isValid ? "yes" : "no");

   return isValid;
}

/**
 * Note: This is called from dput() when d_count is going to 0.
 * Note: We use this to unhash dentries with bad inodes.
 */
int FhgfsOps_deleteDentry(struct dentry* dentry)
{
   // note: dput just wants to know from us whether or not it should delete the dentry
   int shouldBeDeleted = 0; // quasi-boolean (return value)
   struct inode* inode = dentry->d_inode;

   if(inode)
   {
      if(is_bad_inode(inode) )
         shouldBeDeleted = 1; // unhash the "bad inode"-dentry
   }
   else
   {
      // think about: we could unhash negative dentries here
   }

   return shouldBeDeleted;
}

/*
 * Constructs the path from the root dentry (of the mount-point) to an arbitrary hashed dentry.
 *
 * Note: Acquires a buf from the pathBufStore that must be released by the caller.
 * Note: This is safe for paths larger than bufSize, but returns -ENAMETOOLONG in that case
 *
 * @outBuf the buf that must be returned to the pathBufStore of the app
 * @return some offset within the outStoreBuf
 */
char* __FhgfsOps_pathResolveToStoreBuf(NoAllocBufferStore* bufStore, const struct dentry* dentry,
   char** outStoreBuf)
{
   const ssize_t storeBufLen = NoAllocBufferStore_getBufSize(bufStore);
   ssize_t bufLenLeft; // remaining length
   char* currentBufStart;
   int nameLen;

   bufLenLeft = storeBufLen;
   *outStoreBuf = NoAllocBufferStore_waitForBuf(bufStore);

   currentBufStart = (*outStoreBuf) + bufLenLeft;

   *--currentBufStart = '\0';
   bufLenLeft--;

   spin_lock(&dcache_lock);

   while(!IS_ROOT(dentry) )
   {
      nameLen = dentry->d_name.len;
      bufLenLeft -= nameLen + 1;

      if (bufLenLeft < 0)
         goto err_too_long_unlock;

      currentBufStart -= nameLen;

      memcpy(currentBufStart, dentry->d_name.name, nameLen);

      *--currentBufStart = '/';

      dentry = dentry->d_parent;
   }

   spin_unlock(&dcache_lock);

   // return "/" instead of an empty string for the root directory
   if(bufLenLeft == (storeBufLen-1) ) // -1 for the terminating zero
   { // dentry is (mount) root directory
      *--currentBufStart = '/';
   }

   return currentBufStart;

err_too_long_unlock:
   spin_unlock(&dcache_lock);

   return ERR_PTR(-ENAMETOOLONG);
}

// Note: Not needed currently (and d_path() has changed in kernel 2.6.25)
///**
// * Allocates a page that must be free_page'd by the caller.
// * Note: unlike the other path resolve method, this version retrieves the full path (including
// * the path to the mount point)
// *
// * @param outPageBuf the page that must be free_page'd
// * @return points to some location inside the pageBuf
// */
//char* __FhgfsOps_fullpathResolveToPage(struct dentry* dentry, struct vfsmount* mnt,
//   char** outPageBuf)
//{
//   char* resolvedPath;
//
//   *outPageBuf = (char*)__get_free_page(GFP_NOFS);
//
//   resolvedPath = d_path(dentry, mnt, *outPageBuf, PAGE_SIZE);
//
//   return resolvedPath;
//}

