#include <include_closed/app/log/Logger.h>
#include <include_closed/app/App.h>
#include <include_closed/app/config/Config.h>
#include <include_closed/common/toolkit/vector/StrCpyVec.h>
#include <include_open/common/storage/Open_StorageErrors.h>
#include <include_open/common/toolkit/Open_StringTk.h>
#include <include_closed/filesystem/FhgfsOpsHelper.h>
#include <open/os/OsTypeConversion.h>
#include <open/filesystem/ProcFs.h>
#include "FhgfsOpsFile.h"
#include "FhgfsOpsDir.h"
#include "FhgfsOpsInode.h"
#include "FhgfsOpsSuper.h"
#include "FhgfsOps_versions.h"

#include <linux/writeback.h>
#include <linux/mm.h>
#include <linux/mpage.h>
#include <linux/backing-dev.h>
#include <linux/pagemap.h>


#define FHGFSOPS_WRITEPAGES_POINTER_RESERVE   (1024) /* number of page pointers in write buffer */


struct file_operations fhgfs_file_ops =
{
   .open       = FhgfsOps_open,
   .release    = FhgfsOps_release,
   .read       = FhgfsOps_read,
   .write      = FhgfsOps_write,
   .fsync      = FhgfsOps_fsync,
   .flush      = FhgfsOps_flush,
   .llseek     = FhgfsOps_llseek,
   .mmap       = FhgfsOps_mmap,

   #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,17)
      .splice_read  = generic_file_splice_read,
      .splice_write = generic_file_splice_write,
   #endif // LINUX_VERSION_CODE

   #if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,22)
      .sendfile   = generic_file_sendfile, // removed in 2.6.23 (now handled via splice)
   #endif // LINUX_VERSION_CODE
};

struct file_operations fhgfs_file_pagecache_ops =
{
   .open       = FhgfsOps_open,
   .release    = FhgfsOps_release,
   .read       = do_sync_read,
   .write      = do_sync_write,
   .aio_read   = FhgfsOps_aio_read,
   .aio_write  = FhgfsOps_aio_write,
   .fsync      = FhgfsOps_fsync,
   .flush      = FhgfsOps_flush,
   .llseek     = FhgfsOps_llseek,
   .mmap       = FhgfsOps_mmap,

   #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,17)
      .splice_read  = generic_file_splice_read,
      .splice_write = generic_file_splice_write,
   #endif // LINUX_VERSION_CODE

   #if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,22)
      .sendfile   = generic_file_sendfile, // removed in 2.6.23 (now handled via splice)
   #endif // LINUX_VERSION_CODE
};

struct file_operations fhgfs_dir_ops =
{
   .open       = FhgfsOps_opendirIncremental,
   .release    = FhgfsOps_releasedir,
   .readdir    = FhgfsOps_readdirIncremental,
   .read       = generic_read_dir, // just returns the appropriate error code
};

struct address_space_operations fhgfs_address_ops =
{
   .readpage       = FhgfsOps_readpage,
   .writepage      = FhgfsOps_writepage,
   .writepages     = generic_writepages,
   .set_page_dirty = __set_page_dirty_nobuffers,
   //.bmap           = FhgfsOps_bmap, // required for swapon (if used directly - without loop mount)
   .direct_IO      = FhgfsOps_directIO,

   #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,28)
   .prepare_write = FhgfsOps_prepare_write,
   .commit_write  = FhgfsOps_commit_write,
   #else
   .write_begin   = FhgfsOps_write_begin,
   .write_end     = FhgfsOps_write_end,
   #endif // LINUX_VERSION_CODE
};

struct address_space_operations fhgfs_address_pagecache_ops =
{
   .readpage       = FhgfsOps_readpage,
   .readpages      = FhgfsOps_readpages,
   .writepage      = FhgfsOps_writepage,
   .writepages     = FhgfsOps_writepages, //generic_writepages,
   .set_page_dirty = __set_page_dirty_nobuffers,
   //.bmap           = FhgfsOps_bmap, // required for swapon, if used directly (without loop mount)
   .direct_IO      = FhgfsOps_directIO,

   #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,28)
   .prepare_write = FhgfsOps_prepare_write,
   .commit_write  = FhgfsOps_commit_write,
   #else
   .write_begin   = FhgfsOps_write_begin,
   .write_end     = FhgfsOps_write_end,
   #endif // LINUX_VERSION_CODE
};


loff_t FhgfsOps_llseek(struct file *file, loff_t offset, int origin)
{
   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_llseek";

   loff_t retVal = 0;
   struct inode *inode = file->f_mapping->host;

   if(unlikely(Logger_getLogLevel(log) >= 5) )
   {
      NoAllocBufferStore* bufStore = App_getPathBufStore(app);
      char* pathStoreBuf;
      char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, file->f_dentry, &pathStoreBuf);

      Logger_logFormatted(log, 5, logContext, "called. Path: %s", path);

      NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);
   }


   if(origin == SEEK_END)
   { // seek to position relative to end-of-file => update current file size first
      retVal = __FhgfsOps_refreshInode(app, inode);
      if(!retVal)
      { // success
         spin_lock(&inode->i_lock);
         retVal = generic_file_llseek_unlocked(file, offset, origin);
         spin_unlock(&inode->i_lock);
      }
   }
   else
   { // abolute or relative-to-current_pos seeks => generic stuff
      retVal = generic_file_llseek_unlocked(file, offset, origin);
   }

   return retVal;
}


int FhgfsOps_readlink(struct dentry* dentry, char __user* buf, int size)
{
   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_readlink";

   int retVal;
   struct inode* inode = dentry->d_inode;
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);
   OwnerMinInfo entryInfo;

   if(unlikely(Logger_getLogLevel(log) >= 5) )
   {
      NoAllocBufferStore* bufStore = App_getPathBufStore(app);
      char* pathStoreBuf;
      char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);

      Logger_logFormatted(log, 5, logContext, "called. Path: %s", path);

      NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);
   }

   // check user buffer
   if(unlikely(!access_ok(VERIFY_WRITE, buf, size) ) )
      return -EFAULT;

   FhgfsInode_getOwnerMinInfoCopy(fhgfsInode, &entryInfo);

   retVal = FhgfsOpsHelper_readlink(app, &entryInfo, buf, size);


   // clean up
   FhgfsInode_freeOwnerMinInfoVals(&entryInfo);

   return retVal;
}

/**
 * Opens a directory and prepares the handle for incremental readdir().
 */
int FhgfsOps_opendirIncremental(struct inode* inode, struct file* file)
{
   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_opendirIncremental";

   int retVal = 0;
   //struct dentry* dentry = file->f_dentry;
   FsDirInfo* dirInfo;

   if(unlikely(Logger_getLogLevel(log) >= 5) )
   {
      NoAllocBufferStore* bufStore = App_getPathBufStore(app);
      char* pathStoreBuf;
      char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, file->f_dentry, &pathStoreBuf);

      Logger_logFormatted(log, 5, logContext, "called. Path: %s", path);

      NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);
   }

   //retVal = __FhgfsOps_refreshInode(app, inode); // not necessary
   if(!retVal)
   { // success
      dirInfo = FsDirInfo_construct(app);
      __FhgfsOps_setDirInfo(dirInfo, file);
   }

   return retVal;
}


int FhgfsOps_readdirIncremental(struct file* file, void* buf, filldir_t filldir)
{
   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_readdirIncremental";

   int retVal = 0;
   FsDirInfo* dirInfo = __FhgfsOps_getDirInfo(file);
   struct dentry* dentry = file->f_dentry;
   struct inode* inode = dentry->d_inode;
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);
   OwnerMinInfo entryInfo;
   const char* entryName = dentry->d_name.name;

   if(unlikely(Logger_getLogLevel(log) >= 5) )
   {
      NoAllocBufferStore* bufStore = App_getPathBufStore(app);
      char* pathStoreBuf;
      char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, file->f_dentry, &pathStoreBuf);

      Logger_logFormatted(log, 5, logContext, "called. Path: %s", path);

      NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);
   }

   FhgfsInode_getOwnerMinInfoCopy(fhgfsInode, &entryInfo);

   //log.log(5, std::string("dirContents.size: ") + StringTk::intToStr(dirContents->size() ) );;
   //log.log(5, std::string("offset: ") + StringTk::intToStr(offset) );;

   if(file->f_pos < FsDirInfo_getOffset(dirInfo) )
   { // user seeked backwards => re-read the dir contents
      int rereadRes;
      loff_t currentOffset = (file->f_pos >= 2) ? (file->f_pos - 2) : 0; // avoid negative values;
         // -2 because of "." and ".."

      LOG_DEBUG_FORMATTED(log, 5, logContext, "Seeked backwards => re-reading directory: '%s' "
         "(old offset: %lld; new: %lld)", entryName, (long long)FsDirInfo_getOffset(dirInfo),
         (long long)file->f_pos);
      IGNORE_UNUSED_VARIABLE(entryName);

      rereadRes = FhgfsOpsHelper_refreshDirInfoIncremental(app, &entryInfo, currentOffset, dirInfo,
         fhgfs_true);

      //LOG_DEBUG_FORMATTED(log, 4, logContext, "remoting complete. result: %d", (int)rereadRes);

      if(rereadRes)
      { // error occured
         retVal = rereadRes;
         goto clean_up;
      }

      dirInfo = __FhgfsOps_getDirInfo(file);
   }

   switch(file->f_pos)
   {
      case 0:
      {
         ino_t ino = dentry->d_inode->i_ino;
         if(filldir(buf, ".", 1, file->f_pos, ino, DT_DIR) < 0)
            break;
         file->f_pos++;

         LOG_DEBUG_FORMATTED(log, 5, logContext, "filled[%lld]: %s",
            (long long)0, ".");
      }
         /* fall through */
      case 1:
      {
         ino_t inoParent = parent_ino(dentry);
         if(filldir(buf, "..", 2, file->f_pos, inoParent, DT_DIR) < 0)
            break;
         file->f_pos++;

         LOG_DEBUG_FORMATTED(log, 5, logContext, "filled[%lld]: %s",
            (long long)1, "..");
      }
         /* fall through */
      default:
      {
         loff_t currentOffset = file->f_pos - 2; // -2 because of "." and ".."
         StrCpyVec* dirContents = FsDirInfo_getDirContents(dirInfo);
         //size_t dirContentsSize = StrCpyVec_length(dirContents);
         loff_t contentsOffset;
         size_t contentsLength;

         for( ; ; )
         {
            int refreshRes;
            char* currentName;

            refreshRes = FhgfsOpsHelper_refreshDirInfoIncremental(app, &entryInfo,
               currentOffset, dirInfo, fhgfs_false);
            if(refreshRes)
            { // error occured
               retVal = refreshRes;
               break;
            }

            contentsOffset = FsDirInfo_getContentsOffset(dirInfo);
            contentsLength = StrCpyVec_length(dirContents);

            LOG_DEBUG_FORMATTED(log, 5, logContext,
               "stage: %d, currentOffset: %lld, contentsOffset: %lld, contentsLen: %lld",
               (int)FsDirInfo_getStage(dirInfo),
               (long long)currentOffset, (long long)contentsOffset, (long long)contentsLength);


            if( (FsDirInfo_getStage(dirInfo) == ListDirIncStage_END) &&
               (currentOffset >= (contentsOffset + contentsLength) ) )
            {
               LOG_DEBUG(log, 5, logContext, "reached end of dir");
               break;
            }

            currentName = StrCpyVec_at(
               dirContents, currentOffset - contentsOffset);

            //log.log(5, std::string("current iter value: ") + *iter);

            if(filldir(buf, currentName, strlen(currentName), file->f_pos, ~0, DT_UNKNOWN) < 0)
               break;

            LOG_DEBUG_FORMATTED(log, 5, logContext, "filled[%lld]: %s",
               (long long)currentOffset, currentName);

            file->f_pos++;
            currentOffset++;
         }
      }
   } // end of switch

   FsDirInfo_setOffset(dirInfo, file->f_pos);

clean_up:
   // clean-up
   FhgfsInode_freeOwnerMinInfoVals(&entryInfo);

   return retVal;
}

/**
 * Note: This works for _opendir() and for _opendirIncremental().
 */
int FhgfsOps_releasedir(struct inode* inode, struct file* file)
{
   const char* logContext = "FhgfsOps_releasedir";

   FsDirInfo* dirInfo = __FhgfsOps_getDirInfo(file);
   FsObjectInfo* fsObjectInfo = __FhgfsOps_getObjectInfo(file);

   App* app = FsObjectInfo_getApp(fsObjectInfo);
   Logger* log = App_getLogger(app);

   Logger_logFormatted(log, 5, logContext, "called. Path: n/a");


   dirInfo = __FhgfsOps_getDirInfo(file);
   FsObjectInfo_virtualDestruct(fsObjectInfo);

   // clean-up

   return 0;
}

/**
 * Open a file.
 */
int FhgfsOps_open(struct inode* inode, struct file* file)
{
   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);
   Config* cfg = App_getConfig(app);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_open";

   //struct dentry* dentry = d_find_alias(inode); // calls __dget_locked()
   struct dentry* dentry = file->f_dentry;

   int retVal = 0;
   int fhgfsOpenFlags;
   FileHandleType handleType;
   FhgfsOpsErr openRes;
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);

   if(unlikely(Logger_getLogLevel(log) >= 4) )
   {
      NoAllocBufferStore* bufStore = App_getPathBufStore(app);
      char* pathStoreBuf;
      char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);

      Logger_logFormatted(log, 4, logContext, "called. Path: %s", path);

      NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);
   }

   fhgfsOpenFlags = OsTypeConv_openFlagsOsToFhgfs(file->f_flags);

   openRes = FhgfsInode_referenceHandle(fhgfsInode, fhgfsOpenFlags, fhgfs_false, &handleType);

   LOG_DEBUG_FORMATTED(log, 4, logContext, "remoting complete. result: %s",
      FhgfsOpsErr_toErrString(openRes) );

   if(openRes != FhgfsOpsErr_SUCCESS)
   { // error
      retVal = FhgfsOpsErr_toSysErr(openRes);
   }
   else
   { // success => file is open (=> handle open flags)
      FsFileInfo* fileInfo = FsFileInfo_construct(app, fhgfsOpenFlags, handleType);

      // handle O_APPEND
      if(file->f_flags & O_APPEND)
         FsFileInfo_setAppending(fileInfo, fhgfs_true);

      // handle O_DIRECT + disabled caching
      if( (file->f_flags & O_DIRECT) ||
          (Config_getTuneFileCacheTypeNum(cfg) == FILECACHETYPE_None) )
      {
         FsFileInfo_setAllowCaching(fileInfo, fhgfs_false);
      }

      __FhgfsOps_setFileInfo(fileInfo, file);
   }

   return retVal;
}

/**
 * Close a file.
 *
 * Note: We only got one shot, even in case of an error.
 */
int FhgfsOps_release(struct inode* inode, struct file* file)
{
   const char* logContext = "FhgfsOps_release";

   int retVal = 0;
   FhgfsOpsErr closeRes;
   FsFileInfo* fileInfo = __FhgfsOps_getFileInfo(file);
   FsObjectInfo* fsObjectInfo = __FhgfsOps_getObjectInfo(file);
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);

   FileHandleType handleType = FsFileInfo_getHandleType(fileInfo);
   int maxUsedNodeIndex = FsFileInfo_getMaxUsedNodeIndex(fileInfo);

   App* app = FsObjectInfo_getApp(fsObjectInfo);
   Logger* log = App_getLogger(app);

   Logger_logFormatted(log, 4, logContext, "called. Path: n/a");

   if(unlikely(!fileInfo) )
   { // invalid file handle
      return -EBADF;
   }

   closeRes = FhgfsInode_releaseHandle(fhgfsInode, handleType, maxUsedNodeIndex);

   LOG_DEBUG_FORMATTED(log, 4, logContext, "remoting complete. result: %s",
      FhgfsOpsErr_toErrString(closeRes) );

   if(closeRes != FhgfsOpsErr_SUCCESS)
   { // error
      retVal = FhgfsOpsErr_toSysErr(closeRes);
   }

   // note: we free the fileInfo no matter whether the communication succeeded or not
   //    (because _release() won't be called again even if it didn't succeed)

   FsObjectInfo_virtualDestruct( (FsObjectInfo*)fileInfo);
   __FhgfsOps_setFileInfo( (FsFileInfo*)NULL, file);

   return retVal;
}

ssize_t FhgfsOps_read(struct file* file, char __user *buf, size_t size, loff_t* offsetPointer)
{
   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);

   struct inode* inode = file->f_mapping->host;
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);
   RemotingIOInfo ioInfo;
   FsFileInfo* fileInfo = __FhgfsOps_getFileInfo(file);
   ssize_t readRes;

   #ifdef LOG_DEBUG_MESSAGES

      Logger* log = App_getLogger(app);
      const char* logContext = "FhgfsOps_read";

      NoAllocBufferStore* bufStore = App_getPathBufStore(app);
      char* pathStoreBuf;
      char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, file->f_dentry, &pathStoreBuf);

      LOG_DEBUG_FORMATTED(log, 5, logContext, "called. Path: %s (offset: %lld; size: %lld)",
         path, (long long)*offsetPointer, (long long)size);

      // clean-up
      NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);

   #endif // LOG_DEBUG_MESSAGES


   FsFileInfo_getIOInfo(fileInfo, fhgfsInode, &ioInfo);

   readRes = FhgfsOpsHelper_readCached(app, buf, size, *offsetPointer, fileInfo, &ioInfo);
   //readRes = FhgfsOpsRemoting_readfile(buf, size, *offsetPointer, &ioInfo);
   //readRes = FhgfsOpsRemoting_readfileThreaded(buf, size, *offsetPointer, &ioInfo);

   if(unlikely(readRes < 0) )
   { // error
      return FhgfsOpsErr_toSysErr(-readRes);
   }

   *offsetPointer += readRes;

   return readRes;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)

ssize_t FhgfsOps_aio_read(struct kiocb *iocb, char __user *buf, size_t count, loff_t pos)
{

#else

ssize_t FhgfsOps_aio_read(struct kiocb *iocb, const struct iovec *iov, unsigned long nr_segs,
   loff_t pos)
{
   size_t count = iov_length(iov, nr_segs);

#endif // LINUX_VERSION_CODE

   struct file* file = iocb->ki_filp;
   struct address_space* mapping = file->f_mapping;
   struct inode* inode = mapping->host; // (alternative: file->f_dentry->d_inode?)

   App* app = FhgfsOps_getApp(inode->i_sb);

   ssize_t retVal;

#ifdef LOG_DEBUG_MESSAGES

   struct dentry* dentry = file->f_dentry;

   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_aio_read";

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBuf;
   char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);

   LOG_DEBUG_FORMATTED(log, 5, logContext, "called. Path: %s (offset: %lld; size: %lld)",
      path, (long long)pos, (long long)count);

   // clean-up
   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);

#endif // LOG_DEBUG_MESSAGES
   
   IGNORE_UNUSED_VARIABLE(count);

   retVal = __FhgfsOps_revalidateMapping(app, inode);
   if(unlikely(retVal) )
   { // error
      return retVal;
   }

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
   retVal = generic_file_aio_read(iocb, buf, count, pos);
#else
   retVal = generic_file_aio_read(iocb, iov, nr_segs, pos);
#endif // LINUX_VERSION_CODE

   return retVal;
}


ssize_t FhgfsOps_write(struct file* file, const char __user *buf, size_t size,
   loff_t* offsetPointer)
{
   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);

   struct inode* inode = file->f_mapping->host;
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);
   RemotingIOInfo ioInfo;
   FsFileInfo* fileInfo = __FhgfsOps_getFileInfo(file);
   ssize_t writeRes;

   #ifdef LOG_DEBUG_MESSAGES

      Logger* log = App_getLogger(app);
      const char* logContext = "FhgfsOps_write";

      NoAllocBufferStore* bufStore = App_getPathBufStore(app);
      char* pathStoreBuf;
      char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, file->f_dentry, &pathStoreBuf);

      LOG_DEBUG_FORMATTED(log, 5, logContext, "called. Path: %s (offset: %lld; size: %lld)",
         path, (long long)size, (long long)*offsetPointer);

      // clean-up
      NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);

   #endif // LOG_DEBUG_MESSAGES


   if(FsFileInfo_getAppending(fileInfo) )
   { // appending => move file offset to end-of-file before writing
      FhgfsOpsErr statRes;
      fhgfs_stat fhgfsStat;

      NoAllocBufferStore* bufStore = App_getPathBufStore(app);
      char* pathStoreBuf;
      char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, file->f_dentry, &pathStoreBuf);

      statRes = FhgfsOpsRemoting_stat(app, path, &fhgfsStat);
      if(statRes != FhgfsOpsErr_SUCCESS)
         return FhgfsOpsErr_toSysErr(statRes);

      *offsetPointer = fhgfsStat.size;

      // clean-up
      NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);
   }

   FsFileInfo_getIOInfo(fileInfo, fhgfsInode, &ioInfo);

   writeRes = FhgfsOpsHelper_writeCached(app, buf, size, *offsetPointer, fileInfo, &ioInfo);
   //writeRes = FhgfsOpsRemoting_writefile(buf, size, *offsetPointer, &ioInfo);
   //writeRes = FhgfsOpsRemoting_writefileThreaded(buf, size, *offsetPointer, &ioInfo);

   if(unlikely(writeRes < 0) )
   { // error
      return FhgfsOpsErr_toSysErr(-writeRes);
   }

   *offsetPointer += writeRes;

   return writeRes;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)

ssize_t FhgfsOps_aio_write(struct kiocb *iocb, const char __user *buf, size_t count, loff_t pos)
{

#else

ssize_t FhgfsOps_aio_write(struct kiocb *iocb, const struct iovec *iov, unsigned long nr_segs,
   loff_t pos)
{
   size_t count = iov_length(iov, nr_segs);

#endif // LINUX_VERSION_CODE

   struct dentry* dentry = iocb->ki_filp->f_dentry;
   struct inode* inode = dentry->d_inode;

   App* app = FhgfsOps_getApp(dentry->d_sb);

   ssize_t retVal;


#ifdef LOG_DEBUG_MESSAGES

   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_aio_write";

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBuf;
   char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);

   LOG_DEBUG_FORMATTED(log, 5, logContext, "called. Path: %s (offset: %lld; size: %lld)",
      path, (long long)pos, (long long)count);

   // clean-up
   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);

#endif // LOG_DEBUG_MESSAGES


   if(iocb->ki_filp->f_flags & O_APPEND)
   { // O_APPEND => refresh file size
      retVal = __FhgfsOps_refreshInode(app, inode);

      if(retVal)
         return retVal;
   }

   if(!count)
      return 0;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
   retVal = generic_file_aio_write(iocb, buf, count, pos);
#else
   retVal = generic_file_aio_write(iocb, iov, nr_segs, pos);
#endif // LINUX_VERSION_CODE

   if( (retVal >= 0) &&
       (IS_SYNC(inode) || (iocb->ki_filp->f_flags & O_SYNC) ) )
   { // sync I/O => flush and wait
      struct address_space* mapping = inode->i_mapping;

      if(mapping->nrpages)
      {
         int writeRes = filemap_fdatawrite(mapping);
         if(writeRes >= 0)
         {
            int waitRes = filemap_fdatawait(mapping);
            if(waitRes < 0)
               retVal = waitRes;
         }
      }
   } // end of if(sync)

   return retVal;
}



int FhgfsOps_fsync(struct file* file, struct dentry* dentry, int datasync)
{
   App* app = FhgfsOps_getApp(dentry->d_sb);
   Config* cfg = App_getConfig(app);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_fsync";

   int retVal = 0;
   struct inode* inode = file->f_dentry->d_inode;
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);
   RemotingIOInfo ioInfo;
   FhgfsOpsErr flushRes;
   FsObjectInfo* fsObjectInfo;
   FsFileInfo* fileInfo;

   if(unlikely(Logger_getLogLevel(log) >= 5) )
   {
      NoAllocBufferStore* bufStore = App_getPathBufStore(app);
      char* pathStoreBuf;
      char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);

      Logger_logFormatted(log, 5, logContext, "called. Path: %s", path);

      NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);
   }

   fsObjectInfo = __FhgfsOps_getObjectInfo(file);

   // syncing something other than a file?
   if(FsObjectInfo_getObjectType(fsObjectInfo) != FsObjectType_FILE)
      goto clean_up;

   fileInfo = __FhgfsOps_getFileInfo(file);

   retVal = filemap_write_and_wait(inode->i_mapping);
   if(retVal)
      goto clean_up;


   FsFileInfo_getIOInfo(fileInfo, fhgfsInode, &ioInfo);

   flushRes = FhgfsOpsHelper_flushCache(app, fileInfo, &ioInfo, fhgfs_false);
   if(flushRes != FhgfsOpsErr_SUCCESS)
   {
      retVal = FhgfsOpsErr_toSysErr(flushRes);
      goto clean_up;
   }

   if(Config_getTuneRemoteFSync(cfg) )
   {
      FhgfsOpsErr fsyncRes = FhgfsOpsRemoting_fsyncfile(&ioInfo);
      if(fsyncRes != FhgfsOpsErr_SUCCESS)
      {
         retVal = FhgfsOpsErr_toSysErr(fsyncRes);
         goto clean_up;
      }
   }

clean_up:
   // clean-up

   return retVal;
}

int FhgfsOps_mmap(struct file* file, struct vm_area_struct* vma)
{
   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_mmap";

   int retVal;
   struct inode* inode = file->f_dentry->d_inode;

   if(unlikely(Logger_getLogLevel(log) >= 5) )
   {
      NoAllocBufferStore* bufStore = App_getPathBufStore(app);
      char* pathStoreBuf;
      char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, file->f_dentry, &pathStoreBuf);

      Logger_logFormatted(log, 5, logContext, "called. Path: %s", path);

      NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);
   }

   retVal = generic_file_mmap(file, vma);
   if(!retVal)
   {
      retVal = __FhgfsOps_refreshInode(app, inode);
   }

   LOG_DEBUG_FORMATTED(log, 5, logContext, "result: %d", retVal);

   return retVal;
}

/*
 * Note: Called with the page locked and we need to unlock it when done.
 *
 * @return 0 on success, negative linux error code otherwise
 */
int FhgfsOps_readpage(struct file* file, struct page* page)
{
   loff_t offset = page_offset(page);

   int retVal;

#ifdef LOG_DEBUG_MESSAGES
   const char* logContext = "FhgfsOps_readpage";
   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);
   Logger* log = App_getLogger(app);

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBuf;
   char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, file->f_dentry, &pathStoreBuf);

   LOG_DEBUG_FORMATTED(log, 5, logContext, "called. Path: %s", path);

   // clean-up
   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);
#endif // LOG_DEBUG_MESSAGES


   retVal = __FhgfsOps_readpage_sync(file, page, offset);

   unlock_page(page);

   return retVal;
}

/**
 * @return 0 on success, negative linux error code otherwise
 */
int FhgfsOps_readpages(struct file* file, struct address_space* mapping,
   struct list_head* page_list, unsigned num_pages)
{
   int retVal = -EACCES;

   struct inode* inode = mapping->host;
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);
   RemotingIOInfo ioInfo;
   FsFileInfo* fileInfo = __FhgfsOps_getFileInfo(file);

   struct pagevec lru_pvec;
   char* buf = NULL;
   size_t bufSize = 0; // size of the currently allocated buf

   unsigned i;

#ifdef LOG_DEBUG_MESSAGES
   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);
   const char* logContext = "FhgfsOps_readpages";
   Logger* log = App_getLogger(app);

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBuf;
   char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, file->f_dentry, &pathStoreBuf);

   LOG_DEBUG_FORMATTED(log, 5, logContext, "called. Path: %s (num_pages: %u)", path, num_pages);

   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);
#endif // LOG_DEBUG_MESSAGES


   FsFileInfo_getIOInfo(fileInfo, fhgfsInode, &ioInfo);

   pagevec_init(&lru_pvec, 0);


   // walk over all pages
   for(i=0; i < num_pages; )
   {
      unsigned numContPages = 0; // continuous pages
      struct page* firstPage; // first page of the continuous set
      loff_t offset;
      size_t toBeRead;
      struct page* contCheckPage; // current cont-check walk page
      unsigned long nextContPageIndex;
      ssize_t readRes;

      if(list_empty(page_list) )
         break; // no pages remaining (would be strange...)

      firstPage = list_entry(page_list->prev, struct page, lru);
      offset = page_offset(firstPage);;

      // check continuous pages (=> read multiple pages at once)

      // note: first expected next index is the current page index, so it will definitely be a hit
      //    and numContPages will be inc'ed to 1

      nextContPageIndex = list_entry(page_list->prev, struct page, lru)->index;

      list_for_each_entry_reverse(contCheckPage, page_list, lru)
      {
         if(contCheckPage->index == nextContPageIndex)
         { // next index is continuous - good!
            numContPages++;
            nextContPageIndex++;
         }
         else
            break;
      }

      if( (numContPages + i) >  num_pages)
         numContPages = num_pages - i; // we found more pages than requested by caller

      toBeRead = numContPages * PAGE_CACHE_SIZE;

      // prepare read buffer
      if(bufSize < toBeRead)
      {
         SAFE_VFREE(buf);
         buf = vmalloc(toBeRead);
         bufSize = toBeRead;
      }

      LOG_DEBUG_FORMATTED(log, 5, logContext, "toBeRead: %llu;  numContPages: %u",
         toBeRead, numContPages);

      readRes = FhgfsOpsRemoting_readfileThreaded(buf, toBeRead, offset, &ioInfo);
      //readRes = FhgfsOpsRemoting_readfile(buf, toBeRead, offset, fileInfo);

      if(unlikely(readRes < 0) )
      {
         LOG_DEBUG_FORMATTED(log, 1, logContext, "error: %s", FhgfsOpsErr_toErrString(-readRes) );

         retVal = FhgfsOpsErr_toSysErr(-readRes);
         break;
      }
      else
      if(readRes > 0)
      {
         __FhgfsOps_copyToCachePages(mapping, page_list, buf, readRes, &lru_pvec);

         i += (readRes >> PAGE_CACHE_SHIFT);

         if(readRes != toBeRead)
         { // end of file
            LOG_DEBUG_FORMATTED(log, 5, logContext,
               "End of file (readRes < toBeRead) inside page with offset %lld. "
               "(=> skipping remaining pages)",
               offset);
            break;
         }
      }
      else
      { // end of file
         LOG_DEBUG_FORMATTED(log, 5, logContext,
            "End of file (!readRes) at offset %lld. (=> skipping remaining pages)", offset);
         break;
      }

   }

   pagevec_lru_add_file(&lru_pvec);

   // clean-up
   SAFE_VFREE(buf);

   return retVal;
}

/**
 * Note: This also removes the pages from the list and adds them to the page cache.
 *
 * @param page_list dest pages
 * @param srcBuf source buffer
 * @param bufLen srcBuf length
 */
void __FhgfsOps_copyToCachePages(struct address_space* mapping, struct list_head* page_list,
   char* srcBuf, int bufLen, struct pagevec* plru_pvec)
{
   struct page* page;
   char* pageBuf;

   while(bufLen)
   {
      if(list_empty(page_list) ) // no pages remaining (would be strange...)
         break;

      page = list_entry(page_list->prev, struct page, lru);
      list_del(&page->lru);

      if(add_to_page_cache(page, mapping, page->index, GFP_KERNEL) )
      { // adding this one to cache failed => just skip it
         page_cache_release(page);

         srcBuf += PAGE_CACHE_SIZE;
         bufLen -= PAGE_CACHE_SIZE;

         continue;
      }

      pageBuf = kmap_atomic(page, KM_USER0);

      if(PAGE_CACHE_SIZE > bufLen)
      { // partial page => zero-fill the rest of the page
         memcpy(pageBuf, srcBuf, bufLen);
         memset(pageBuf + bufLen, 0, PAGE_CACHE_SIZE - bufLen);

         bufLen = 0;
      }
      else
      { // full page => just copy the current srcBuf part
         memcpy(pageBuf, srcBuf, PAGE_CACHE_SIZE);

         bufLen -= PAGE_CACHE_SIZE;
      }

      kunmap_atomic(pageBuf, KM_USER0);

      flush_dcache_page(page);
      SetPageUptodate(page);
      unlock_page(page);

      if(!pagevec_add(plru_pvec, page) )
         __pagevec_lru_add_file(plru_pvec);

      srcBuf += PAGE_CACHE_SIZE;
   }

}


/**
 * @return 0 on success
 */
int FhgfsOps_writepage(struct page *page, struct writeback_control *wbc)
{
   // Note: this is a writeback method, so we have no file handle here! (only the inode)

   // note: make sure that write-mapping is enabled in fhgfsops_mmap!!!

   const char* logContext = "FhgfsOps_writepage";

   int retVal = 0;

   page_cache_get(page);
   if(!PageUptodate(page) )
   {
      printk_fhgfs_debug(KERN_WARNING, "%s: page not up-to-date (=> partial page write)\n",
         logContext);
      IGNORE_UNUSED_VARIABLE(logContext);
   }

   // set writeback-flag and clear dirty-flag.
   // note: the alternative is redirty_page_for_writepage() in case of an error
   set_page_writeback(page);

   retVal = __FhgfsOps_writepage_async(page, 0, PAGE_CACHE_SIZE);

   SetPageUptodate(page);
   unlock_page(page);
   end_page_writeback(page);
   page_cache_release(page);


   return retVal;
}

/**
 * @return 0 on success, negative linux error code otherwise
 */
int FhgfsOps_writepages(struct address_space* mapping, struct writeback_control* wbc)
{
   App* app = FhgfsOps_getApp(mapping->host->i_sb);
   const char* logContext = "FhgfsOps_writepages";
   Logger* log = App_getLogger(app);

   // note: we require here that the buf size is at least "pagePointerReserveSize + PAGE_CACHE_SIZE"
   NoAllocBufferStore* pagedIOBufStore = App_getPagedIOBufStore(app);
   char* pagedIOBuf; // split into multiple sub-buffers
   const size_t pagedIOBufSize = NoAllocBufferStore_getBufSize(pagedIOBufStore);
   const size_t pagePointerReserveSize = FHGFSOPS_WRITEPAGES_POINTER_RESERVE * sizeof(struct page*);
   const size_t maxNumWritePages = MIN(
      (pagedIOBufSize - pagePointerReserveSize) >> PAGE_CACHE_SHIFT,
      FHGFSOPS_WRITEPAGES_POINTER_RESERVE);

   void** writePages; // the mapped pages that we want to write in one parallel rpc
   char* writeBuf; // the concatenation buffer for the writePages

   int retVal = 0;
   struct backing_dev_info* bdi = mapping->backing_dev_info;

   struct inode* inode = mapping->host;
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);
   FileHandleType handleType;
   fhgfs_bool referenceErrLogged = fhgfs_false; // avoid logging same error multiple times
   FhgfsOpsErr referenceRes = FhgfsOpsErr_INTERNAL; // success/error value controls release call
   FhgfsOpsErr releaseRes;
   int maxUsedNodeIndex = -1;
   RemotingIOInfo ioInfo;

   int done = 0; // loop-terminator (used by inner loop to terminate outer loop) (quasi-boolean)
   int cycled = 0; // used to wrap around once at the end for range_cyclic (quasi-boolean)

   pgoff_t end = -1; // inclusive end page index
   pgoff_t index; // current page index
   pgoff_t writeback_index = 0;
   int range_whole = 0; // 1 if whole file range should be written (quasi-boolean)
   long nr_to_write = wbc->nr_to_write; // number of pages that should be written

   struct pagevec pvec;

   #ifdef LOG_DEBUG_MESSAGES
   {
      const char* logContext = "FhgfsOps_writepages";
      Logger* log = App_getLogger(app);

      struct dentry* dentry = d_find_alias(inode); // calls dget_locked

      if(!dentry)
      {
         LOG_DEBUG_FORMATTED(log, 5, logContext, "called. Path: %s (nr_to_write: %llu = %lluKB)",
            "n/a (no dentry found)", (unsigned long long)nr_to_write,
            (unsigned long long)(nr_to_write << PAGE_CACHE_SHIFT) / 1024);
      }
      else
      {
         NoAllocBufferStore* bufStore = App_getPathBufStore(app);
         char* pathStoreBuf;
         char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);

         LOG_DEBUG_FORMATTED(log, 5, logContext, "called. Path: %s (nr_to_write: %llu = %lluKB)",
            path, (unsigned long long)nr_to_write,
            (unsigned long long)(nr_to_write << PAGE_CACHE_SHIFT) / 1024);

         NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);
         dput(dentry);
      }
   }
   #endif // LOG_DEBUG_MESSAGES


   if(wbc->nonblocking && bdi_write_congested(bdi) )
   {
      wbc->encountered_congestion = 1;
      return 0;
   }

   pagedIOBuf = NoAllocBufferStore_instantBuf(pagedIOBufStore);
   if(!pagedIOBuf)
   {
      if(wbc->sync_mode != WB_SYNC_NONE)
         return generic_writepages(mapping, wbc); // we must start IO immediately

      pagedIOBuf = NoAllocBufferStore_waitForBuf(pagedIOBufStore);
   }

   // init pagedIOBuf sub-buffers
   writePages = (void**)&pagedIOBuf[0];
   writeBuf = &pagedIOBuf[pagePointerReserveSize];

   pagevec_init(&pvec, 0);

   // note: 2.6.18 introduced range_cyclic and renamed start/end to range_start/end.

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,17)
   if(wbc->sync_mode == WB_SYNC_NONE)
   {
      writeback_index = mapping->writeback_index; // start from prev offset
      index = writeback_index;
   }
   else
   { // called for sync => whole file sweep
      index = 0;
      cycled = 1; // already at beginning => no wrap-around needed
      range_whole = 1;
   }

   if(wbc->start || wbc->end)
   {
      index = wbc->start >> PAGE_CACHE_SHIFT;
      end = wbc->end >> PAGE_CACHE_SHIFT;
      cycled = 1;
   }
#else
   if(wbc->range_cyclic)
   {
      writeback_index = mapping->writeback_index; // start from prev offset
      index = writeback_index;

      if(index == 0)
         cycled = 1; // already at beginning => no wrap-around needed

      end = -1; // (note: end is unsigned)
   }
   else
   {
      index = wbc->range_start >> PAGE_CACHE_SHIFT;
      end = wbc->range_end >> PAGE_CACHE_SHIFT;

      if( (wbc->range_start == 0) && (wbc->range_end == FHGFS_LLONG_MAX) )
         range_whole = 1; // whole file sweep

      cycled = 1; // 1 to skip range_cyclic wrap-around
   }
#endif // LINUX_VERSION_CODE

   // one-time loop (for range_cyclic)
   // note: typical kernel code impls this with "goto retry" (but that's too much pasta for us)
   for( ; ; )
   {
      while(!done && (index <= end) )
      {
         int firstPageIndex = -1; // first page index in pvec for rpc group (-1 means "not set yet")
         pgoff_t nextContPageIndex = 0; // (0 means "not set yet")
         unsigned num_pages; // remaining pages
         unsigned i; // walks over number of pages (in pvec)
         struct page* page;
         loff_t offset = 0; // first page offset
         int pageFillLen;
         size_t toBeWritten = 0; // number of rpc bytes to write
         unsigned numWritePages = 0;  // number of rpc pages for write
         ssize_t writeRes = 0;

         num_pages = pagevec_lookup_tag(&pvec, mapping, &index, PAGECACHE_TAG_DIRTY,
            min(end - index, (pgoff_t)PAGEVEC_SIZE - 1) + 1);
         if(!num_pages)
            break;

         for(i = 0; i < num_pages; i++)
         {
            loff_t i_size;

            page = pvec.pages[i];

            // at this point we hold neither mapping->tree_lock nor lock on the page itself:
            // the page may be truncated or invalidated (changing page->mapping to NULL),
            // or even swizzled back from swapper_space to tmpfs file mapping.
            // however, page->index will not change because we have a reference on the page.

            if(page->index > end)
            { // behind end => get out (note: can't be range_cyclic, because then "end == -1")
               done = 1;
               break;
            }

            if(firstPageIndex < 0)
               lock_page(page);
            else
            if(!trylock_page(page) )
               break; // we already got at least one page

            if(page->mapping != mapping)
            { // page concurrently disappeared (has been truncated or invalidated)
               unlock_page(page);
               break;
            }

            if(!PageDirty(page) )
            { // someone already wrote the page
               unlock_page(page);
               break;
            }

            if(nextContPageIndex && (page->index != nextContPageIndex) )
            { // non-consecutive page
               unlock_page(page);
               break;
            }

            if(wbc->sync_mode != WB_SYNC_NONE)
               wait_on_page_writeback(page);

            if(PageWriteback(page) || !clear_page_dirty_for_io(page) )
            {
               unlock_page(page);
               break;
            }

            set_page_writeback(page);

            i_size = mapping->host->i_size;

            if(page_offset(page) >= i_size)
            { // page offset behind end of file
               done = 1;
               unlock_page(page);
               end_page_writeback(page);
               break;
            }

            page_cache_get(page);

            pageFillLen = min(i_size - page_offset(page), (loff_t)PAGE_CACHE_SIZE);

            writePages[numWritePages] = kmap(page);
            memcpy(&writeBuf[numWritePages << PAGE_CACHE_SHIFT], writePages[numWritePages],
               pageFillLen);
            toBeWritten += pageFillLen;
            numWritePages++;

            if(firstPageIndex < 0)
            { // firstPageIndex (and offset) not set yet => init
               firstPageIndex = i;
               offset = page_offset(page);
            }

            nextContPageIndex = page->index + 1;

            if( (numWritePages + 1) >= maxNumWritePages)
               break; // next page would be too much for our buffer

            if(pageFillLen < PAGE_CACHE_SIZE)
               break; // partial page (may only occur at the end of a rpc group)

         } // end of for num_pages loop

         if(numWritePages)
         { // we found some pages that need to be written
            LOG_DEBUG_FORMATTED(log, 5, logContext, "toBeWritten: %llu;  numWritePages: %u",
               toBeWritten, numWritePages);

            if(referenceRes != FhgfsOpsErr_SUCCESS)
            { // no file handle reference yet => get one
               referenceRes = FhgfsInode_referenceHandle(fhgfsInode, OPENFILE_ACCESS_WRITE,
                  fhgfs_true, &handleType);

               if(unlikely(referenceRes != FhgfsOpsErr_SUCCESS) )
               { // error getting reference
                  if(!referenceErrLogged)
                  {
                     Logger_logErrFormatted(log, logContext, "Unable to get file handle. "
                        "ErrMsg: %s", FhgfsOpsErr_toErrString(referenceRes) );
                     referenceErrLogged = fhgfs_true;
                  }

                  retVal = FhgfsOpsErr_toSysErr(referenceRes);
                  mapping_set_error(mapping, retVal);
               }
               else
               { // success => got reference
                  FhgfsInode_getRefIOInfo(fhgfsInode, handleType,
                     FhgfsInode_handleTypeToOpenFlags(handleType), &maxUsedNodeIndex, &ioInfo);
               }
            }

            if(likely(referenceRes == FhgfsOpsErr_SUCCESS) )
            { // remoting
               writeRes = FhgfsOpsRemoting_writefileThreaded(writeBuf, toBeWritten, offset,
                  &ioInfo);
               if(unlikely(writeRes < 0) )
               { // error occured
                  LOG_DEBUG_FORMATTED(log, 1, logContext, "error: %s",
                     FhgfsOpsErr_toErrString(-writeRes) );

                  retVal = FhgfsOpsErr_toSysErr(-writeRes);
                  mapping_set_error(mapping, retVal);
               }
            }

            for(i = 0; i < numWritePages; i++)
            { // release the written pages
               page = pvec.pages[firstPageIndex + i];

               if(retVal)
                  SetPageError(page); // general write error
               else
               if( (writeRes < toBeWritten) &&
                   (page->index >= ( (offset+writeRes) >> PAGE_CACHE_SHIFT) ) )
                  SetPageError(page); // page belongs to unwritten area

               kunmap(page);
               unlock_page(page);
               end_page_writeback(page);
               page_cache_release(page);
            }

            nr_to_write -= numWritePages;

            if(nr_to_write <= 0)
               done = 1; // no pages left => we're done

            index = nextContPageIndex;
         }
         else
         { // "numWritePages==0" => need to re-find the pages we skipped
            index = pvec.pages[0]->index + 1;
         }

         pagevec_release(&pvec);
      } // end of while loop

      if(cycled || done)
         break;

      // range_cyclic: we hit the last page and there is more work to be done
      //    => wrap back to the start of the file
      cycled = 1;
      index = 0;
      end = writeback_index - 1;

   } // end of for (retry) loop


#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,17)

   if(range_whole)
      mapping->writeback_index = index;

   wbc->nr_to_write = nr_to_write;

#else

   #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,28)
      if(!wbc->no_nrwrite_index_update)
   #endif // LINUX_VERSION_CODE
      {
         if(wbc->range_cyclic || (range_whole && (nr_to_write > 0) ) )
            mapping->writeback_index = index;

         wbc->nr_to_write = nr_to_write;
      }

#endif // LINUX_VERSION_CODE


   // clean-up

   if(referenceRes == FhgfsOpsErr_SUCCESS)
   { // we got a reference => release it
      releaseRes = FhgfsInode_releaseHandle(fhgfsInode, handleType, *ioInfo.pMaxUsedNodeIndex);
      if(unlikely(releaseRes != FhgfsOpsErr_SUCCESS) )
      { // error getting file handle
         Logger_logErrFormatted(log, logContext, "Unable to release file handle. ErrMsg: %s",
            FhgfsOpsErr_toErrString(releaseRes) );

         if(!retVal)
         { // no other error yet
            retVal = FhgfsOpsErr_toSysErr(releaseRes);
            mapping_set_error(mapping, retVal);
         }

         goto cleanup_buf;
      }
   }


cleanup_buf:
   NoAllocBufferStore_addBuf(pagedIOBufStore, pagedIOBuf);

   return retVal;
}


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,28)

/**
 * Note: We never return an error code here, currently.
 *
 * @param from page inner offset start
 * @param to page inner offset end
 * @return 0 on success
 */
int FhgfsOps_prepare_write(struct file* file, struct page* page, unsigned from, unsigned to)
{
   int retVal = 0;

   loff_t fileSize;
   loff_t offset;

#ifdef LOG_DEBUG_MESSAGES
   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_prepare_write";

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBuf;
   char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, file->f_dentry, &pathStoreBuf);

   LOG_DEBUG_FORMATTED(log, 5, logContext, "called. Path: %s; from: %u; to: %u", path, from, to);
   IGNORE_UNUSED_VARIABLE(logContext);
   IGNORE_UNUSED_VARIABLE(log);

   // clean-up
   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);
#endif // LOG_DEBUG_MESSAGES


   if(PageUptodate(page) )
      goto clean_up;

   if( (to == PAGE_CACHE_SIZE) && (from == 0) )
   { // we're writing a full page
      // full page => will be up to date => no need to read-update the page from the server
      SetPageUptodate(page);
      goto clean_up;
   }

   // we're not writing a full page => try to read-update the page

   offset = page_offset(page);
   fileSize = i_size_read(page->mapping->host);

   LOG_DEBUG_FORMATTED(log, 5, logContext, "fileSize: %lld; offset: %lld",
      (long long)fileSize, (long long)offset);

   if( (offset >= fileSize) ||
       ( (from == 0) && (offset + to) >= fileSize) )
   {
      // we don't need to read data beyond the end of the file
      //    => zero it, and set the page uptodate
      simple_prepare_write(file, page, from, to);
      SetPageUptodate(page);
   }
   else
   if( (file->f_flags & O_ACCMODE) != O_WRONLY)
   {
      // file open for reading also => read-update the page
      retVal = __FhgfsOps_readpage_sync(file, page, offset);
   }
   else
   { // Write-only file handle => we cannot use it to read-update this page.
      // We could try to get a different open file handle, but the user data will be written out by
      // commit_write, so it's fine.
   }


   // clean-up
clean_up:

   return 0;
}

/**
 * @param from page inner offset start
 * @param to page inner offset end
 *
 * @return 0 in success
 */
int FhgfsOps_commit_write(struct file* file, struct page* page, unsigned from, unsigned to)
{
   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_commit_write";

   int retVal = 0;
   struct inode* inode = page->mapping->host;
   loff_t pageStart = page_offset(page);
   loff_t endPos = pageStart + to;

#ifdef LOG_DEBUG_MESSAGES

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBuf;
   char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, file->f_dentry, &pathStoreBuf);

   LOG_DEBUG_FORMATTED(log, 5, logContext, "called. Path: %s", path);
   LOG_DEBUG_FORMATTED(log, 5, logContext, "pageStart: %lld; from: %u; to: %u",
      (long long)pageStart, from, to);

   // clean up
   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);

#endif // LOG_DEBUG_MESSAGES


   // check fileSize and update it to page endPos if required
   spin_lock(&inode->i_lock);
   if(endPos > inode->i_size)
   {
      i_size_write(inode, endPos);
   }
   spin_unlock(&inode->i_lock);

   // write the page directly or mark it for async writing
   if(!PageUptodate(page) )
   { // write now
      if(unlikely(to < from) )
      {
         Logger_logFormatted(log, 2, logContext, "Illegal offsets. from: %d; to: %d", from, to);
         goto clean_up;
      }

      retVal = __FhgfsOps_writepage_sync(file, page, from, to-from);
      if(likely(retVal > 0) )
         retVal = 0;
   }
   else
   { // mark for async writing via writepage
      set_page_dirty(page);
   }


clean_up:
   // clean-up

   LOG_DEBUG_FORMATTED(log, 5, logContext, "complete. retVal: %d", retVal);

   return retVal;
}

#else // LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,28)

/**
 * @param fsdata can be used to hand any data over to write_end() (but note that the old
 * prepare_write() doesn't have this)
 * @return 0 on success
 */
int FhgfsOps_write_begin(struct file* file, struct address_space* mapping,
   loff_t pos, unsigned len, unsigned flags, struct page** pagep, void** fsdata)
{
   pgoff_t index = pos >> PAGE_CACHE_SHIFT;
   loff_t offset = pos & (PAGE_CACHE_SIZE - 1);
   loff_t page_start = pos & PAGE_MASK;

   loff_t i_size;
   struct page* page;

   int retVal = 0;

#ifdef LOG_DEBUG_MESSAGES

   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_write_begin";

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBuf;
   char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, file->f_dentry, &pathStoreBuf);

   LOG_DEBUG_FORMATTED(log, 5, logContext, "called. Path: %s; (offset: %lld; page_start: %lld; "
      "len: %u)", path, (long long)offset, (long long)page_start, len);

   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);

#endif // LOG_DEBUG_MESSAGES

   page = grab_cache_page_write_begin(mapping, index, flags);
   if(!page)
   {
      retVal = -ENOMEM;
      goto clean_up;
   }

   if(PageUptodate(page) )
      goto clean_up;

   if(len == PAGE_CACHE_SIZE)
   {
      // two possibilities:
      // a) full page write => no need to read-update the page from the server
      // b) short write (with offset) => will lead to sync write
      goto clean_up;
   }

   i_size = i_size_read(mapping->host);
   if( (page_start >= i_size) ||
       (!offset && ( (pos + len) >= i_size) ) )
   {
      // we don't need to read data beyond the end of the file
      //    => zero it, and set the page up-to-date

      zero_user_segments(page, 0, offset, offset + len, PAGE_CACHE_SIZE);

      // note: PageChecked means rest of the page (to which is not being written) is up-to-date.
      //       so when our data is written, the whole page is up-to-date.
      SetPageChecked(page);

      goto clean_up;
   }

   if( (file->f_flags & O_ACCMODE) != O_WRONLY)
   {
      // file handle allows reading => read-update the page
      // (in case of an error, PG_uptodate is not set and hence write_end() will do a sync write)
      __FhgfsOps_readpage_sync(file, page, page_start);
   }
   else
   { // write-only file handle => we cannot use it to read-update this page.
      // We could try to get a different open file handle, but the user data will be written out by
      // commit_write, so it's fine.
   }


clean_up:
   // clean-up

   *pagep = page;

   return retVal;
}

/**
 * @param copied the amount that was able to be copied ("copied==len" is always true if
 * write_begin() was called with the AOP_FLAG_UNINTERRUPTIBLE flag)
 * @param fsdata whatever write_begin() set here
 * @return < 0 on failure, number of bytes copied into pagecache (<= 'copied') otherwise
 */
int FhgfsOps_write_end(struct file* file, struct address_space* mapping,
   loff_t pos, unsigned len, unsigned copied, struct page* page, void* fsdata)
{
   FsFileInfo* fileInfo = __FhgfsOps_getFileInfo(file);

   struct inode* inode = mapping->host;
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);

   int retVal;

#ifdef LOG_DEBUG_MESSAGES

   FsObjectInfo* fsObjectInfo = __FhgfsOps_getObjectInfo(file);
   App* app = FsObjectInfo_getApp(fsObjectInfo);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_write_end";

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBuf;
   char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, file->f_dentry, &pathStoreBuf);

   LOG_DEBUG_FORMATTED(log, 5, logContext, "called. Path: %s", path);
   LOG_DEBUG_FORMATTED(log, 5, logContext, "pos: %lld; len: %u; copied: %u",
      (long long)pos, len, copied);

   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);

#endif // LOG_DEBUG_MESSAGES


   if(PageChecked(page) )
   {
      // note: see write_begin() for meaning of PageChecked()

      if(copied == len)
         SetPageUptodate(page);

      ClearPageChecked(page);
   }
   else
   if(!PageUptodate(page) && (copied == PAGE_CACHE_SIZE) )
      SetPageUptodate(page);


   if(!PageUptodate(page) )
   {
      unsigned offset = pos & (PAGE_CACHE_SIZE - 1);
      char* buf = kmap(page);
      RemotingIOInfo ioInfo;
      ssize_t writeRes;

      FsFileInfo_getIOInfo(fileInfo, fhgfsInode, &ioInfo);

      writeRes = FhgfsOpsRemoting_writefile(&buf[offset], copied, pos, &ioInfo);
      if(likely(writeRes > 0) )
      {
         retVal = writeRes;
         pos += writeRes;
      }
      else
         retVal = FhgfsOpsErr_toSysErr(-writeRes);

      kunmap(page);
   }
   else
   {
      retVal = copied;
      pos += copied;
      set_page_dirty(page);
   }

   if(likely(retVal > 0) )
   {
      spin_lock(&inode->i_lock);
      if(pos > inode->i_size)
         i_size_write(inode, pos);
      spin_unlock(&inode->i_lock);
   }

   unlock_page(page);
   page_cache_release(page);

   // clean-up

   LOG_DEBUG_FORMATTED(log, 5, logContext, "complete. retVal: %d", retVal);

   return retVal;
}

#endif // LINUX_VERSION_CODE


/**
 * Called by the VFS to map a logical block offset within object to physical block number.
 *
 * @param mapping address space info
 * @param block the block to be mapped
 * @return disk address for the block or 0 on hole or error
 */
sector_t FhgfsOps_bmap(struct address_space* mapping, sector_t block)
{
   //printk_fhgfs(FHGFS_KERN_INFO, "FhgfsOps_bmap called. block: %lld\n", (long long)block);
      // debug in

   // note: +1 to avoid returning 0 for the very first block
   return block+1;
}

/**
 * Note: This method must be present because otherwise the kernel rejects open with O_DIRECT in
 * fs/open.c. It is only called by the generic file read/write routines, so it should actually
 * never be called for buffered IO.
 * Note: This is currently a very slow path, because we submit and retrieve every single page
 * seperately.
 *
 * @param rw whether this is a read or a write {READ, WRITE}
 * @param iocb I/O control block with open file handle
 * @param iov I/O buffer vectors (array)
 * @param pos file offset
 * @param nr_segs length of iov array
 */
ssize_t FhgfsOps_directIO(int rw, struct kiocb *iocb, const struct iovec *iov, loff_t pos,
   unsigned long nr_segs)
{
   struct file* file = iocb->ki_filp;
   FsFileInfo* fileInfo = __FhgfsOps_getFileInfo(file);
   struct dentry* dentry = file->f_dentry;
   struct inode* inode = dentry->d_inode;
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);
   RemotingIOInfo ioInfo;

   ssize_t remotingRes;
   ssize_t retVal = 0;

   unsigned long i;

#ifdef LOG_DEBUG_MESSAGES
   const char* logContext = "FhgfsOps_directIO";
   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBuf;
   char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, file->f_dentry, &pathStoreBuf);

   LOG_DEBUG_FORMATTED(log, 5, logContext, "called. Path: %s (%s, pos: %lld, nr_seqs: %lld)",
      path, (rw == WRITE) ? "WRITE" : "READ", (long long)pos, (long long)nr_segs);

   // clean-up
   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);
#endif // LOG_DEBUG_MESSAGES


   FsFileInfo_getIOInfo(fileInfo, fhgfsInode, &ioInfo);

   for(i=0; i < nr_segs; i++)
   {
      const struct iovec* currentIOV = &iov[i];

      remotingRes = (rw == WRITE) ?
         FhgfsOpsRemoting_writefile(currentIOV->iov_base, currentIOV->iov_len, pos, &ioInfo) :
         FhgfsOpsRemoting_readfile(currentIOV->iov_base, currentIOV->iov_len, pos, &ioInfo);

      if(unlikely(remotingRes < 0) )
      { // error occured
         LOG_DEBUG_FORMATTED(log, 1, logContext, "error: %s",
            FhgfsOpsErr_toErrString(-remotingRes) );

         retVal = FhgfsOpsErr_toSysErr(-remotingRes);
         break;
      }

      retVal += remotingRes;
      pos += remotingRes;

      if(remotingRes != (ssize_t)currentIOV->iov_len)
         break; // read end of file or no write space left
   }

   return retVal;
}



/**
 * Note: Reads a locked page and unlocks it when done
 *
 * @param offset file offset for the page
 * @return 0 on success, negative linux error code otherwise
 */
int __FhgfsOps_readpage_sync(struct file* file, struct page* page, loff_t offset)
{
   int retVal = 0;
   char* buf;
   struct inode* inode = file->f_mapping->host;
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);
   RemotingIOInfo ioInfo;
   ssize_t readRes;

   FsFileInfo* fileInfo = __FhgfsOps_getFileInfo(file);

#ifdef LOG_DEBUG_MESSAGES

   FsObjectInfo* fsObjectInfo = __FhgfsOps_getObjectInfo(file);
   App* app = FsObjectInfo_getApp(fsObjectInfo);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_readpage_sync";

   LOG_DEBUG_FORMATTED(log, 5, logContext, "offset: %lld", (long long)offset);

#endif // LOG_DEBUG_MESSAGES

   page_cache_get(page);
   buf = kmap(page);

   FsFileInfo_getIOInfo(fileInfo, fhgfsInode, &ioInfo);

   readRes = FhgfsOpsRemoting_readfile(buf, PAGE_CACHE_SIZE, offset, &ioInfo);
   if(unlikely(readRes < 0) )
   { // error occured
      retVal = FhgfsOpsErr_toSysErr(-readRes);
      goto clean_up;
   }

   if(readRes < PAGE_CACHE_SIZE)
   { // did not receive the whole page => zero the remaining bytes
      memset(&buf[readRes], 0, PAGE_CACHE_SIZE - readRes);
   }

   flush_dcache_page(page);
   SetPageUptodate(page);

   retVal = 0;

clean_up:
   kunmap(page);
   page_cache_release(page);

   return retVal;
}

/**
 * @param pageoffset is the inner-page offset (relative to the page start)
 */
int __FhgfsOps_writepage_sync(struct file* file, struct page* page, unsigned pageOffset,
   unsigned count)
{
   int retVal = 0;

   struct inode* inode = file->f_mapping->host;
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);
   RemotingIOInfo ioInfo;
   char* buf = kmap(page);
   loff_t offset = page_offset(page) + pageOffset;
   ssize_t writeRes;

   FsFileInfo* fileInfo = __FhgfsOps_getFileInfo(file);

#ifdef LOG_DEBUG_MESSAGES

   FsObjectInfo* fsObjectInfo = __FhgfsOps_getObjectInfo(file);
   App* app = FsObjectInfo_getApp(fsObjectInfo);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_writepage_sync";

   LOG_DEBUG_FORMATTED(log, 5, logContext, "offset: %lld", (long long)offset);

#endif // LOG_DEBUG_MESSAGES

   FsFileInfo_getIOInfo(fileInfo, fhgfsInode, &ioInfo);

   writeRes = FhgfsOpsRemoting_writefile(&buf[pageOffset], count, offset, &ioInfo);
   if(unlikely(writeRes < 0) )
   { // error occured
      retVal = FhgfsOpsErr_toSysErr(-writeRes);
   }

   // clean-up
   kunmap(page);

   return retVal;
}

/**
 * Note: Will check i_size and adjust "to" accordingly
 *
 * @return 0 on success
 */
int __FhgfsOps_writepage_async(struct page* page, unsigned from, unsigned to)
{
   App* app;

   int retVal = 0;
   struct address_space* mapping = page->mapping;
   struct inode* inode;
   loff_t i_size;
   loff_t offset = page_offset(page);
   char* buf = NULL;

   if(!mapping || !mapping->host)
   { // page disappeared (truncated or invalidated)
      retVal = -EFAULT;
      goto err_exit;
   }

   inode = mapping->host; // (note: depends on mapping-check above)
   i_size = inode->i_size;
   app = FhgfsOps_getApp(inode->i_sb);

   #ifdef LOG_DEBUG_MESSAGES
   {
      const char* logContext = "FhgfsOps_writepage_async";
      Logger* log = App_getLogger(app);

      struct dentry* dentry = d_find_alias(inode); // calls dget_locked

      if(!dentry)
      {
         LOG_DEBUG_FORMATTED(log, 5, logContext, "called. Path: %s; offset: %lld, from: %u; to: %u",
            "n/a (no dentry found)", (long long)offset, from, to);
      }
      else
      {
         NoAllocBufferStore* bufStore = App_getPathBufStore(app);
         char* pathStoreBuf;
         char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);

         LOG_DEBUG_FORMATTED(log, 5, logContext, "called. Path: %s; offset: %lld, from: %u; to: %u",
            path, (long long)offset, from, to);

         NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);
         dput(dentry);
      }
   }
   #endif // LOG_DEBUG_MESSAGES


   offset += (loff_t)from;
   buf = kmap(page);

   if( (to > PAGE_CACHE_SIZE) || (from > to) )
   {
      retVal = -EIO;
      goto err_unmap;
   }

   if(offset > i_size)
   { // racing with truncate
      retVal = 0; // ignore
      goto err_unmap;
   }

   // check to make sure that we are not extending the file
   if( (i_size - offset) < (loff_t)to)
      to = (unsigned)(i_size - offset);

   retVal = __FhgfsOps_writeStateless(inode, &buf[from], to-from, offset);

   // clean-up

err_unmap:
   kunmap(page);

err_exit:
   return retVal;
}

ssize_t __FhgfsOps_writeStateless(struct inode* inode, const char __user *buf, size_t size,
   loff_t offset)
{
   int retVal = -FHGFS_EREMOTEIO;

   //App* app = FhgfsOps_getApp(inode->i_sb);
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);

   FileHandleType handleType;
   RemotingIOInfo ioInfo;
   int maxUsedNodeIndex = -1;

   FhgfsOpsErr referenceRes;
   ssize_t writeRes;
   FhgfsOpsErr releaseRes;


   // open file

   referenceRes = FhgfsInode_referenceHandle(fhgfsInode, OPENFILE_ACCESS_WRITE, fhgfs_true,
      &handleType);
   if(referenceRes != FhgfsOpsErr_SUCCESS)
   { // error
      return FhgfsOpsErr_toSysErr(referenceRes);
   }

   // write file

   FhgfsInode_getRefIOInfo(fhgfsInode, handleType, FhgfsInode_handleTypeToOpenFlags(handleType),
      &maxUsedNodeIndex, &ioInfo);

   writeRes = FhgfsOpsRemoting_writefile(buf, size, offset, &ioInfo);
   if(writeRes < 0)
   { // error
      FhgfsInode_releaseHandle(fhgfsInode, handleType, maxUsedNodeIndex);

      retVal = FhgfsOpsErr_toSysErr(-writeRes);
      goto err_exit;
   }

   retVal = writeRes;

   // close the file

   releaseRes = FhgfsInode_releaseHandle(fhgfsInode, handleType, maxUsedNodeIndex);
   if(releaseRes != FhgfsOpsErr_SUCCESS)
   { // error
      retVal = FhgfsOpsErr_toSysErr(releaseRes);
   }

err_exit:

   return retVal;
}
