#include <include_closed/app/log/Logger.h>
#include <include_closed/app/App.h>
#include <include_closed/app/config/Config.h>
#include <include_closed/common/toolkit/vector/StrCpyVec.h>
#include <include_open/common/storage/Open_StorageErrors.h>
#include <include_open/common/toolkit/Open_StringTk.h>
#include <include_closed/filesystem/FhgfsOpsHelper.h>
#include <include_closed/common/toolkit/MetadataTk.h>
#include <open/os/OsTypeConversion.h>
#include "FhgfsOpsSuper.h"
#include "FhgfsOpsDir.h"
#include "FhgfsOpsInode.h"
#include "FhgfsOpsFile.h"

#include <linux/namei.h>
#include <linux/backing-dev.h>


static struct kmem_cache* FhgfsInodeCache = NULL;


struct inode_operations fhgfs_file_inode_ops =
{
   .getattr    = FhgfsOps_getattr,
   .setattr    = FhgfsOps_setattr,
   .truncate   = FhgfsOps_truncate,
};

struct inode_operations fhgfs_file_pagecache_inode_ops =
{
   .getattr    = FhgfsOps_getattr,
   .setattr    = FhgfsOps_setattr,
};

struct inode_operations fhgfs_symlink_inode_ops =
{
   .readlink      = generic_readlink, //FhgfsOps_readlink, // default should be fine
   .follow_link   = FhgfsOps_follow_link,
   .put_link      = FhgfsOps_put_link,
};

static struct inode_operations fhgfs_dir_inode_ops =
{
   .lookup     = FhgfsOps_lookup,
   .create     = FhgfsOps_create,
   .link       = FhgfsOps_link,
   .unlink     = FhgfsOps_unlink,
   .symlink    = FhgfsOps_symlink,
   .mkdir      = FhgfsOps_mkdir,
   .rmdir      = FhgfsOps_rmdir,
   .rename     = FhgfsOps_rename,
   .getattr    = FhgfsOps_getattr,
   .setattr    = FhgfsOps_setattr,
};

//static struct backing_dev_info fhgfs_backing_dev_info =
//{
//   .ra_pages   = 0, //VM_MAX_READAHEAD * 1024 / PAGE_CACHE_SIZE, /* read-ahead pages */
//   .state      = 0,
//   .capabilities  = BDI_CAP_MAP_COPY,
//   .unplug_io_fn  = default_unplug_io_fn,
//};


/**
 * Find out whether a directory entry exists. Add the dentry and create/associate the
 * corresponding inode in case it does exist.
 *
 * @param dentry the thing that we are looking for
 * @return ERR_PTR(x), with x==0 if successful, negative linux error code otherwise
 */
struct dentry* FhgfsOps_lookup(struct inode* parentDir, struct dentry* dentry,
   struct nameidata* nameidata)
{
   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_lookup";

   int retVal = 0;
   FhgfsOpsErr statRes;
   fhgfs_stat fhgfsStat;
   FhgfsInode* fhgfsInode = FHGFS_INODE(parentDir);
   OwnerMinInfo parentInfo;
   OwnerMinInfo newEntryInfo;
   const char* entryName = dentry->d_name.name;
   fhgfs_bool freeNewEntryInfo = fhgfs_false;

   if(unlikely(Logger_getLogLevel(log) >= 5) )
   {
      NoAllocBufferStore* bufStore = App_getPathBufStore(app);
      char* pathStoreBuf;
      char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);

      Logger_logFormatted(log, 5, logContext, "called. Path: %s", path);

      NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);
   }

   if(dentry->d_name.len > NAME_MAX)
   {
      retVal = -ENAMETOOLONG;
      goto clean_up;
   }

   if(IS_ROOT(dentry) )
   { // madame root deserves special treatment, of course...
      MetadataTk_getRootMinInfoCopy(app, &newEntryInfo);
      freeNewEntryInfo = fhgfs_true;

      statRes = FhgfsOpsRemoting_statDirect(app, &newEntryInfo, &fhgfsStat);
   }
   else
   { // just a normal subentry of some dir => lookup entryInfo and stat it
      FhgfsInode_getOwnerMinInfoCopy(fhgfsInode, &parentInfo);

      statRes = FhgfsOpsRemoting_lookupSubentry(app, &parentInfo, entryName, &newEntryInfo);

      if(statRes == FhgfsOpsErr_SUCCESS)
      {
         freeNewEntryInfo = fhgfs_true;
         statRes = FhgfsOpsRemoting_statDirect(app, &newEntryInfo, &fhgfsStat);
      }
   }


   if(statRes != FhgfsOpsErr_SUCCESS)
   { // error
      switch(statRes)
      {
         case FhgfsOpsErr_PATHNOTEXISTS:
         {
            dentry->d_op = &fhgfs_dentry_ops;
            d_add(dentry, NULL);
         } break;

         default:
         {
            retVal = FhgfsOpsErr_toSysErr(statRes);
         } break;
      }
   }
   else
   { // success (entry exists)
      struct kstat kstat;
      struct inode* newInode;

      OsTypeConv_kstatFhgfsToOs(&fhgfsStat, &kstat);

      kstat.ino = iunique(dentry->d_sb, 2);

      newInode = __FhgfsOps_newInode(parentDir->i_sb, &kstat, 0, &newEntryInfo);
      if(!newInode)
         retVal = -EACCES;
      else
      {
         dentry->d_op = &fhgfs_dentry_ops;
         d_add(dentry, newInode);
      }

      freeNewEntryInfo = fhgfs_false; // newEntryInfo owned or freed by _newInode()
   }

clean_up:
   // clean-up
   if(!IS_ROOT(dentry) )
      FhgfsInode_freeOwnerMinInfoVals(&parentInfo);

   if(freeNewEntryInfo)
      FhgfsInode_freeOwnerMinInfoVals(&newEntryInfo);


   return ERR_PTR(retVal);
}

int FhgfsOps_getattr(struct vfsmount* mnt, struct dentry* dentry, struct kstat* kstat)
{
   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_getattr";

   int retVal = 0;
   struct inode* inode = dentry->d_inode;

   if(unlikely(Logger_getLogLevel(log) >= 5) )
   {
      NoAllocBufferStore* bufStore = App_getPathBufStore(app);
      char* pathStoreBuf;
      char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);

      Logger_logFormatted(log, 5, logContext, "called. Path: %s", path);

      NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);
   }

   retVal = __FhgfsOps_refreshInode(app, inode);
   if(!retVal)
   {
      generic_fillattr(dentry->d_inode, kstat);

      kstat->blksize = FHGFS_INODE_BLOCKSIZE;

      if(IS_ROOT(dentry) )
         kstat->ino = FHGFS_INODE_ROOT_INO; // root => assign the constant root inode number
   }

   // clean-up

   return retVal;
}


int FhgfsOps_setattr(struct dentry* dentry, struct iattr* iattr)
{
   // note: called for chmod(), chown() and utime()
   // note: changed fields can be determined by the iattr->ia_valid flags (ATTR_...)

   App* app = FhgfsOps_getApp(dentry->d_sb);
   Config* cfg = App_getConfig(app);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_setattr";

   int retVal = 0;
   SettableFileAttribs fhgfsAttr;
   int validFhgfsAttribs;
   FhgfsOpsErr setAttrRes = FhgfsOpsErr_SUCCESS;
   int refreshRes;
   struct inode* inode = dentry->d_inode;
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);
   OwnerMinInfo entryInfo;

   if(unlikely(Logger_getLogLevel(log) >= 5) )
   {
      NoAllocBufferStore* bufStore = App_getPathBufStore(app);
      char* pathStoreBuf;
      char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);

      Logger_logFormatted(log, 5, logContext, "called. Path: %s", path);

      NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);
   }

   if( (retVal = inode_change_ok(inode, iattr) ) < 0)
      return retVal;

   if(iattr->ia_valid & ATTR_SIZE)
   { // make sure we only update size of regular files
      if(!S_ISREG(inode->i_mode) )
         iattr->ia_valid &= ~ATTR_SIZE;
   }

   if(S_ISREG(inode->i_mode) )
   { // flush all dirty file data
      filemap_write_and_wait(inode->i_mapping);
   }

   FhgfsInode_getOwnerMinInfoCopy(fhgfsInode, &entryInfo);

   OsTypeConv_iattrOsToFhgfs(iattr, &fhgfsAttr, &validFhgfsAttribs);

   if(validFhgfsAttribs)
      setAttrRes = FhgfsOpsRemoting_setAttr(app, &entryInfo, &fhgfsAttr, validFhgfsAttribs);

   // note: Remoting_setAttr() on the metadata server doesn't handle truncation (we have a separate
   //    direct-to-servers call to make it faster)

   if(setAttrRes == FhgfsOpsErr_SUCCESS)
   { // all right so far => handle truncation
      if(iattr->ia_valid & ATTR_SIZE)
      {
         setAttrRes = FhgfsOpsRemoting_truncfile(app, &entryInfo, iattr->ia_size);

         if( (setAttrRes == FhgfsOpsErr_SUCCESS) &&
             (Config_getTuneFileCacheTypeNum(cfg) == FILECACHETYPE_Paged) )
         {
            FhgfsOps_vmtruncate(inode, iattr->ia_size);
         }
      }
   }

   if(setAttrRes != FhgfsOpsErr_SUCCESS)
      retVal = FhgfsOpsErr_toSysErr(setAttrRes);

   // update local inode
   refreshRes = __FhgfsOps_refreshInode(app, inode);
   if(!retVal)
      retVal = refreshRes;

   // clean-up
   FhgfsInode_freeOwnerMinInfoVals(&entryInfo);

   return retVal;
}

int FhgfsOps_mkdir(struct inode* dir, struct dentry* dentry, int mode)
{
   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_mkdir";

   int retVal = 0;
   FhgfsOpsErr mkRes;
   FhgfsInode* fhgfsInode = FHGFS_INODE(dir);
   OwnerMinInfo parentInfo;
   OwnerMinInfo newEntryInfo;
   const char* entryName = dentry->d_name.name;

   if(unlikely(Logger_getLogLevel(log) >= 5) )
   {
      NoAllocBufferStore* bufStore = App_getPathBufStore(app);
      char* pathStoreBuf;
      char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);

      Logger_logFormatted(log, 5, logContext, "called. Path: %s", path);

      NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);
   }

   FhgfsInode_getOwnerMinInfoCopy(fhgfsInode, &parentInfo);

   mode |= S_IFDIR; // just make sure this is a dir

   mkRes = FhgfsOpsRemoting_mkdir(app, &parentInfo, entryName, currentFsUserID, currentFsGroupID,
      mode, &newEntryInfo);

   if(mkRes != FhgfsOpsErr_SUCCESS)
   {
      retVal = FhgfsOpsErr_toSysErr(mkRes);
   }
   else
   { // remote success => create the local inode
      retVal = __FhgfsOps_instantiateInode(dentry, &newEntryInfo);
   }

   // clean-up
   FhgfsInode_freeOwnerMinInfoVals(&parentInfo);

   return retVal;
}

int FhgfsOps_rmdir(struct inode* dir, struct dentry* dentry)
{
   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_rmdir";

   int retVal = 0;
   FhgfsOpsErr rmRes;
   FhgfsInode* fhgfsInode = FHGFS_INODE(dir);
   OwnerMinInfo parentInfo;
   const char* entryName = dentry->d_name.name;

   if(unlikely(Logger_getLogLevel(log) >= 5) )
   {
      NoAllocBufferStore* bufStore = App_getPathBufStore(app);
      char* pathStoreBuf;
      char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);

      Logger_logFormatted(log, 5, logContext, "called. Path: %s", path);

      NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);
   }

   FhgfsInode_getOwnerMinInfoCopy(fhgfsInode, &parentInfo);

   rmRes = FhgfsOpsRemoting_rmdir(app, &parentInfo, entryName);

   if(rmRes != FhgfsOpsErr_SUCCESS)
   {
      retVal = FhgfsOpsErr_toSysErr(rmRes);
   }
   else
   { // remote success
      clear_nlink(dentry->d_inode);
   }


   // clean-up
   FhgfsInode_freeOwnerMinInfoVals(&parentInfo);

   return retVal;
}

int FhgfsOps_create(struct inode* dir, struct dentry* dentry, int mode, struct nameidata* nameidata)
{
   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_create";

   int retVal = 0;
   FhgfsOpsErr mkRes;
   FhgfsInode* fhgfsInode = FHGFS_INODE(dir);
   OwnerMinInfo parentInfo;
   OwnerMinInfo newEntryInfo;
   const char* entryName = dentry->d_name.name;

   if(Logger_getLogLevel(log) >= 4)
   {
      NoAllocBufferStore* bufStore = App_getPathBufStore(app);
      char* pathStoreBuf;
      char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);

      Logger_logFormatted(log, 4, logContext, "called. Path: %s", path);

      NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);
   }

   FhgfsInode_getOwnerMinInfoCopy(fhgfsInode, &parentInfo);

   mode |=  S_IFREG; // just make sure we create a file here

   mkRes = FhgfsOpsRemoting_mkfile(app, &parentInfo, entryName,
      currentFsUserID, currentFsGroupID, mode, &newEntryInfo);

   if(mkRes != FhgfsOpsErr_SUCCESS)
   {
      if( (mkRes != FhgfsOpsErr_EXISTS) || (nameidata->intent.open.flags & O_EXCL) )
      { // Err_EXISTS is only an error if O_EXCL was specified

         d_drop(dentry); // avoid leaving a negative dentry behind

         retVal = FhgfsOpsErr_toSysErr(mkRes);
      }
   }

   if(!retVal) // keep this separate because of the O_EXCL special case above
   { // remote success => create local inode
      retVal = __FhgfsOps_instantiateInode(dentry, &newEntryInfo);
   }

   // clean-up
   FhgfsInode_freeOwnerMinInfoVals(&parentInfo);

   return retVal;
}


int FhgfsOps_unlink(struct inode* dir, struct dentry* dentry)
{
   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_unlink";

   int retVal = 0;
   FhgfsOpsErr unlinkRes;
   FhgfsInode* fhgfsInode = FHGFS_INODE(dir);
   OwnerMinInfo parentInfo;
   const char* entryName = dentry->d_name.name;

   if(Logger_getLogLevel(log) >= 4)
   {
      NoAllocBufferStore* bufStore = App_getPathBufStore(app);
      char* pathStoreBuf;
      char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);

      Logger_logFormatted(log, 4, logContext, "called. Path: %s", path);

      NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);
   }

   FhgfsInode_getOwnerMinInfoCopy(fhgfsInode, &parentInfo);

   unlinkRes = FhgfsOpsRemoting_unlinkfile(app, &parentInfo, entryName);

   if(unlinkRes != FhgfsOpsErr_SUCCESS)
   {
      retVal = FhgfsOpsErr_toSysErr(unlinkRes);

      if(retVal == -ENOENT)
         d_drop(dentry);
   }
   else
   { // remote success
      if(dentry->d_inode)
         drop_nlink(dentry->d_inode);
   }

   // clean-up
   FhgfsInode_freeOwnerMinInfoVals(&parentInfo);

   return retVal;
}


int FhgfsOps_symlink(struct inode* dir, struct dentry* dentry, const char* to)
{
   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_symlink";

   int retVal;
   int mode = S_IFLNK | S_IRWXU | S_IRWXG | S_IRWXO;
   FhgfsInode* fhgfsInode = FHGFS_INODE(dir);
   OwnerMinInfo parentInfo;
   OwnerMinInfo newEntryInfo;
   const char* entryName = dentry->d_name.name;

   if(unlikely(Logger_getLogLevel(log) >= 5) )
   {
      NoAllocBufferStore* bufStore = App_getPathBufStore(app);
      char* pathStoreBuf;
      char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);

      Logger_logFormatted(log, 5, logContext, "called. Path: %s; To: %s", path, to);

      NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);
   }

   FhgfsInode_getOwnerMinInfoCopy(fhgfsInode, &parentInfo);

   retVal = FhgfsOpsHelper_symlink(app, &parentInfo, entryName, to,
      currentFsUserID, currentFsGroupID, mode, &newEntryInfo);

   if(retVal)
   { // error occured
      d_drop(dentry); // avoid leaving a negative dentry behind
   }
   else
   { // remote success => create local inode
      retVal = __FhgfsOps_instantiateInode(dentry, &newEntryInfo);
   }

   // clean-up
   FhgfsInode_freeOwnerMinInfoVals(&parentInfo);

   return retVal;
}

int FhgfsOps_link(struct dentry* dentryFrom, struct inode* inode, struct dentry* dentryTo)
{
   App* app = FhgfsOps_getApp(dentryFrom->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_link";

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBuf;
   char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentryFrom, &pathStoreBuf);

   Logger_logFormatted(log, 2, logContext, "called - NOT IMPLEMENTED YET. From: %s", path);


   // clean-up
   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);

   return -EPERM; // do not allow creation of hardlink
}

void* FhgfsOps_follow_link(struct dentry* dentry, struct nameidata* nd)
{
   void* retVal = NULL;

   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_follow_link";

   int readRes;
   char* bufPage = (char*)__os_get_free_page();
   char* destination = bufPage;

   struct inode* inode = dentry->d_inode;
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);
   OwnerMinInfo entryInfo;

   if(unlikely(Logger_getLogLevel(log) >= 5) )
   {
      NoAllocBufferStore* bufStore = App_getPathBufStore(app);
      char* pathStoreBuf;
      char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);

      Logger_logFormatted(log, 5, logContext, "called. Path: %s", path);

      NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);
   }

   FhgfsInode_getOwnerMinInfoCopy(fhgfsInode, &entryInfo);

   readRes = FhgfsOpsHelper_readlink(app, &entryInfo, bufPage, PAGE_SIZE-1);

   if(readRes < 0)
   {
      destination = ERR_PTR(readRes);
   }
   else
      bufPage[readRes] = 0;

   if(readRes == (PAGE_SIZE-1) )
   { // link destination too long
      destination = ERR_PTR(-ENAMETOOLONG);
   }

   // store link destination
   nd_set_link(nd, destination);

   // clean-up
   FhgfsInode_freeOwnerMinInfoVals(&entryInfo);

   if(IS_ERR(destination) )
      os_free_page( (unsigned long)bufPage);

   // Note: free_page() is called by the put_link method in the success case

   return retVal;
}


void FhgfsOps_put_link(struct dentry *dentry, struct nameidata* nd, void* p)
{
   char* destination = nd_get_link(nd);

   if(!IS_ERR(destination) )
      os_free_page( (unsigned long)destination);
}


int FhgfsOps_rename(struct inode* inodeDirFrom, struct dentry* dentryFrom,
   struct inode* inodeDirTo, struct dentry* dentryTo)
{
   App* app = FhgfsOps_getApp(dentryFrom->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_rename";

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBufFrom;
   char* pathStoreBufTo;

   char* pathFrom = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentryFrom, &pathStoreBufFrom);
   char* pathTo = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentryTo, &pathStoreBufTo);

   int retVal = 0;
   FhgfsOpsErr renameRes;

   Logger_logFormatted(log, 5, logContext, "called. From: %s; To: %s",
      pathFrom, pathTo);

   renameRes = FhgfsOpsRemoting_rename(app, pathFrom, pathTo);

   if(renameRes != FhgfsOpsErr_SUCCESS)
   {
      retVal = FhgfsOpsErr_toSysErr(renameRes);
   }
   else
   { // remote success
      // nothing to be done here right now
   }

   LOG_DEBUG_FORMATTED(log, 4, logContext, "remoting complete. result: %d", (int)renameRes);

   // clean-up
   NoAllocBufferStore_addBuf(bufStore, pathStoreBufFrom);
   NoAllocBufferStore_addBuf(bufStore, pathStoreBufTo);

   return retVal;
}


/**
 * Note: i_size of the inode is set to the desired size before this is called
 */
void FhgfsOps_truncate(struct inode* inode)
{
   App* app = FhgfsOps_getApp(inode->i_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_truncate";

   int retVal = 0;
   FhgfsOpsErr truncRes;
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);
   OwnerMinInfo entryInfo;

   loff_t size = i_size_read(inode);

   if(unlikely(Logger_getLogLevel(log) >= 5) )
   {
      struct dentry* dentry = d_find_alias(inode); // calls dget_locked

      NoAllocBufferStore* bufStore = App_getPathBufStore(app);
      char* pathStoreBuf;
      char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);

      Logger_logFormatted(log, 5, logContext, "called. Path: %s", path);

      NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);
      dput(dentry);
   }

   FhgfsInode_getOwnerMinInfoCopy(fhgfsInode, &entryInfo);

   truncRes = FhgfsOpsRemoting_truncfile(app, &entryInfo, size);

   if(truncRes != FhgfsOpsErr_SUCCESS)
   {
      retVal = FhgfsOpsErr_toSysErr(truncRes);
   }

   // clean-up
   FhgfsInode_freeOwnerMinInfoVals(&entryInfo);

   return; // note: this is an internal function which gives us no way to return a result.
}

/**
 * Note: This is almost a copy of general vmtruncate(), just with inode->i_lock around the i_size
 * updates.
 *
 * @param offset file offset for truncation
 */
int FhgfsOps_vmtruncate(struct inode* inode, loff_t offset)
{
   if(i_size_read(inode) < offset)
   {
      unsigned long limit;

      limit = current->signal->rlim[RLIMIT_FSIZE].rlim_cur;
      if (limit != RLIM_INFINITY && offset > limit)
         goto out_sig;

      if (offset > inode->i_sb->s_maxbytes)
         goto out_big;

      spin_lock(&inode->i_lock);
      i_size_write(inode, offset);
      spin_unlock(&inode->i_lock);
   }
   else
   {
      struct address_space *mapping = inode->i_mapping;

      /*
       * truncation of in-use swapfiles is disallowed - it would
       * cause subsequent swapout to scribble on the now-freed
       * blocks.
       */
      if(IS_SWAPFILE(inode) )
         return -ETXTBSY;

      spin_lock(&inode->i_lock);
      i_size_write(inode, offset);
      spin_unlock(&inode->i_lock);

      /*
       * unmap_mapping_range is called twice, first simply for
       * efficiency so that truncate_inode_pages does fewer
       * single-page unmaps.  However after this first call, and
       * before truncate_inode_pages finishes, it is possible for
       * private pages to be COWed, which remain after
       * truncate_inode_pages finishes, hence the second
       * unmap_mapping_range call must be made for correctness.
       */
      unmap_mapping_range(mapping, offset + PAGE_SIZE - 1, 0, 1);
      truncate_inode_pages(mapping, offset);
      unmap_mapping_range(mapping, offset + PAGE_SIZE - 1, 0, 1);
   }

   return 0;

out_sig:
   send_sig(SIGXFSZ, current, 0);
out_big:
   return -EFBIG;
}

/**
 * Note: Call this once during module init (and remember to call _destroyInodeCache() )
 */
fhgfs_bool FhgfsOps_initInodeCache()
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,17)
   unsigned long cacheFlags = SLAB_RECLAIM_ACCOUNT;
#else
   unsigned long cacheFlags = SLAB_RECLAIM_ACCOUNT | SLAB_MEM_SPREAD;
#endif // LINUX_VERSION_CODE

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)
   FhgfsInodeCache = kmem_cache_create("fhgfs_inode_cache", sizeof(FhgfsInode), 0,
      cacheFlags, FhgfsOps_initInodeOnce, NULL);
#else
   FhgfsInodeCache = kmem_cache_create("fhgfs_inode_cache", sizeof(FhgfsInode), 0,
      cacheFlags, FhgfsOps_initInodeOnce);
#endif // LINUX_VERSION_CODE


   if(FhgfsInodeCache == NULL)
      return fhgfs_false;

   return fhgfs_true;
}

void FhgfsOps_destroyInodeCache()
{
   if(FhgfsInodeCache)
      kmem_cache_destroy(FhgfsInodeCache);
}


struct inode* FhgfsOps_alloc_inode(struct super_block *sb)
{
   FhgfsInode* fhgfsInode;

   fhgfsInode = kmem_cache_alloc(FhgfsInodeCache, GFP_KERNEL);
   if(unlikely(!fhgfsInode) )
      return NULL;

   FhgfsInode_allocInit(fhgfsInode);

   fhgfsInode->vfs_inode.i_blkbits = FHGFS_INODE_BLOCKBITS;

   return (struct inode*)fhgfsInode;
}

void FhgfsOps_destroy_inode(struct inode* inode)
{
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);

   FhgfsInode_destroyUninit(fhgfsInode);

   kmem_cache_free(FhgfsInodeCache, inode);
}

/**
 * Creates a new inode, inits it from the kstat, inits the ops (depending on the mode)
 * and hashes it.
 * Note: Make sure everything is set in the kstat _before_ you call this, because we hash
 * the inode in here.
 * Note2: This should not be called directly. Instead, do the remote stuff first and then
 * use the _instantiateInode()-wrapper
 *
 * @param kstat must have a valid .ino (inode number)
 * @param dev set to 0 if not required
 * @param entryInfo contained values (not the struct itself!) will be owned by the inode
 * (or kfreed on error)
 * @return NULL if not successful
 */
struct inode* __FhgfsOps_newInode(struct super_block* sb, struct kstat* kstat, dev_t dev,
   OwnerMinInfo* entryInfo)
{
   App* app = FhgfsOps_getApp(sb);
   Config* cfg = App_getConfig(app);

   FhgfsInode* fhgfsInode;

   struct inode* inode = new_inode(sb);
   if(unlikely(!inode) )
      goto err_cleanup_entryInfo;

   fhgfsInode = FHGFS_INODE(inode);

   __FhgfsOps_applyStatDataToInodeUnlocked(kstat, inode); // (nobody is using inode yet => unlocked)

   inode->i_ino = kstat->ino; // pre-set by caller

   inode->i_flags |= S_NOATIME | S_NOCMTIME; // timestamps updated by server

   fhgfsInode->entryInfo = *entryInfo;

   switch (kstat->mode & S_IFMT)
   {
      case S_IFREG:
      {
         if(Config_getTuneFileCacheTypeNum(cfg) == FILECACHETYPE_Paged)
         { // with pagecache
            inode->i_op = &fhgfs_file_pagecache_inode_ops;
            inode->i_fop = &fhgfs_file_pagecache_ops;
            inode->i_data.a_ops = &fhgfs_address_pagecache_ops;

            //inode->i_mapping->backing_dev_info = &fhgfs_backing_dev_info; // debug in
         }
         else
         { // no pagecache (=> either none or simple cache)
            inode->i_op = &fhgfs_file_inode_ops;
            inode->i_fop = &fhgfs_file_ops;
            inode->i_data.a_ops = &fhgfs_address_ops;
         }

      } break;
      case S_IFDIR:
      {
         inode->i_op = &fhgfs_dir_inode_ops;
         inode->i_fop = &fhgfs_dir_ops;

         /* directory inodes start off with i_nlink == 2 (for "." entry) */
         //inode->i_nlink = 2; // not necessary - servers send the nlink count
      } break;
      case S_IFLNK:
      {
         //inode->i_op = &page_symlink_inode_operations;
         inode->i_op = &fhgfs_symlink_inode_ops;
      } break;

      default:
      {
         init_special_inode(inode, kstat->mode, dev);
      } break;
   }

   insert_inode_hash(inode);

   return inode;

err_cleanup_entryInfo:
   FhgfsInode_freeOwnerMinInfoVals(entryInfo);

   return NULL;
}

/**
 * Retrieves the attribs from the metadata node, generates an inode number (ID) and instantiates
 * a local version of the inode for the dentry.
 *
 * Note: This is a wrapper for _newInode() that also retrieves the the entry attributes (so the
 * entry must exist already on the server).
 *
 * @param entryInfo contained values (not the struct itself!) will be owned by the inode
 * (or kfreed on error)
 * @param dentry the new entry
 * @return 0 on success, negative linux error code otherwise
 */
int __FhgfsOps_instantiateInode(struct dentry* dentry, OwnerMinInfo* entryInfo)
{
   App* app = FhgfsOps_getApp(dentry->d_sb);
   int retVal = 0;
   fhgfs_stat fhgfsStat;
   FhgfsOpsErr statRes;

   statRes = FhgfsOpsRemoting_statDirect(app, entryInfo, &fhgfsStat);

   if(statRes != FhgfsOpsErr_SUCCESS)
   { // error
      retVal = FhgfsOpsErr_toSysErr(statRes);
   }
   else
   { // success (entry exists on server)
      struct kstat kstat;
      struct inode* newInode;

      OsTypeConv_kstatFhgfsToOs(&fhgfsStat, &kstat);

      kstat.ino = iunique(dentry->d_sb, 2);

      newInode = __FhgfsOps_newInode(dentry->d_sb, &kstat, 0, entryInfo);
      if(!newInode)
         retVal = -EACCES;
      else
      { // new inode created
         d_instantiate(dentry, newInode);
      }
   }

   return retVal;
}

/**
 * Called when we want to check whether the inode has changed on the server (or to update
 * the inode attribs after we've made changes to a file/dir).
 * (If it has changed, we invalidate the caches.)
 *
 * Note: This method will indirectly aqcuire the i_lock spinlock.
 *
 * @return 0 on success (validity), negative linux error code if no longer valid
 */
int __FhgfsOps_refreshInode(App* app, struct inode* inode)
{
   const char* logContext = "FhgfsOps_refreshInode";
   Logger* log = App_getLogger(app);
   Config* cfg = App_getConfig(app);

   int retVal = 0;
   struct kstat kstat;
   FhgfsOpsErr statRes;
   fhgfs_stat fhgfsStat;
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);
   OwnerMinInfo entryInfo;

   time_t oldMTime;
   loff_t oldSize;
   unsigned cacheElapsedMS;
   fhgfs_bool mtimeSizeInvalidate;
   fhgfs_bool timeoutInvalidate;

   LOG_DEBUG_FORMATTED(log, 5, logContext, "called.");
   IGNORE_UNUSED_VARIABLE(log);
   IGNORE_UNUSED_VARIABLE(logContext);


   if(inode->i_mapping)
   { // flush out file contents for correct file size
      int flushRes = filemap_fdatawrite(inode->i_mapping);
      if(flushRes)
         return retVal;
   }

   FhgfsInode_getOwnerMinInfoCopy(fhgfsInode, &entryInfo);

   statRes = FhgfsOpsRemoting_statDirect(app, &entryInfo, &fhgfsStat);
   if(statRes != FhgfsOpsErr_SUCCESS)
   { // error or entry doesn't exist
      retVal = FhgfsOpsErr_toSysErr(statRes);
      goto clean_up;
   }

   // entry exists
   OsTypeConv_kstatFhgfsToOs(&fhgfsStat, &kstat);

   if(inode->i_ino == FHGFS_INODE_ROOT_INO)
   { // root node => deserves special handling (less checking)
      __FhgfsOps_applyStatDataToInode(&kstat, inode);

      goto clean_up;
   }

   // any existing node

   // check whether it is still the same object type
   if(unlikely( (inode->i_mode & S_IFMT) != (kstat.mode & S_IFMT) ) )
   { // object type changed => limit damage by marking it as bad
      // Note: This is quite impossible since we're verifying by entryID and an ID should
         // never ever be assigned to another object type.

      umode_t savedMode = inode->i_mode; // save mode

      make_bad_inode(inode);

      inode->i_mode = savedMode; // restore mode

      printk_fhgfs(KERN_WARNING, "%s: Object type changed for EntryID '%s'. Strange...\n",
         logContext, entryInfo.entryID);

      // invalidate cached pages
      if(!S_ISDIR(inode->i_mode) )
      {
         invalidate_remote_inode(inode);
      }

      retVal = -ENOENT;
      goto clean_up;
   }


   spin_lock(&inode->i_lock); // I _ L O C K

   oldMTime = inode->i_mtime.tv_sec;
   oldSize = i_size_read(inode);

   __FhgfsOps_applyStatDataToInodeUnlocked(&kstat, inode);

   mtimeSizeInvalidate =
      (inode->i_mtime.tv_sec != oldMTime) || (i_size_read(inode) != oldSize);
   cacheElapsedMS = Time_elapsedMS(&fhgfsInode->dataCacheTime);
   timeoutInvalidate = cacheElapsedMS > Config_getTunePageCacheValidityMS(cfg);

   if(S_ISDIR(inode->i_mode) )
   {
      Time_setToNow(&fhgfsInode->dataCacheTime);
   }
   else
   if(mtimeSizeInvalidate || timeoutInvalidate)
   { // file contents changed => invalidate non-dirty pages
      spin_unlock(&inode->i_lock); // I _ U N L O C K
      invalidate_remote_inode(inode); // might sleep => unlocked
      spin_lock(&inode->i_lock); // I _ R E L O C K

      Time_setToNow(&fhgfsInode->dataCacheTime); // after invalidate_... to avoid race condition
   }

   spin_unlock(&inode->i_lock); // I _ U N L O C K


   // clean up
clean_up:
   FhgfsInode_freeOwnerMinInfoVals(&entryInfo);

   return retVal;
}

/**
 * Note: Refreshes dir attribs only if cache validity timeout expired.
 *
 * @return 0 on success (validity), negative linux error code if no longer valid
 */
int __FhgfsOps_revalidateDirInode(App* app, struct inode* inode)
{
   const char* logContext = "FhgfsOps_revalidateMapping";
   Logger* log = App_getLogger(app);
   Config* cfg = App_getConfig(app);

   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);

   unsigned cacheValidityMS = Config_getTuneAttribCacheValidityMS(cfg);
   unsigned cacheElapsedMS;
   fhgfs_bool timeoutInvalidate;
   fhgfs_bool needRefresh = fhgfs_false;

   LOG_DEBUG_FORMATTED(log, 5, logContext, "called.");
   IGNORE_UNUSED_VARIABLE(log);
   IGNORE_UNUSED_VARIABLE(logContext);

   if(!cacheValidityMS) // caching disabled
      return __FhgfsOps_refreshInode(app, inode);

   spin_lock(&inode->i_lock); // I _ L O C K

   cacheElapsedMS = Time_elapsedMS(&fhgfsInode->dataCacheTime);
   timeoutInvalidate = cacheElapsedMS > cacheValidityMS;

   if(timeoutInvalidate)
      needRefresh = fhgfs_true;

   spin_unlock(&inode->i_lock); // I _ U N L O C K

   if(needRefresh)
      return __FhgfsOps_refreshInode(app, inode);

   return 0;
}

/**
 * Note: Refreshes inode and mapping only if cache validity timeout expired.
 *
 * @return 0 on success (validity), negative linux error code if no longer valid
 */
int __FhgfsOps_revalidateMapping(App* app, struct inode* inode)
{
   const char* logContext = "FhgfsOps_revalidateMapping";
   Logger* log = App_getLogger(app);
   Config* cfg = App_getConfig(app);

   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);

   unsigned cacheElapsedMS;
   fhgfs_bool timeoutInvalidate;
   fhgfs_bool needRefresh = fhgfs_false;

   LOG_DEBUG_FORMATTED(log, 5, logContext, "called.");
   IGNORE_UNUSED_VARIABLE(log);
   IGNORE_UNUSED_VARIABLE(logContext);

   spin_lock(&inode->i_lock); // I _ L O C K

   cacheElapsedMS = Time_elapsedMS(&fhgfsInode->dataCacheTime);
   timeoutInvalidate = cacheElapsedMS > Config_getTunePageCacheValidityMS(cfg);

   if(!S_ISDIR(inode->i_mode) && timeoutInvalidate)
      needRefresh = fhgfs_true;

   spin_unlock(&inode->i_lock); // I _ U N L O C K

   if(needRefresh)
      return __FhgfsOps_refreshInode(app, inode);

   return 0;
}
