#include <include_closed/app/log/Logger.h>
#include <include_closed/app/App.h>
#include <include_closed/app/config/Config.h>
#include <open/filesystem/ProcFs.h>
#include <include_closed/common/toolkit/MetadataTk.h>
#include <include_closed/net/filesystem/FhgfsOpsRemoting.h>
#include "FhgfsOps_versions.h"
#include "FhgfsOpsSuper.h"
#include "FhgfsOpsInode.h"
#include "FhgfsOpsFile.h"



static struct file_system_type fhgfs_fs_type =
{
   .name       = "fhgfs",
   .owner      = THIS_MODULE,
   .get_sb     = __FhgfsOps_getSB,
   .kill_sb    = FhgfsOps_killSB,
   //.fs_flags   = FS_BINARY_MOUNTDATA, // not required currently
};

static struct super_operations fhgfs_super_ops =
{
   .statfs        = FhgfsOps_statfs,
   .alloc_inode   = FhgfsOps_alloc_inode,
   .destroy_inode = FhgfsOps_destroy_inode,
   .drop_inode    = generic_delete_inode,
   .put_super     = FhgfsOps_putSuper,
};


/**
 * Creates and initializes the per-mount application object.
 */
int __FhgfsOps_initApp(struct super_block* sb, char* rawMountOptions)
{
   MountConfig* mountConfig;
   fhgfs_bool parseRes;
   App* app;
   int appRes;

   // create mountConfig (parse from mount options)
   mountConfig = MountConfig_construct();

   parseRes = MountConfig_parseFromRawOptions(mountConfig, rawMountOptions);
   if(!parseRes)
   {
      MountConfig_destruct(mountConfig);
      return APPCODE_INVALID_CONFIG;
   }


   //printk_fhgfs(FHGFS_KERN_INFO, "Initializing App...\n"); // debug in

   app = App_construct(mountConfig);

   //printk_fhgfs(FHGFS_KERN_INFO, "Running App...\n"); // debug in

   appRes = App_run(app);

   if(appRes != APPCODE_NO_ERROR)
   { // error occured => clean up
      printk_fhgfs_debug(FHGFS_KERN_INFO, "Stopping App...\n");

      App_stop(app);

      printk_fhgfs_debug(FHGFS_KERN_INFO, "Cleaning up...\n");

      App_destruct(app);

      printk_fhgfs_debug(FHGFS_KERN_INFO, "App destructed.\n");

      return appRes;
   }

   FhgfsOps_setApp(sb, app);

   ProcFs_createEntries(app);

   // app running
   printk_fhgfs(FHGFS_KERN_INFO, "FhGFS mount ready.\n");

   return appRes;
}

/**
 * Stops and destroys the per-mount application object.
 */
void __FhgfsOps_uninitApp(App* app)
{
   //printk_fhgfs(FHGFS_KERN_INFO, "fhgfs_client preparing exit...\n");

   ProcFs_removeEntries(app);

   //printk_fhgfs(FHGFS_KERN_INFO, "Stopping App...\n"); // debug in

   App_stop(app);

   //printk_fhgfs(FHGFS_KERN_INFO, "Cleaning up...\n");

   App_destruct(app);

   //printk_fhgfs(FHGFS_KERN_INFO, "App destructed.\n");

   printk_fhgfs(FHGFS_KERN_INFO, "FhGFS unmounted.\n");
}

int FhgfsOps_registerFilesystem()
{
   return register_filesystem(&fhgfs_fs_type);
}

int FhgfsOps_unregisterFilesystem()
{
   return unregister_filesystem(&fhgfs_fs_type);
}

int __FhgfsOps_fillSuper(struct super_block* sb, void* rawMountOptions, int silent)
{
   //Logger* log = App_getLogger(Program_getApp() );
   //const char* logContext = "FhgfsOps_fillSuper";

   struct inode* rootInode;
   struct dentry* rootDentry;
   struct kstat kstat;
   OwnerMinInfo entryInfo;

   // init per-mount app object

   if(__FhgfsOps_initApp(sb, rawMountOptions) )
      return -EPERM;

   // set up super block data

   sb->s_maxbytes = MAX_LFS_FILESIZE;
   sb->s_blocksize = PAGE_CACHE_SIZE;
   sb->s_blocksize_bits = PAGE_CACHE_SHIFT;
   sb->s_magic = FHGFS_MAGIC;
   sb->s_op = &fhgfs_super_ops;
   sb->s_time_gran = 1000000000; // granularity of c/m/atime in ns
   sb->s_flags |= MS_NODIRATIME;

   //Logger_logFormatted(log, 5, logContext, "MS_NOEXEC specified: %d",
   //   (int)(sb->s_flags & MS_NOEXEC) );

   // init root inode

   memset(&kstat, 0, sizeof(struct kstat) );

   kstat.ino = 2; /* traditional root inode number */
   kstat.mode = S_IFDIR | 0777; // allow access for everyone
   kstat.atime = kstat.mtime = kstat.ctime = current_fs_time(sb);
   kstat.uid = currentFsUserID;
   kstat.gid = currentFsGroupID;
   kstat.blksize = FHGFS_INODE_BLOCKSIZE; /*PAGE_CACHE_SIZE*/
   kstat.nlink = 1;

   // root ownerInfo is always updated when someone asks for it (so we just set dummy values here)
   entryInfo.ownerNodeID = StringTk_strDup("");
   entryInfo.entryID = StringTk_strDup("");

   rootInode = __FhgfsOps_newInode(sb, &kstat, 0, &entryInfo);
   if(!rootInode)
   {
      __FhgfsOps_uninitApp(FhgfsOps_getApp(sb) );
      return -ENOMEM;
   }

   rootDentry = d_alloc_root(rootInode);
   if(!rootDentry)
   {
      __FhgfsOps_uninitApp(FhgfsOps_getApp(sb) );
      iput(rootInode);
      return -ENOMEM;
   }

   rootDentry->d_time = jiffies;
   sb->s_root = rootDentry;

   return 0;
}

void FhgfsOps_putSuper(struct super_block* sb)
{
   App* app = FhgfsOps_getApp(sb);

   if(app)
      __FhgfsOps_uninitApp(app);
}

void FhgfsOps_killSB(struct super_block* sb)
{
   // this is just a debug helper block (for dentry reference counters)
   /*
   printk_fhgfs(FHGFS_KERN_INFO, "calling autfs4_force_release\n"); // debug add
   __FhgfsOps_forceDentryRelease(sb); // debug add
   printk_fhgfs(FHGFS_KERN_INFO, "after autfs4_force_release\n"); // debug add
   */

   kill_anon_super(sb);
}

/**
 * Releases dentry references (during unmount).
 *
 * Note: This is meant to be just a debug helper for dentry reference counters. If everything is
 * okay with the reference counters, then this method is not needed and shouldn't be used.
 */
void __FhgfsOps_forceDentryRelease(struct super_block *sb)
{
   struct dentry* this_parent = sb->s_root;
   struct list_head* next;

   if (!sb->s_root)
      return;

   spin_lock(&dcache_lock);
repeat:
   next = this_parent->d_subdirs.next;
resume:
   while(next != &this_parent->d_subdirs)
   {
      struct dentry *dentry = list_entry(next, struct dentry, d_u.d_child);

      /* Negative dentry - don't care */
      if (!dentry->d_inode)
      {
         next = next->next;
         continue;
      }

      if(!list_empty(&dentry->d_subdirs) )
      {
         this_parent = dentry;
         goto repeat;
      }

      next = next->next;

      spin_unlock(&dcache_lock);

      printk_fhgfs(FHGFS_KERN_INFO, "cleaning up dentry %p %.*s (d_count: %lld)\n",
         dentry, (int)dentry->d_name.len, dentry->d_name.name,
         (long long)atomic_read(&dentry->d_count) ); // debug comment


      while(atomic_read(&dentry->d_count) >= 1) // we need >=1 here (not > 1)!
         dput(dentry);

      spin_lock(&dcache_lock);
   }

   if(this_parent != sb->s_root)
   {
      struct dentry *dentry = this_parent;

      next = this_parent->d_u.d_child.next;
      this_parent = this_parent->d_parent;

      spin_unlock(&dcache_lock);

      printk_fhgfs(FHGFS_KERN_INFO, "parent dentry %p %.*s (d_count: %lld)\n",
         dentry, (int)dentry->d_name.len, dentry->d_name.name,
         (long long)atomic_read(&dentry->d_count) ); // debug comment

      while(atomic_read(&dentry->d_count) > 1) // we need >1 here (not >= 1)!
         dput(dentry);

      spin_lock(&dcache_lock);

      goto resume;
   }

   spin_unlock(&dcache_lock);
}
