#include <include_closed/toolkit/NoAllocBufferStore.h>
#include <include_closed/net/filesystem/FhgfsOpsRemoting.h>
#include <include_closed/filesystem/FhgfsOpsHelper.h>
#include "FhgfsOps_versions.h"
#include "FhgfsOpsFile.h"
#include "FhgfsOpsDir.h"
#include "FhgfsOpsInode.h"
#include "FhgfsOpsSuper.h"

#include <linux/namei.h>


#ifndef KERNEL_HAS_STATFS_DENTRY

int FhgfsOps_statfs(struct super_block* sb, struct kstatfs* kstatfs)
{

#else

int FhgfsOps_statfs(struct dentry* dentry, struct kstatfs* kstatfs)
{
   struct super_block* sb = dentry->d_sb;

#endif // KERNEL_HAS_STATFS_DENTRY


   App* app = FhgfsOps_getApp(sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_statfs";


   int retVal = -EREMOTEIO;
   FhgfsOpsErr statRes;

   int64_t sizeTotal = 0;
   int64_t sizeFree = 0;

   Logger_logFormatted(log, 5, logContext, "called.");

   memset(kstatfs, 0, sizeof(struct kstatfs) );

   kstatfs->f_type      = FHGFS_MAGIC;

   kstatfs->f_namelen   = NAME_MAX;
   kstatfs->f_files     = 0; // total used file nodes (not supported currently)
   kstatfs->f_ffree     = 0; // free file nodes (not supported currently)

   kstatfs->f_frsize    = sb->s_blocksize; // should be same as f_bsize
   kstatfs->f_bsize     = sb->s_blocksize;

   statRes = FhgfsOpsRemoting_statStoragePath(app, &sizeTotal, &sizeFree);

   LOG_DEBUG_FORMATTED(log, 4, logContext, "remoting complete. result: %d", (int)statRes);

   retVal = FhgfsOpsErr_toSysErr(statRes);

   if(statRes == FhgfsOpsErr_SUCCESS)
   { // success => assign received values

      // note: f_blocks, f_bfree, and f_bavail are reported in units of f_bsize.

      unsigned char blockBits = sb->s_blocksize_bits; // just a shorter name
      unsigned long blockSizeDec = (1 << blockBits) - 1; // for rounding

      kstatfs->f_blocks = (sizeTotal + blockSizeDec) >> blockBits;
      kstatfs->f_bfree  = (sizeFree  + blockSizeDec) >> blockBits; // free for superuser
      kstatfs->f_bavail = (sizeFree  + blockSizeDec) >> blockBits; // available for non-superuser
   }

   return retVal;
}



#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,17)

struct super_block* __FhgfsOps_getSB(struct file_system_type *fs_type,
   int flags, const char *dev_name, void *data)
{
   return get_sb_nodev(fs_type, flags, data, __FhgfsOps_fillSuper);
}

#else

int __FhgfsOps_getSB(struct file_system_type *fs_type,
   int flags, const char *dev_name, void *data, struct vfsmount *mnt)
{
   return get_sb_nodev(fs_type, flags, data, __FhgfsOps_fillSuper, mnt);
}

#endif // LINUX_VERSION_CODE


/**
 * Note: Called by close(2) system call.
 */
#ifndef KERNEL_HAS_FLUSH_OWNER
int FhgfsOps_flush(struct file *file)
#else
int FhgfsOps_flush(struct file *file, fl_owner_t id)
#endif // KERNEL_HAS_FLUSH_OWNER
{
   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_flush";

   struct inode* inode = file->f_dentry->d_inode;
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);
   FsFileInfo* fileInfo = __FhgfsOps_getFileInfo(file);
   RemotingIOInfo ioInfo;
   FhgfsOpsErr flushRes;
   int retVal = 0;

   Logger_logFormatted(log, 5, logContext, "called. Path: %s", "n/a");

   //printk_fhgfs_debug(FHGFS_KERN_INFO, "waiting for flush page writes\n"); // debug in (delete me)

   retVal = filemap_write_and_wait(inode->i_mapping);

   //printk_fhgfs_debug(FHGFS_KERN_INFO, "flush page writes complete\n"); // debug in (delete me)

   FsFileInfo_getIOInfo(fileInfo, fhgfsInode, &ioInfo);

   flushRes = FhgfsOpsHelper_flushCache(app, fileInfo, &ioInfo, fhgfs_true);
   if(flushRes != FhgfsOpsErr_SUCCESS)
   { // error
      if(!retVal)
         retVal = FhgfsOpsErr_toSysErr(flushRes);
   }

   return retVal;
}

/**
 * Note: Called for new cache objects (see FhgfsOps_initInodeCache() )
 */
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,21)

void FhgfsOps_initInodeOnce(void* inode, struct kmem_cache* cache, unsigned long flags)
{
   // note: no, this is not a typo. since kernel version 2.6.22, we're no longer checking
   //    the slab_... flags (even though the argument still exists in the method signature).

   if( (flags & (SLAB_CTOR_VERIFY|SLAB_CTOR_CONSTRUCTOR) ) == SLAB_CTOR_CONSTRUCTOR)
   { // fresh construction => initialize object
      FhgfsInode* fhgfsInode = (FhgfsInode*)inode;

      FhgfsInode_initOnce(fhgfsInode);

      inode_init_once(&fhgfsInode->vfs_inode);
   }
}

#else

   #if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,23)
   void FhgfsOps_initInodeOnce(void* inode, struct kmem_cache* cache, unsigned long flags)
   #elif LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,26)
   void FhgfsOps_initInodeOnce(struct kmem_cache* cache, void* inode)
   #else
   void FhgfsOps_initInodeOnce(void* inode)
   #endif // LINUX_VERSION_CODE

   {
      FhgfsInode* fhgfsInode = (FhgfsInode*)inode;

      FhgfsInode_initOnce(fhgfsInode);

      inode_init_once(&fhgfsInode->vfs_inode);
   }

#endif // LINUX_VERSION_CODE


#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,26)

/**
 * Note: This is just an exact copy of the kernel method (from kernel 2.6.30), which was introduced
 * in 2.6.27.
 *
 * generic_file_llseek_unlocked - lockless generic llseek implementation
 * @file:   file structure to seek on
 * @offset: file offset to seek to
 * @origin: type of seek
 *
 * Updates the file offset to the value specified by @offset and @origin.
 * Locking must be provided by the caller.
 */
loff_t generic_file_llseek_unlocked(struct file *file, loff_t offset, int origin)
{
   struct inode *inode = file->f_mapping->host;

   switch (origin)
   {
      case SEEK_END:
         offset += inode->i_size;
         break;
      case SEEK_CUR:
         /*
          * Here we special-case the lseek(fd, 0, SEEK_CUR)
          * position-querying operation.  Avoid rewriting the "same"
          * f_pos value back to the file because a concurrent read(),
          * write() or lseek() might have altered it
          */
         if (offset == 0)
            return file->f_pos;
         offset += file->f_pos;
         break;
   }

   if (offset < 0 || offset > inode->i_sb->s_maxbytes)
      return -EINVAL;

   /* Special lock needed here? */
   if (offset != file->f_pos)
   {
      file->f_pos = offset;
      file->f_version = 0;
   }

   return offset;
}

#endif // LINUX_VERSION_CODE

#ifndef KERNEL_HAS_DROP_NLINK

/**
 * Note: This is just an exact copy of the kernel method (from kernel 2.6.30), which was introduced
 * in 2.6.19.
 *
 * drop_nlink - directly drop an inode's link count
 * @inode: inode
 *
 * This is a low-level filesystem helper to replace any
 * direct filesystem manipulation of i_nlink.  In cases
 * where we are attempting to track writes to the
 * filesystem, a decrement to zero means an imminent
 * write when the file is truncated and actually unlinked
 * on the filesystem.
 */
void drop_nlink(struct inode *inode)
{
   inode->i_nlink--;
}

#endif // KERNEL_HAS_DROP_NLINK

#ifndef KERNEL_HAS_CLEAR_NLINK
/**
 * Note: This is just an exact copy of the kernel method (from kernel 2.6.30), which was introduced
 * in 2.6.19.
 *
 * clear_nlink - directly zero an inode's link count
 * @inode: inode
 *
 * This is a low-level filesystem helper to replace any
 * direct filesystem manipulation of i_nlink.  See
 * drop_nlink() for why we care about i_nlink hitting zero.
 */
void clear_nlink(struct inode *inode)
{
   inode->i_nlink = 0;
}

#endif // KERNEL_HAS_CLEAR_NLINK

