#include <include_closed/filesystem/ProcFsHelper.h>
#include "ProcFs.h"

#include <linux/proc_fs.h>
#include <linux/stat.h>


#define FHGFS_PROC_DIR_NAME      "fs/fhgfs"
#define FHGFS_PROC_NAMEBUF_LEN   4096

#define FHGFS_PROC_ENTRY_CONFIG              "config"
#define FHGFS_PROC_ENTRY_MGMTNODES           "mgmt_nodes"
#define FHGFS_PROC_ENTRY_METANODES           "meta_nodes"
#define FHGFS_PROC_ENTRY_STORAGENODES        "storage_nodes"
#define FHGFS_PROC_ENTRY_CLIENTINFO          "client_info"
#define FHGFS_PROC_ENTRY_REMOVEMGMTNODE      "remove_mgmt_node"
#define FHGFS_PROC_ENTRY_REMOVEMETANODE      "remove_meta_node"
#define FHGFS_PROC_ENTRY_REMOVESTORAGENODE   "remove_storage_node"
#define FHGFS_PROC_ENTRY_RETRIESENABLED      "conn_retries_enabled"


/**
 * Creates the general parent dir. Meant to be called only once at module load.
 */
void ProcFs_createGeneralDir()
{
   struct proc_dir_entry* procDir;
   
   procDir = proc_mkdir(FHGFS_PROC_DIR_NAME, NULL);
   if(!procDir)
      printk_fhgfs(FHGFS_KERN_INFO, "Failed to create proc dir: " FHGFS_PROC_DIR_NAME "\n");
   
   
}

/**
 * Removes the general parent dir. Meant to be called only once at module unload.
 */
void ProcFs_removeGeneralDir()
{
   remove_proc_entry(FHGFS_PROC_DIR_NAME, NULL);
}

/**
 * Creates the dir and entries for a specific mountpoint.
 * Note: Uses sessionID to create a unique dir for the mountpoint.
 */
void ProcFs_createEntries(App* app)
{
   struct proc_dir_entry* procDir;
   struct proc_dir_entry* currentEntry;
   
   char* dirNameBuf = vmalloc(FHGFS_PROC_NAMEBUF_LEN);
   
   const char* sessionID = ProcFsHelper_getSessionID(app);
   
   scnprintf(dirNameBuf, FHGFS_PROC_NAMEBUF_LEN, FHGFS_PROC_DIR_NAME "/%s", sessionID);

   // create unique directory for this clientID
   procDir = proc_mkdir(dirNameBuf, NULL);
   if(!procDir)
      goto clean_up;

   
   // create entries
   
   currentEntry = create_proc_read_entry(FHGFS_PROC_ENTRY_CONFIG, 0444, procDir,
      ProcFs_read_config, app);
   currentEntry = create_proc_read_entry(FHGFS_PROC_ENTRY_MGMTNODES, 0444, procDir,
      ProcFs_read_mgmtNodes, app);
   currentEntry = create_proc_read_entry(FHGFS_PROC_ENTRY_METANODES, 0444, procDir,
      ProcFs_read_metaNodes, app);
   currentEntry = create_proc_read_entry(FHGFS_PROC_ENTRY_STORAGENODES, 0444, procDir,
      ProcFs_read_storageNodes, app);
   currentEntry = create_proc_read_entry(FHGFS_PROC_ENTRY_CLIENTINFO, 0444, procDir,
      ProcFs_read_clientInfo, app);
   
   currentEntry = create_proc_entry(FHGFS_PROC_ENTRY_REMOVEMGMTNODE, 0664, procDir);
   if(currentEntry)
   {
      currentEntry->read_proc = ProcFs_read_nothing;
      currentEntry->write_proc = ProcFs_write_removeMgmtNode;
      currentEntry->data = app;
   }

   currentEntry = create_proc_entry(FHGFS_PROC_ENTRY_REMOVEMETANODE, 0664, procDir);
   if(currentEntry)
   {
      currentEntry->read_proc = ProcFs_read_nothing;
      currentEntry->write_proc = ProcFs_write_removeMetaNode;
      currentEntry->data = app;
   }
   
   currentEntry = create_proc_entry(FHGFS_PROC_ENTRY_REMOVESTORAGENODE, 0664, procDir);
   if(currentEntry)
   {
      currentEntry->read_proc = ProcFs_read_nothing;
      currentEntry->write_proc = ProcFs_write_removeStorageNode;
      currentEntry->data = app;
   }

   currentEntry = create_proc_entry(FHGFS_PROC_ENTRY_RETRIESENABLED, 0664, procDir);
   if(currentEntry)
   {
      currentEntry->read_proc = ProcFs_read_connRetriesEnabled;
      currentEntry->write_proc = ProcFs_write_connRetriesEnabled;
      currentEntry->data = app;
   }
   
clean_up:
   SAFE_VFREE(dirNameBuf);
}

/**
 * Removes the dir and entries for a specific mountpoint.
 * Note: Uses sessionID for unique dir of the mountpoint.
 */
void ProcFs_removeEntries(App* app)
{
   char* dirNameBuf = vmalloc(FHGFS_PROC_NAMEBUF_LEN);
   char* entryNameBuf = vmalloc(FHGFS_PROC_NAMEBUF_LEN);
   
   const char* sessionID = ProcFsHelper_getSessionID(app);
   
   scnprintf(dirNameBuf, FHGFS_PROC_NAMEBUF_LEN, FHGFS_PROC_DIR_NAME "/%s", sessionID);

   // remove entries
   scnprintf(entryNameBuf, FHGFS_PROC_NAMEBUF_LEN, "%s/%s",
      dirNameBuf, FHGFS_PROC_ENTRY_CONFIG);
   remove_proc_entry(entryNameBuf, NULL);

   scnprintf(entryNameBuf, FHGFS_PROC_NAMEBUF_LEN, "%s/%s",
      dirNameBuf, FHGFS_PROC_ENTRY_MGMTNODES);
   remove_proc_entry(entryNameBuf, NULL);

   scnprintf(entryNameBuf, FHGFS_PROC_NAMEBUF_LEN, "%s/%s",
      dirNameBuf, FHGFS_PROC_ENTRY_METANODES);
   remove_proc_entry(entryNameBuf, NULL);

   scnprintf(entryNameBuf, FHGFS_PROC_NAMEBUF_LEN, "%s/%s",
      dirNameBuf, FHGFS_PROC_ENTRY_STORAGENODES);
   remove_proc_entry(entryNameBuf, NULL);

   scnprintf(entryNameBuf, FHGFS_PROC_NAMEBUF_LEN, "%s/%s",
      dirNameBuf, FHGFS_PROC_ENTRY_CLIENTINFO);
   remove_proc_entry(entryNameBuf, NULL);
   
   scnprintf(entryNameBuf, FHGFS_PROC_NAMEBUF_LEN, "%s/%s",
      dirNameBuf, FHGFS_PROC_ENTRY_REMOVEMGMTNODE);
   remove_proc_entry(entryNameBuf, NULL);

   scnprintf(entryNameBuf, FHGFS_PROC_NAMEBUF_LEN, "%s/%s",
      dirNameBuf, FHGFS_PROC_ENTRY_REMOVEMETANODE);
   remove_proc_entry(entryNameBuf, NULL);

   scnprintf(entryNameBuf, FHGFS_PROC_NAMEBUF_LEN, "%s/%s",
      dirNameBuf, FHGFS_PROC_ENTRY_REMOVESTORAGENODE);
   remove_proc_entry(entryNameBuf, NULL);

   scnprintf(entryNameBuf, FHGFS_PROC_NAMEBUF_LEN, "%s/%s",
      dirNameBuf, FHGFS_PROC_ENTRY_RETRIESENABLED);
   remove_proc_entry(entryNameBuf, NULL);


   // remove unique dir
   remove_proc_entry(dirNameBuf, NULL);

   
   SAFE_VFREE(dirNameBuf);
   SAFE_VFREE(entryNameBuf);
}

/**
 * Does not return anything to the reading process.
 * 
 * @param data specified at entry creation
 */
int ProcFs_read_nothing(char* buf, char** start, off_t offset, int size, int* eof, void* data)
{
   *eof = 1;
   return 0;
}

/**
 * @param data specified at entry creation
 */
int ProcFs_read_config(char* buf, char** start, off_t offset, int size, int* eof, void* data)
{
   return ProcFsHelper_read_config(buf, start, offset, size, eof, (App*)data);
}

/**
 * @param data specified at entry creation
 */
int ProcFs_read_mgmtNodes(char* buf, char** start, off_t offset, int size, int* eof, void* data)
{
   return ProcFsHelper_read_nodes(buf, start, offset, size, eof,
      App_getMgmtNodes( (App*)data) );
}

/**
 * @param data specified at entry creation
 */
int ProcFs_read_metaNodes(char* buf, char** start, off_t offset, int size, int* eof, void* data)
{
   return ProcFsHelper_read_nodes(buf, start, offset, size, eof,
      App_getMetaNodes( (App*)data) );
}

/**
 * @param data specified at entry creation
 */
int ProcFs_read_storageNodes(char* buf, char** start, off_t offset, int size, int* eof, void* data)
{
   return ProcFsHelper_read_nodes(buf, start, offset, size, eof,
      App_getStorageNodes( (App*)data) );
}

int ProcFs_read_clientInfo(char* buf, char** start, off_t offset, int size, int* eof,
   void* data)
{
   return ProcFsHelper_read_clientInfo(buf, start, offset, size, eof, (App*)data);
}

/**
 * @param data specified at entry creation
 */
int ProcFs_write_removeMgmtNode(struct file* file, const char __user *buf,
   unsigned long count, void* data)
{
   // check user buffer
   if(unlikely(!access_ok(VERIFY_READ, buf, count) ) )
      return -EFAULT;
   
   return ProcFsHelper_write_removeNode(buf, count, (App*)data,
      App_getMgmtNodes( (App*)data) );
}

/**
 * @param data specified at entry creation
 */
int ProcFs_write_removeMetaNode(struct file* file, const char __user *buf,
   unsigned long count, void* data)
{
   // check user buffer
   if(unlikely(!access_ok(VERIFY_READ, buf, count) ) )
      return -EFAULT;
   
   return ProcFsHelper_write_removeNode(buf, count, (App*)data,
      App_getMetaNodes( (App*)data) );
}

/**
 * @param data specified at entry creation
 */
int ProcFs_write_removeStorageNode(struct file* file, const char __user *buf,
   unsigned long count, void* data)
{
   // check user buffer
   if(unlikely(!access_ok(VERIFY_READ, buf, count) ) )
      return -EFAULT;
   
   return ProcFsHelper_write_removeNode(buf, count, (App*)data,
      App_getStorageNodes( (App*)data) );
}

int ProcFs_read_connRetriesEnabled(char* buf, char** start, off_t offset, int size, int* eof,
   void* data)
{
   return ProcFsHelper_read_connRetriesEnabled(buf, start, offset, size, eof, (App*)data);
}

/**
 * @param data specified at entry creation
 */
int ProcFs_write_connRetriesEnabled(struct file* file, const char __user *buf,
   unsigned long count, void* data)
{
   // check user buffer
   if(unlikely(!access_ok(VERIFY_READ, buf, count) ) )
      return -EFAULT;
   
   return ProcFsHelper_write_connRetriesEnabled(buf, count, (App*)data);
   
   return -EPERM;
}
