#ifdef FHGFS_OPENTK_IBVERBS

#include "IBVSocket.h"

#include <linux/sched.h>
#include <linux/poll.h>
#include <linux/socket.h>


#define IBVSOCKET_CONN_TIMEOUT_MS         5000
#define IBVSOCKET_COMPLETION_TIMEOUT_MS   300000 /* this also includes send completion wait times */


void IBVSocket_init(IBVSocket* _this)
{
   memset(_this, 0, sizeof(*_this) );

   _this->sockValid = fhgfs_false;

   //Mutex_init(&_this->mutex);
   //Mutex_init(&_this->connStateMutex);
   //Condition_init(&_this->connStateCond);

   init_waitqueue_head(&_this->eventWaitQ);

   _this->cm_id = rdma_create_id(__IBVSocket_cmaHandler, _this, RDMA_PS_TCP);
   if(!_this->cm_id)
   {
      printk_fhgfs(KERN_INFO, "%d:%s: rdma_create_id failed\n", __LINE__, __func__);
      return;
   }

   _this->connState = IBVSOCKETCONNSTATE_UNCONNECTED;

   _this->sockValid = fhgfs_true;

   return;
}

EXPORT_SYMBOL(IBVSocket_init);


IBVSocket* IBVSocket_construct()
{
   IBVSocket* _this = (IBVSocket*)kmalloc(sizeof(*_this), GFP_KERNEL);

   IBVSocket_init(_this);

   return _this;
}

EXPORT_SYMBOL(IBVSocket_construct);

void IBVSocket_uninit(IBVSocket* _this)
{
   __IBVSocket_close(_this);

   //Condition_uninit(&_this->connStateCond);
   //Mutex_uninit(&_this->connStateMutex);
   //Mutex_uninit(&_this->mutex);
}

EXPORT_SYMBOL(IBVSocket_uninit);

void IBVSocket_destruct(IBVSocket* _this)
{
   IBVSocket_uninit(_this);

   kfree(_this);
}

EXPORT_SYMBOL(IBVSocket_destruct);

fhgfs_bool IBVSocket_rdmaDevicesExist()
{
   // Note: We use this (currently) just to inform the higher levels
   //    about availability of RDMA functionality
   return fhgfs_true;
}

EXPORT_SYMBOL(IBVSocket_rdmaDevicesExist);


fhgfs_bool IBVSocket_connectByIP(IBVSocket* _this, struct in_addr* ipaddress, unsigned short port,
   IBVCommConfig* commCfg)
{
   struct sockaddr_in sin;
   fhgfs_bool stateChangeRes;

   // Note: The rest of the connect procedure is invoked through the cmaHandler when the
   //    corresponding asynchronous events arrive => we just have to wait for the connState
   //    to change here

   __IBVSocket_setConnState(_this, IBVSOCKETCONNSTATE_CONNECTING);


   // resolve IP address ...
   // (and automatically also resolves route on success)

   sin.sin_addr.s_addr = ipaddress->s_addr;
   sin.sin_family = AF_INET;
   sin.sin_port = htons(port);

   if(rdma_resolve_addr(_this->cm_id, NULL, (struct sockaddr*)&sin, IBVSOCKET_CONN_TIMEOUT_MS) )
   {
      printk_fhgfs_debug(KERN_INFO, "%d:%s: rdma_resolve_addr failed\n", __LINE__, __func__);
      goto err_invalidateSock;
   }

   // wait for async event
   stateChangeRes = __IBVSocket_waitForConnStateChange(_this, IBVSOCKETCONNSTATE_CONNECTING);
   if(!stateChangeRes ||
      (__IBVSocket_getConnState(_this) != IBVSOCKETCONNSTATE_ROUTERESOLVED) )
      goto err_invalidateSock;


   // establish connection...

   // (handler calls rdma_connect() )
   if(__IBVSocket_routeResolvedHandler(_this, _this->cm_id, commCfg, &_this->commContext) )
   {
      printk_fhgfs_debug(KERN_INFO, "%d:%s: route resolved handler failed\n", __LINE__, __func__);
      goto err_invalidateSock;
   }

   // wait for async event
   stateChangeRes = __IBVSocket_waitForConnStateChange(_this, IBVSOCKETCONNSTATE_ROUTERESOLVED);
   if(!stateChangeRes ||
      (__IBVSocket_getConnState(_this) != IBVSOCKETCONNSTATE_ESTABLISHED) )
      goto err_invalidateSock;


   // connected

   return fhgfs_true;


err_invalidateSock:
   _this->errState = -1;

   return fhgfs_false;
}


EXPORT_SYMBOL(IBVSocket_connectByIP);

/**
 * @return fhgfs_true on success
 */
fhgfs_bool IBVSocket_bind(IBVSocket* _this, unsigned short port)
{
   struct in_addr ipAddr;

   ipAddr.s_addr = INADDR_ANY;

   return IBVSocket_bindToAddr(_this, &ipAddr, port);
}

EXPORT_SYMBOL(IBVSocket_bind);

fhgfs_bool IBVSocket_bindToAddr(IBVSocket* _this, struct in_addr* ipAddr, unsigned short port)
{
   struct sockaddr_in bindAddr;

   bindAddr.sin_family = AF_INET;
   bindAddr.sin_addr = *ipAddr;
   bindAddr.sin_port = htons(port);

   if(rdma_bind_addr(_this->cm_id, (struct sockaddr*)&bindAddr) )
   {
      //printk_fhgfs_debug(KERN_INFO, "%s:%d rdma_bind_addr failed (port: %d)\n",
      //   __func__, __LINE__, (int)port); // debug in
      goto err_invalidateSock;
   }

   return fhgfs_true;


err_invalidateSock:
   _this->errState = -1;

   return fhgfs_false;
}

EXPORT_SYMBOL(IBVSocket_bindToAddr);

/**
 * @return fhgfs_true on success
 */
fhgfs_bool IBVSocket_listen(IBVSocket* _this)
{
   if(rdma_listen(_this->cm_id, 0) )
   {
      printk_fhgfs(KERN_INFO, "%d:%s: rdma_listen failed\n", __LINE__, __func__);
      goto err_invalidateSock;
   }

   return fhgfs_true;


err_invalidateSock:
   _this->errState = -1;

   return fhgfs_false;
}


EXPORT_SYMBOL(IBVSocket_listen);

fhgfs_bool IBVSocket_shutdown(IBVSocket* _this)
{
   IBVCommContext* commContext = _this->commContext;

   unsigned numWaitWrites = 0;
   unsigned numWaitReads = 0;
   int timeoutMS = 250;

   if(_this->errState)
      return fhgfs_true; // fhgfs_true, because the conn is down anyways

   if(!commContext)
      return fhgfs_true; // this socket has never been connected

   if(commContext->incompleteSend.numAvailable)
   { // wait for all incomplete sends
      int waitRes;

      waitRes = __IBVSocket_waitForTotalSendCompletion(_this,
         &commContext->incompleteSend.numAvailable, &numWaitWrites, &numWaitReads, timeoutMS);
      if(waitRes < 0)
      {
         printk_fhgfs(KERN_INFO, "%s: Waiting for incomplete send requests failed\n", __func__);
         return fhgfs_false;
      }
   }

   //__IBVSocket_disconnect(_this, fhgfs_false); // note: read the method comments before using this

   return fhgfs_true;
}

EXPORT_SYMBOL(IBVSocket_shutdown);

/**
 * Continues an incomplete former recv() by returning immediately available data from the
 * corresponding buffer.
 */
ssize_t __IBVSocket_recvContinueIncomplete(IBVSocket* _this, char* buf, size_t bufLen)
{
   IBVCommContext* commContext = _this->commContext;
   int completedOffset = commContext->incompleteRecv.completedOffset;
   size_t availableLen = commContext->incompleteRecv.wc.byte_len - completedOffset;
   size_t bufIndex = commContext->incompleteRecv.wc.wr_id - IBVSOCKET_RECV_WORK_ID_OFFSET;

   if(unlikely(_this->errState) )
      return -1;

   if(availableLen <= bufLen)
   { // old data fits completely into buf
      IGNORE_RETURN_VALUE(
         __copy_to_user(buf, &(commContext->recvBufs)[bufIndex][completedOffset], availableLen) );

      commContext->incompleteRecv.isAvailable = 0;

      if(__IBVSocket_postRecv(_this, _this->commContext, bufIndex) )
         goto err_invalidateSock;

      return availableLen;
   }
   else
   { // still too much data for the buf => copy partially
      IGNORE_RETURN_VALUE(
         __copy_to_user(buf, &(commContext->recvBufs)[bufIndex][completedOffset], bufLen) );

      commContext->incompleteRecv.completedOffset += bufLen;

      return bufLen;
   }


err_invalidateSock:
   _this->errState = -1;

   return -1;
}


ssize_t IBVSocket_recv(IBVSocket* _this, char* buf, size_t bufLen, int flags)
{
   const int timeoutMS = 1024*1024;
   ssize_t recvTRes;

   do
   {
      recvTRes = IBVSocket_recvT(_this, buf, bufLen, flags, timeoutMS);
   } while(!recvTRes);

   return recvTRes;
}

EXPORT_SYMBOL(IBVSocket_recv);


/**
 * @return number of received bytes on success, 0 on timeout, -1 on error
 */
ssize_t IBVSocket_recvT(IBVSocket* _this, char* buf, size_t bufLen, int flags, int timeoutMS)
{
   IBVCommContext* commContext = _this->commContext;
   int recvWCRes;

   // check whether an old buffer has not been fully read yet
   if(commContext->incompleteRecv.isAvailable)
      return __IBVSocket_recvContinueIncomplete(_this, buf, bufLen);

   // recv new buffer
   recvWCRes = __IBVSocket_recvWC(_this, timeoutMS);
   if(unlikely(recvWCRes <= 0) )
   {
      if(!recvWCRes)
         return 0; // timeout

      goto err_invalidateSock; // error occured
   }

   // recvWC was positive => we're guaranteed to have an incomplete recv buf availabe

   return __IBVSocket_recvContinueIncomplete(_this, buf, bufLen);


err_invalidateSock:
   _this->errState = -1;

   return -1;
}

EXPORT_SYMBOL(IBVSocket_recvT);

/**
 * @flags supports MSG_DONTWAIT
 */
ssize_t IBVSocket_send(IBVSocket* _this, const char* buf, size_t bufLen, int flags)
{
   IBVCommContext* commContext = _this->commContext;
   int postRes;
   size_t postedLen = 0;
   int currentPostLen;
   int waitRes;

   unsigned numWaitWrites = 0;
   unsigned numWaitReads = 0;
   int timeoutMS = IBVSOCKET_COMPLETION_TIMEOUT_MS;

   // handle flags
   if(flags & MSG_DONTWAIT)
   { // send only as much as we can without blocking

      // note: we adept the bufLen variable as necessary here for simplicity

      int checkSendRes;
      size_t bufNumLeft;
      size_t bufLenLeft;

      checkSendRes = IBVSocket_nonblockingSendCheck(_this);
      if(!checkSendRes)
      { // sorry, we can't send non-blocking currently
         return 0;
      }
      else
      if(unlikely(checkSendRes < 0) )
         goto err_invalidateSock;

      // buffers available => adept bufLen (if necessary)

      bufNumLeft = commContext->commCfg.bufNum - commContext->incompleteSend.numAvailable;
      bufLenLeft = bufNumLeft * commContext->commCfg.bufSize;

      bufLen = MIN(bufLen, bufLenLeft);
   }

   // send data cut in buf-sized pieces...

   do
   {
      if(commContext->incompleteSend.forceWaitForAll ||
         (commContext->incompleteSend.numAvailable == commContext->commCfg.bufNum) )
      { // wait for all (!) incomplete sends
         waitRes = __IBVSocket_waitForTotalSendCompletion(_this,
            &commContext->incompleteSend.numAvailable, &numWaitWrites, &numWaitReads, timeoutMS);
         if(waitRes <= 0)
            goto err_invalidateSock;

         commContext->incompleteSend.forceWaitForAll = fhgfs_false;
      }


      currentPostLen = MIN(bufLen-postedLen, commContext->commCfg.bufSize);

      //memcpy( (commContext->sendBufs)[commContext->incompleteSend.numAvailable], &buf[postedLen],
      //   currentPostLen);
      IGNORE_RETURN_VALUE(
         __copy_from_user( (commContext->sendBufs)[commContext->incompleteSend.numAvailable],
            &buf[postedLen], currentPostLen) );

      postRes = __IBVSocket_postSend(_this,
         commContext->incompleteSend.numAvailable, currentPostLen);
      if(postRes)
         goto err_invalidateSock;

      commContext->incompleteSend.numAvailable++;

      postedLen += currentPostLen;

   } while(postedLen < bufLen);

   return (ssize_t)bufLen;


err_invalidateSock:
   _this->errState = -1;

   return -1;
}

EXPORT_SYMBOL(IBVSocket_send);

/*
int __IBVSocket_registerBuf(IBVCommContext* commContext, char* buf, size_t bufLen,
   struct ib_mr** outMR)
{
   // Note: IB spec says:
   //    "The consumer is not allowed to assign remote-write or remote-atomic to
   //    a memory region that has not been assigned local-write."
   enum ib_access_flags accessFlags = (enum ib_access_flags)
      (IB_ACCESS_REMOTE_READ | IB_ACCESS_REMOTE_WRITE | IB_ACCESS_LOCAL_WRITE);

   *outMR = ib_reg_mr(commContext->pd, buf, bufLen, accessFlags);
   if(!*outMR)
   {
      printk_fhgfs(KERN_INFO, "%d:%s: Couldn't allocate MR\n", __LINE__, __func__);
      return -1;
   }

   return 0;
}
*/

/**
 * @param buf must be physically contiguous (kmalloc'ed)
 */
int __IBVSocket_registerBuf(IBVCommContext* commContext, char* buf, size_t bufLen, u64* outBufDMA)
{
   int retVal = 0;

   *outBufDMA = ib_dma_map_single(commContext->pd->device, buf, bufLen, DMA_BIDIRECTIONAL);
   if(ib_dma_mapping_error(commContext->pd->device, *outBufDMA) )
   {
      printk_fhgfs(KERN_INFO, "%d:%s: Couldn't map mem region for DMA\n", __LINE__, __func__);

      retVal = -ENOMEM;
      *outBufDMA = (u64)NULL;
   }

   return retVal;
}


/**
 * @return kmalloc'ed buffer
 */
char* __IBVSocket_allocAndRegisterBuf(IBVCommContext* commContext, size_t bufLen, u64* outBufDMA)
{
   char* buf;
   int registerRes;

   buf = (char*)kmalloc(bufLen, GFP_KERNEL);
   if(!buf)
   {
      printk_fhgfs(KERN_INFO, "%d:%s: Couldn't allocate work buf.\n", __LINE__, __func__);
      return NULL;
   }

   memset(buf, 0, bufLen);

   registerRes = __IBVSocket_registerBuf(commContext, buf, bufLen, outBufDMA);
   if(registerRes < 0)
   {
      kfree(buf);
      return NULL;
   }

   return buf;
}

fhgfs_bool __IBVSocket_createCommContext(IBVSocket* _this, struct rdma_cm_id* cm_id,
   IBVCommConfig* commCfg, IBVCommContext** outCommContext)
{
   IBVCommContext* commContext;
   int registerControlRes;
   int registerControlResetRes;
   struct ib_qp_init_attr qpInitAttr;
   int qpRes;
   unsigned i;

   commContext = (IBVCommContext*)kzalloc(sizeof(*commContext), GFP_KERNEL);
   if(!commContext)
      goto err_cleanup;

   // prepare recv and send event notification

   init_waitqueue_head(&commContext->recvCompWaitQ);
   init_waitqueue_head(&commContext->sendCompWaitQ);

   atomic_set(&commContext->recvCompEventCount, 0);
   atomic_set(&commContext->sendCompEventCount, 0);

   // protection domain...

   commContext->pd = ib_alloc_pd(cm_id->device);
   if(IS_ERR(commContext->pd) )
   {
      printk_fhgfs(KERN_INFO, "%s:%d: Couldn't allocate PD. ErrCode: %d\n", __func__, __LINE__,
         (int)PTR_ERR(commContext->pd) );

      commContext->pd = NULL;
      goto err_cleanup;
   }

   // DMA system mem region...

   // (Note: IB spec says:
   //    "The consumer is not allowed to assign remote-write or remote-atomic to
   //    a memory region that has not been assigned local-write.")

   commContext->dmaMR = ib_get_dma_mr(commContext->pd,
      IB_ACCESS_LOCAL_WRITE| IB_ACCESS_REMOTE_READ| IB_ACCESS_REMOTE_WRITE);
   if(IS_ERR(commContext->dmaMR) )
   {
      printk_fhgfs(KERN_INFO, "%s:%d: ib_get_dma_mr failed. ErrCode: %d\n", __func__, __LINE__,
         (int)PTR_ERR(commContext->dmaMR) );

      commContext->dmaMR = NULL;
      goto err_cleanup;
   }

   // alloc and register buffers...

   commContext->commCfg = *commCfg;

   commContext->recvBufs = (char**)kzalloc(commCfg->bufNum * sizeof(char*), GFP_KERNEL);
   commContext->recvBufsDMA = (u64*)kzalloc(commCfg->bufNum * sizeof(u64), GFP_KERNEL);

   for(i=0; i < commCfg->bufNum; i++)
   {
      commContext->recvBufs[i] = __IBVSocket_allocAndRegisterBuf(commContext, commCfg->bufSize,
         &commContext->recvBufsDMA[i] );
      if(!commContext->recvBufs[i])
      {
         printk_fhgfs(KERN_INFO, "%s:%d: Couldn't prepare recvBuf #%d\n", __func__, __LINE__, i+1);
         goto err_cleanup;
      }
   }

   commContext->sendBufs = (char**)kzalloc(commCfg->bufNum * sizeof(char*), GFP_KERNEL);
   commContext->sendBufsDMA = (u64*)kzalloc(commCfg->bufNum * sizeof(u64), GFP_KERNEL);

   for(i=0; i < commCfg->bufNum; i++)
   {
      commContext->sendBufs[i] = __IBVSocket_allocAndRegisterBuf(commContext, commCfg->bufSize,
         &commContext->sendBufsDMA[i] );
      if(!commContext->sendBufs[i])
      {
         printk_fhgfs(KERN_INFO, "%s:%d: Couldn't prepare sendBuf #%d\n", __func__, __LINE__, i+1);
         goto err_cleanup;
      }
   }

   registerControlRes = __IBVSocket_registerBuf(commContext,
      (char*)&commContext->numUsedSendBufs, sizeof(commContext->numUsedSendBufs),
      &commContext->numUsedSendBufsDMA);
   if(registerControlRes < 0)
   {
      printk_fhgfs(KERN_INFO, "%s:%d: Couldn't register control memory region\n",
         __func__, __LINE__);
      goto err_cleanup;
   }

   registerControlResetRes = __IBVSocket_registerBuf(commContext,
      (char*)&commContext->numUsedSendBufsReset, sizeof(commContext->numUsedSendBufsReset),
      &commContext->numUsedSendBufsResetDMA);
   if(registerControlResetRes < 0)
   {
      printk_fhgfs(KERN_INFO, "%s:%d: Couldn't register control memory reset region\n",
         __func__, __LINE__);
      goto err_cleanup;
   }

   // create completion queues...

   commContext->recvCQ = __IBVSocket_createCompletionQueue(cm_id->device,
      __IBVSocket_recvCompletionHandler, __IBVSocket_cqRecvEventHandler,
      _this, commCfg->bufNum);
   if(IS_ERR(commContext->recvCQ) )
   {
      printk_fhgfs(KERN_INFO, "%s:%d: Couldn't create recv CQ. ErrCode: %d\n",
         __func__, __LINE__, (int)PTR_ERR(commContext->recvCQ) );
      goto err_cleanup;
   }

   // note: 1+commCfg->bufNum here for the checkConnection() RDMA read
   commContext->sendCQ = __IBVSocket_createCompletionQueue(cm_id->device,
      __IBVSocket_sendCompletionHandler, __IBVSocket_cqSendEventHandler,
      _this, 1+commCfg->bufNum);
   if(IS_ERR(commContext->sendCQ) )
   {
      printk_fhgfs(KERN_INFO, "%s:%d: Couldn't create send CQ. ErrCode: %d\n",
         __func__, __LINE__, (int)PTR_ERR(commContext->sendCQ) );
      goto err_cleanup;
   }

   // note: 1+commCfg->bufNum here for the checkConnection() RDMA read
   memset(&qpInitAttr, 0, sizeof(qpInitAttr) );

   qpInitAttr.event_handler = __IBVSocket_qpEventHandler;
   qpInitAttr.send_cq = commContext->sendCQ;
   qpInitAttr.recv_cq = commContext->recvCQ;
   qpInitAttr.qp_type = IB_QPT_RC;
   qpInitAttr.sq_sig_type = IB_SIGNAL_REQ_WR;
   qpInitAttr.cap.max_send_wr = 1+commCfg->bufNum;
   qpInitAttr.cap.max_recv_wr = commCfg->bufNum;
   qpInitAttr.cap.max_send_sge = 1;
   qpInitAttr.cap.max_recv_sge = 1;
   qpInitAttr.cap.max_inline_data = 0;

   qpRes = rdma_create_qp(cm_id, commContext->pd, &qpInitAttr);
   if(qpRes)
   {
      printk_fhgfs(KERN_INFO, "%s:%d: Couldn't create QP. ErrCode: %d\n",
         __func__, __LINE__, qpRes);
      goto err_cleanup;
   }

   commContext->qp = cm_id->qp;

   // post initial recv buffers...

   for(i=0; i < commCfg->bufNum; i++)
   {
      __IBVSocket_postRecv(_this, commContext, i);
   }

   // prepare event notification...

   // initial event notification requests
   if(ib_req_notify_cq(commContext->recvCQ, IB_CQ_NEXT_COMP) )
   {
      printk_fhgfs(KERN_INFO, "%s:%d: Couldn't request CQ notification\n", __func__, __LINE__);
      goto err_cleanup;
   }

   if(ib_req_notify_cq(commContext->sendCQ, IB_CQ_NEXT_COMP) )
   {
      printk_fhgfs(KERN_INFO, "%s:%d: Couldn't request CQ notification\n", __func__, __LINE__);
      goto err_cleanup;
   }

   *outCommContext = commContext;
   return fhgfs_true;


   //  error handling

err_cleanup:
   __IBVSocket_cleanupCommContext(commContext);

   *outCommContext = NULL;
   return fhgfs_false;
}

void __IBVSocket_cleanupCommContext(IBVCommContext* commContext)
{
   int i;
   struct ib_device* dev;


   if(!commContext)
      return;

   dev = commContext->pd ? commContext->pd->device : NULL;

   if(!dev)
      goto cleanup_no_dev;

   if(commContext->qp)
   {
      if(ib_destroy_qp(commContext->qp) )
         printk_fhgfs(KERN_INFO, "Couldn't destroy QP\n");
   }

   if(commContext->sendCQ)
      ib_destroy_cq(commContext->sendCQ);

   if(commContext->recvCQ)
      ib_destroy_cq(commContext->recvCQ);

   if(commContext->numUsedSendBufsDMA)
   {
      ib_dma_unmap_single(dev, commContext->numUsedSendBufsDMA,
         sizeof(commContext->numUsedSendBufs), DMA_BIDIRECTIONAL);
   }

   if(commContext->numUsedSendBufsResetDMA)
   {
      ib_dma_unmap_single(dev, commContext->numUsedSendBufsResetDMA,
         sizeof(commContext->numUsedSendBufsReset), DMA_BIDIRECTIONAL);
   }

   for(i=0; i < commContext->commCfg.bufNum; i++)
   {
      if(commContext->recvBufsDMA[i] )
      {
         ib_dma_unmap_single(dev,
            commContext->recvBufsDMA[i], commContext->commCfg.bufSize, DMA_BIDIRECTIONAL);
      }

      if(commContext->recvBufs)
         SAFE_KFREE(commContext->recvBufs[i] );

      if(commContext->sendBufsDMA[i] )
      {
         ib_dma_unmap_single(dev,
            commContext->sendBufsDMA[i], commContext->commCfg.bufSize, DMA_BIDIRECTIONAL);
      }

      if(commContext->sendBufs)
         SAFE_KFREE(commContext->sendBufs[i] );
   }

   SAFE_KFREE(commContext->recvBufs);
   SAFE_KFREE(commContext->recvBufsDMA);
   SAFE_KFREE(commContext->sendBufs);
   SAFE_KFREE(commContext->sendBufsDMA);

   if(commContext->dmaMR)
      ib_dereg_mr(commContext->dmaMR);

   if(commContext->pd)
      ib_dealloc_pd(commContext->pd);


cleanup_no_dev:

   kfree(commContext);
}

/**
 * Initializes a (local) IBVCommDest.
 */
void __IBVSocket_initCommDest(IBVCommContext* commContext, IBVCommDest* outDest)
{
   memcpy(outDest->verificationStr, IBVSOCKET_PRIVATEDATA_STR, IBVSOCKET_PRIVATEDATA_STR_LEN);

   //outDest->protocolVersion = IBVSOCKET_PRIVATEDATA_PROTOCOL_VER;
   Serialization_serializeUInt64( (char*)&outDest->protocolVersion,
      IBVSOCKET_PRIVATEDATA_PROTOCOL_VER);

   //outDest->rkey = commContext->dmaMR->rkey;
   Serialization_serializeUInt( (char*)&outDest->rkey, commContext->dmaMR->rkey);

   //outDest->vaddr = commContext->numUsedSendBufsDMA;
   Serialization_serializeUInt64( (char*)&outDest->vaddr, commContext->numUsedSendBufsDMA);

   //outDest->recvBufNum = commContext->commCfg.bufNum;
   Serialization_serializeUInt( (char*)&outDest->recvBufNum, commContext->commCfg.bufNum);

   //outDest->recvBufSize = commContext->commCfg.bufSize;
   Serialization_serializeUInt( (char*)&outDest->recvBufSize, commContext->commCfg.bufSize);
}

/**
 * Parses and checks a (remote) IBVCommDest.
 *
 * @param buf should usually be the private_data of the connection handshake
 * @param outDest will be kmalloced (if fhgfs_true is returned) and needs to be kfree'd by the
 * caller
 * @return fhgfs_true if data is okay, fhgfs_false otherwise
 */
fhgfs_bool __IBVSocket_parseCommDest(const void* buf, size_t bufLen, IBVCommDest** outDest)
{
   IBVCommDest* dest = NULL;
   uint64_t destProtoVer;
   unsigned destProtoLen;
   unsigned destRKey;
   unsigned destRKeyLen;
   uint64_t destVAddr;
   unsigned destVAddrLen;
   unsigned destRecvBufNum;
   unsigned destRecvBufNumLen;
   unsigned destRecvBufSize;
   unsigned destRecvBufSizeLen;

   *outDest = NULL;


   // Note: "bufLen < ..." (and not "!="), because there might be some extra padding
   if(!buf || (bufLen < sizeof(*dest) ) )
   {
      printk_fhgfs(KERN_INFO, "%d:%s: bad private data. ptr: %p len: %d\n",
         __LINE__, __func__, buf, (int)bufLen);

      return fhgfs_false;
   }

   dest = (IBVCommDest*)kmalloc(sizeof(*dest), GFP_ATOMIC);
   if(!dest)
      return fhgfs_false;

   memcpy(dest, buf, sizeof(*dest) );

   if(memcmp(dest->verificationStr, IBVSOCKET_PRIVATEDATA_STR, IBVSOCKET_PRIVATEDATA_STR_LEN) )
      goto err_cleanup;


   Serialization_deserializeUInt64(
      (const char*)&dest->protocolVersion, 8, &destProtoVer, &destProtoLen);

   dest->protocolVersion = destProtoVer; // replace with native-endian version

   if(destProtoVer != IBVSOCKET_PRIVATEDATA_PROTOCOL_VER)
      goto err_cleanup;


   Serialization_deserializeUInt(
      (const char*)&dest->rkey, 8L, &destRKey, &destRKeyLen);

   dest->rkey = destRKey; // replace with native-endian version

   Serialization_deserializeUInt64(
      (const char*)&dest->vaddr, 8L, &destVAddr, &destVAddrLen);

   dest->vaddr = destVAddr; // replace with native-endian version

   Serialization_deserializeUInt(
      (const char*)&dest->recvBufNum, 8L, &destRecvBufNum, &destRecvBufNumLen);

   dest->recvBufNum = destRecvBufNum; // replace with native-endian version

   Serialization_deserializeUInt(
      (const char*)&dest->recvBufSize, 8L, &destRecvBufSize, &destRecvBufSizeLen);

   dest->recvBufSize = destRecvBufSize; // replace with native-endian version


   *outDest = dest;

   return fhgfs_true;


err_cleanup:
   SAFE_KFREE(dest);

   return fhgfs_false;
}


/**
 * @param commContext passed seperately because it's not the _this->commContext during
 *    accept() of incoming connections
 * @return 0 on success, -1 on error
 */
int __IBVSocket_postRecv(IBVSocket* _this, IBVCommContext* commContext, size_t bufIndex)
{
   struct ib_sge list;
   struct ib_recv_wr wr;
   struct ib_recv_wr* bad_wr;
   int postRes;

   list.addr = (u64)commContext->recvBufsDMA[bufIndex];
   list.length = commContext->commCfg.bufSize;
   list.lkey = commContext->dmaMR->lkey;

   wr.next = NULL;
   wr.wr_id = bufIndex + IBVSOCKET_RECV_WORK_ID_OFFSET;
   wr.sg_list = &list;
   wr.num_sge = 1;

   postRes = ib_post_recv(commContext->qp, &wr, &bad_wr);
   if(unlikely(postRes) )
   {
      printk_fhgfs(KERN_INFO, "%d:%s: ib_post_recv failed. ErrCode: %d\n",
         __LINE__, __func__, postRes);

      return -1;
   }

   return 0;
}

/**
 * @param localBuf DMA pointer
 * @return 0 on success, -1 on error
 */
int __IBVSocket_postWrite(IBVSocket* _this, IBVCommDest* remoteDest, u64 localBuf, int bufLen)
{
   IBVCommContext* commContext = _this->commContext;
   struct ib_sge list;
   struct ib_send_wr wr;
   struct ib_send_wr *bad_wr;
   int postRes;
   int waitRes;

   int timeoutMS = IBVSOCKET_COMPLETION_TIMEOUT_MS;
   unsigned numWaitWrites = 1;
   unsigned numWaitReads = 0;

   list.addr = (u64)localBuf;
   list.length = bufLen;
   list.lkey = commContext->dmaMR->lkey;

   wr.wr.rdma.remote_addr = remoteDest->vaddr;
   wr.wr.rdma.rkey = remoteDest->rkey;

   wr.wr_id      = IBVSOCKET_WRITE_WORK_ID;
   wr.sg_list    = &list;
   wr.num_sge    = 1;
   wr.opcode     = IB_WR_RDMA_WRITE;
   wr.send_flags = IB_SEND_SIGNALED;
   wr.next       = NULL;

   postRes = ib_post_send(commContext->qp, &wr, &bad_wr);
   if(unlikely(postRes) )
   {
      printk_fhgfs(KERN_INFO, "%d:%s: ib_post_send() failed. ErrCode: %d\n",
         __LINE__, __func__, postRes);

      return -1;
   }

   waitRes = __IBVSocket_waitForTotalSendCompletion(_this,
      &commContext->incompleteSend.numAvailable, &numWaitWrites, &numWaitReads, timeoutMS);
   if(unlikely(waitRes <= 0) )
      return -1;

   commContext->incompleteSend.forceWaitForAll = fhgfs_false;

   return 0;
}

/**
 * @param localBuf DMA pointer
 * @return 0 on success, -1 on error
 */
int __IBVSocket_postRead(IBVSocket* _this, IBVCommDest* remoteDest,
   u64 localBuf, int bufLen)
{
   IBVCommContext* commContext = _this->commContext;
   struct ib_sge list;
   struct ib_send_wr wr;
   struct ib_send_wr *bad_wr;
   int postRes;
   int waitRes;

   int timeoutMS = IBVSOCKET_COMPLETION_TIMEOUT_MS;
   unsigned numWaitWrites = 0;
   unsigned numWaitReads = 1;

   list.addr = (u64)localBuf;
   list.length = bufLen;
   list.lkey = commContext->dmaMR->lkey;

   wr.wr.rdma.remote_addr = remoteDest->vaddr;
   wr.wr.rdma.rkey = remoteDest->rkey;

   wr.wr_id      = IBVSOCKET_READ_WORK_ID;
   wr.sg_list    = &list;
   wr.num_sge    = 1;
   wr.opcode     = IB_WR_RDMA_READ;
   wr.send_flags = IB_SEND_SIGNALED;
   wr.next       = NULL;

   postRes = ib_post_send(commContext->qp, &wr, &bad_wr);
   if(unlikely(postRes) )
   {
      printk_fhgfs(KERN_INFO, "%d:%s: ib_post_send() failed. ErrCode: %d\n",
         __LINE__, __func__, postRes);

      return -1;
   }

   waitRes = __IBVSocket_waitForTotalSendCompletion(_this,
      &commContext->incompleteSend.numAvailable, &numWaitWrites, &numWaitReads, timeoutMS);
   if(unlikely(waitRes <= 0) )
      return -1;

   commContext->incompleteSend.forceWaitForAll = fhgfs_false;

   return 0;
}

/**
 * @return 0 on success, -1 on error
 */
int __IBVSocket_postSend(IBVSocket* _this, size_t bufIndex, int bufLen)
{
   IBVCommContext* commContext = _this->commContext;
   struct ib_sge list;
   struct ib_send_wr wr;
   struct ib_send_wr *bad_wr;
   int postRes;

   list.addr   = (u64)commContext->sendBufsDMA[bufIndex];
   list.length = bufLen;
   list.lkey   = commContext->dmaMR->lkey;

   wr.wr_id      = bufIndex + IBVSOCKET_SEND_WORK_ID_OFFSET;
   wr.sg_list    = &list;
   wr.num_sge    = 1;
   wr.opcode     = IB_WR_SEND;
   wr.send_flags = IB_SEND_SIGNALED;
   wr.next       = NULL;

   postRes = ib_post_send(commContext->qp, &wr, &bad_wr);
   if(unlikely(postRes) )
   {
      printk_fhgfs(KERN_INFO, "%d:%s: ib_post_send() failed. ErrCode: %d\n",
         __LINE__, __func__, postRes);

      return -1;
   }

   return 0;
}


/**
 * Note: The received wc is returned via commContext::incompleteRecv.
 *
 * @param timeoutMS 0 for non-blocking
 * @return 1 on success, 0 on timeout, -1 on error
 */
int __IBVSocket_recvWC(IBVSocket* _this, int timeoutMS)
{
   //IBVCommContext* commContext = _this->commContext;
   //int reqNotifyRecvRes;
   int recvRes;
   int waitRes;

   for( ; ; )
   {
      recvRes = IBVSocket_nonblockingRecvCheck(_this);
      if(recvRes > 0)
         return recvRes;
      else
      if(!recvRes)
      { // no incoming data yet => wait
         waitRes = __IBVSocket_waitForRecvCompletionEvent(_this, timeoutMS);
         if(unlikely(waitRes <= 0) )
            return waitRes;

         continue;
      }
      else
      if(unlikely(recvRes < 0) )
         return -1;
   }
}


/**
 * @param oldRecvCount old recvCompEventCount
 * @return 1 on available data, 0 on timeout, -1 on error
 */
int __IBVSocket_waitForRecvCompletionEvent(IBVSocket* _this, int timeoutMS)
{
   IBVCommContext* commContext = _this->commContext;
   long waitRes;
   int checkRes;

   for( ; ; )
   {
      // Note: We use pollTimeoutMS to check the conn every few secs (otherwise we might
      //    wait for a very long time in case the other side disconnected silently)
      int pollTimeoutMS = MIN(10000, timeoutMS);
      long pollTimeoutJiffies = __IBVSocket_msToJiffiesSchedulable(pollTimeoutMS);

      waitRes = wait_event_interruptible_timeout(commContext->recvCompWaitQ,
         IBVSocket_nonblockingRecvCheck(_this), pollTimeoutJiffies);

      if(unlikely(waitRes == -ERESTARTSYS) )
      { // signal pending
         printk_fhgfs_debug(KERN_INFO, "%s:%d: wait for recvCompEvent ended by pending signal\n",
            __func__, __LINE__);

         return -1;
      }

      if(likely(IBVSocket_nonblockingRecvCheck(_this) ) )
      { // we got something

         //printk_fhgfs(KERN_INFO "%s:%d: received a completion event\n",
         //    __func__, __LINE__); // debug in

         return 1;
      }

      // timeout

      //printk_fhgfs(KERN_INFO "%s: wait timed out\n", __func__); // debug in

      checkRes = IBVSocket_checkConnection(_this);
      if(checkRes < 0)
         return -1;

      timeoutMS -= pollTimeoutMS;
      if(!timeoutMS)
         return 0;

   } // end of for-loop

}


/**
 * @param oldSendCount old sendCompEventCount
 * @return 1 on available data, 0 on timeout, -1 on error
 */
int __IBVSocket_waitForSendCompletionEvent(IBVSocket* _this, int oldSendCount, int timeoutMS)
{
   IBVCommContext* commContext = _this->commContext;
   long waitRes;

   for( ; ; )
   {
      // Note: We use pollTimeoutMS to check the conn every few secs (otherwise we might
      //    wait for a very long time in case the other side disconnected silently)
      int pollTimeoutMS = MIN(10000, timeoutMS);
      long pollTimeoutJiffies = __IBVSocket_msToJiffiesSchedulable(pollTimeoutMS);

      waitRes = wait_event_interruptible_timeout(commContext->sendCompWaitQ,
         atomic_read(&commContext->sendCompEventCount) != oldSendCount, pollTimeoutJiffies);

      if(unlikely(waitRes == -ERESTARTSYS) )
      { // signal pending
         printk_fhgfs_debug(KERN_INFO, "%s:%d: wait for sendCompEvent ended by pending signal\n",
            __func__, __LINE__);

         return -1;
      }

      if(unlikely(atomic_read(&commContext->sendCompEventCount) == oldSendCount) )
      { // timeout
         //printk_fhgfs(KERN_INFO "%s: wait timed out\n", __func__); // debug in

         timeoutMS -= pollTimeoutMS;
         if(!timeoutMS)
            return 0;

         continue;
      }
      else
      {
         //printk_fhgfs(KERN_INFO "%s:%d: received a completion event\n",
         //    __func__, __LINE__); // debug in

         return 1;
      }

   } // end of for-loop
}


/**
 * @param numSendElements also used as out-param to return the remaining number
 * @param timeoutMS 0 for non-blocking; this is a soft timeout that is reset after each received
 * completion
 * @return 1 if all completions received, 0 if completions missing (in case you wanted non-blocking)
 * or -1 in case of an error.
 */
int __IBVSocket_waitForTotalSendCompletion(IBVSocket* _this,
   unsigned* numSendElements, unsigned* numWriteElements, unsigned* numReadElements, int timeoutMS)
{
   IBVCommContext* commContext = _this->commContext;
   int numElements;
   int waitRes;
   int oldSendCount;
   int i;
   size_t bufIndex;
   struct ib_wc wc[2];

   do
   {
      oldSendCount = atomic_read(&commContext->sendCompEventCount);

      numElements = ib_poll_cq(commContext->sendCQ, 2, wc);
      if(unlikely(numElements < 0) )
      {
         printk_fhgfs(KERN_INFO, "%s:%d: bad ib_poll_cq result: %d\n", __func__, __LINE__,
            numElements);

         return -1;
      }
      else
      if(!numElements)
      { // no completions available yet => wait
         if(!timeoutMS)
            return 0;

         waitRes = __IBVSocket_waitForSendCompletionEvent(_this, oldSendCount, timeoutMS);
         if(likely(waitRes > 0) )
            continue;

         return waitRes;
      }

      // we got something...

      // for each completion element
      for(i=0; i < numElements; i++)
      {
         if(unlikely(wc[i].status != IB_WC_SUCCESS) )
         {
            printk_fhgfs(KERN_INFO, "%d:%s: bad wc status %d\n", __LINE__, __func__, wc[i].status);

            return -1;
         }

         switch(wc[i].opcode)
         {
            case IB_WC_SEND:
            {
               bufIndex = wc[i].wr_id - IBVSOCKET_SEND_WORK_ID_OFFSET;

               if(unlikely(bufIndex >= commContext->commCfg.bufNum) )
               {
                  printk_fhgfs(KERN_INFO, "%d:%s: bad send completion wr_id 0x%x\n", __LINE__,
                     __func__, (int)wc[i].wr_id);

                  return -1;
               }

               if(likely(*numSendElements) )
                  (*numSendElements)--;
               else
               {
                  printk_fhgfs(KERN_INFO, "%s:%d: received bad/unexpected send completion\n",
                     __func__, __LINE__);

                  return -1;
               }

            } break;

            case IB_WC_RDMA_WRITE:
            {
               if(unlikely(wc[i].wr_id != IBVSOCKET_WRITE_WORK_ID) )
               {
                  printk_fhgfs(KERN_INFO, "%d:%s: bad write completion wr_id 0x%x\n",
                     __LINE__, __func__, (int)wc[i].wr_id);

                  return -1;
               }

               if(likely(*numWriteElements) )
                  (*numWriteElements)--;
               else
               {
                  printk_fhgfs(KERN_INFO, "%s:%d: received bad/unexpected RDMA write completion\n",
                     __func__, __LINE__);

                  return -1;
               }
            } break;

            case IB_WC_RDMA_READ:
            {
               if(unlikely(wc[i].wr_id != IBVSOCKET_READ_WORK_ID) )
               {
                  printk_fhgfs(KERN_INFO, "%d:%s: bad read completion wr_id 0x%x\n", __LINE__,
                     __func__, (int)wc[i].wr_id);

                  return -1;
               }

               if(likely(*numReadElements) )
                  (*numReadElements)--;
               else
               {
                  printk_fhgfs(KERN_INFO, "%s:%d: received bad/unexpected RDMA read completion\n",
                     __func__, __LINE__);

                  return -1;
               }
            } break;

            default:
            {
               printk_fhgfs(KERN_INFO, "%s:%d: received bad/unexpected completion opcode %d\n",
                  __func__, __LINE__, wc[i].opcode);

               return -1;
            } break;

         } // end of switch

      } // end of for-loop

   } while(*numSendElements || *numWriteElements || *numReadElements);

   return 1;
}

/**
 * @return 0 on success, -1 on error
 */
int IBVSocket_checkConnection(IBVSocket* _this)
{
   int postRes;
   IBVCommContext* commContext = _this->commContext;

   // note: we read a remote value into the numUsedSendBufsReset field, which is actually
   //    meant for something else, so we need to reset the value afterwards

   //printk_fhgfs_debug(KERN_INFO, "%d:%s: post rdma_read to check connection...\n",
   //   __LINE__, __func__); // debug in

   postRes = __IBVSocket_postRead(_this, _this->remoteDest,
      commContext->numUsedSendBufsResetDMA, sizeof(commContext->numUsedSendBufsReset) );
   if(postRes)
      return -1;

   commContext->numUsedSendBufsReset = 0;

   //printk_fhgfs_debug(KERN_INFO, "%d:%s: rdma_read succeeded\n", __LINE__, __func__); // debug in

   return 0;
}

EXPORT_SYMBOL(IBVSocket_checkConnection);

/**
 * @return <0 on error, 0 if recv would block, >0 if recv would not block
 */
int IBVSocket_nonblockingRecvCheck(IBVSocket* _this)
{
   IBVCommContext* commContext = _this->commContext;
   struct ib_wc* wc;
   int numEvents;
   size_t bufIndex;

   if(commContext->incompleteRecv.isAvailable)
      return 1;

   // recv one packet (if available) and add it as incompleteRecv

   wc = &commContext->incompleteRecv.wc;

   numEvents = ib_poll_cq(_this->commContext->recvCQ, 1, wc);
   if(unlikely(numEvents < 0) )
   {
      printk_fhgfs(KERN_INFO, "%s: poll CQ failed %d\n", __func__, numEvents);
      goto err_invalidateSock;
   }

   if(!numEvents)
      return 0;

   // we got something...

   if(unlikely(wc->status != IB_WC_SUCCESS) )
   {
      printk_fhgfs(KERN_INFO, "%s: Failed. (wc_status: %d; wr_id: %d)\n",
         __func__, (int)wc->status, (int)wc->wr_id);
      goto err_invalidateSock;
   }

   bufIndex = wc->wr_id - IBVSOCKET_RECV_WORK_ID_OFFSET;

   if(unlikely(bufIndex >= commContext->commCfg.bufNum) )
   {
      printk_fhgfs(KERN_INFO, "%s: Completion for unknown/invalid wr_id %d\n",
         __func__, (int)wc->wr_id);
      goto err_invalidateSock;
   }

   // prepare to continue later
   commContext->incompleteRecv.completedOffset = 0;
   commContext->incompleteRecv.isAvailable = 1;

   return 1;

err_invalidateSock:
   _this->errState = -1;
   return -1;
}

EXPORT_SYMBOL(IBVSocket_nonblockingRecvCheck);

/**
 * @return <0 on error, 0 if send would block, >0 if send would not block
 */
int IBVSocket_nonblockingSendCheck(IBVSocket* _this)
{
   IBVCommContext* commContext = _this->commContext;
   int waitRes;

   unsigned numWaitWrites = 0;
   unsigned numWaitReads = 0;
   int timeoutMS = 0;


   if(!commContext->incompleteSend.forceWaitForAll)
   {
      if(commContext->incompleteSend.numAvailable < commContext->commCfg.bufNum)
         return 1;

      // no immediate buffers available => check for completions...
      commContext->incompleteSend.forceWaitForAll = fhgfs_true;
   }

   // we have to wait for completions before we can send...

   waitRes = __IBVSocket_waitForTotalSendCompletion(_this,
      &commContext->incompleteSend.numAvailable, &numWaitWrites, &numWaitReads, timeoutMS);

   if(waitRes > 0)
   {
      commContext->incompleteSend.forceWaitForAll = fhgfs_false;
      return 1;
   }
   else
      return waitRes;
}

EXPORT_SYMBOL(IBVSocket_nonblockingSendCheck);

/**
 * Note: Call this only once with finishPoll==fhgfs_true (=> non-blocking) or multiple times with
 *    finishPoll==fhgfs_true in the last call from the current thread (for cleanup).
 * Note: It's safe to call this multiple times with finishPoll==fhgfs_true.
 *
 * @param events the event flags you are interested in (POLL...)
 * @param finishPoll true for cleanup if you don't call poll again from this thread; (it's also ok
 *    to set this to true if you call poll only once and want to avoid blocking)
 * @return mask revents mask (like poll(), POLL... flags), but only the events you requested or
 *    the error events
 */
unsigned long IBVSocket_poll(IBVSocket* _this, short events, fhgfs_bool finishPoll)
{
   IBVCommContext* commContext = _this->commContext;
   unsigned long revents = 0; // return value

   // note: there are two possible uses for finishPoll==fhgfs_true:
   //    1) this method is called multiple times and finishPoll is fhgfs_true in the last loop
   //       (for cleanup)
   //    2) this method is called only once non-blocking and finishPoll is fhgfs_true to avoid
   //       the blocking

   // note: it's good to call prepare_to_wait more than once for the same thread (e.g. if
   //    caller woke up from schedule() and decides to sleep again)

   // note: condition needs to be re-checked after prepare_to_wait to avoid the race when it became
   //    true between the initial check and the call to prepare_to_wait

   // note: we assume that after we returned a positive result, the caller will not try to sleep

   if(events & POLLIN)
   {
      // initial check and wait preparations

      if(IBVSocket_nonblockingRecvCheck(_this) )
      { // immediate data available => no need to prepare wait
         revents |= POLLIN;
      }
      else
      if(!finishPoll)
      { // no incoming data and caller is planning to schedule()
         if(!commContext->recvWaitInitialized)
         {
            init_wait(&commContext->recvWait);
            commContext->recvWaitInitialized = fhgfs_true;
         }

         prepare_to_wait(&commContext->recvCompWaitQ, &commContext->recvWait, TASK_INTERRUPTIBLE);

         if(IBVSocket_nonblockingRecvCheck(_this) )
            revents |= POLLIN;
      }

      // cleanup

      if(finishPoll && commContext->recvWaitInitialized)
      {
         finish_wait(&commContext->recvCompWaitQ, &commContext->recvWait);
         commContext->recvWaitInitialized = fhgfs_false;
      }
   }

   if(events & POLLOUT)
   {
      // initial check and wait preparations

      if(IBVSocket_nonblockingSendCheck(_this) )
      { // immediate data available => no need to prepare wait
         revents |= POLLOUT;
      }
      else
      if(!finishPoll)
      { // no incoming data and caller is planning to schedule()
         if(!commContext->sendWaitInitialized)
         {
            init_wait(&commContext->sendWait);
            commContext->sendWaitInitialized = fhgfs_true;
         }

         prepare_to_wait(&commContext->sendCompWaitQ, &commContext->sendWait, TASK_INTERRUPTIBLE);

         if(IBVSocket_nonblockingSendCheck(_this) )
            revents |= POLLOUT;

      }

      // cleanup

      if(finishPoll && commContext->sendWaitInitialized)
      {
         finish_wait(&commContext->sendCompWaitQ, &commContext->sendWait);
         commContext->sendWaitInitialized = fhgfs_false;
      }
   }

   if(unlikely(_this->errState) )
      revents |= POLLERR;

   return revents;
}

EXPORT_SYMBOL(IBVSocket_poll);

fhgfs_bool IBVSocket_getSockValid(IBVSocket* _this)
{
   return _this->sockValid;
}

EXPORT_SYMBOL(IBVSocket_getSockValid);

/**
 * Note: this is provided simply because it exists, but somehow the connection tear-down process
 *    often works better (without delay) if this explicit disconnect is avoided. In fact, it seems
 *    a receiver might sometimes even hang infinitely if this is used.
 */
void __IBVSocket_disconnect(IBVSocket* _this, fhgfs_bool waitForEvent)
{
   int disconnectRes;

   disconnectRes = rdma_disconnect(_this->cm_id);
   if(disconnectRes)
   {
      printk_fhgfs(KERN_INFO, "%d:%s: rdma disconnect error\n", __LINE__, __func__);

      return;
   }

}

void __IBVSocket_close(IBVSocket* _this)
{
   SAFE_KFREE(_this->remoteDest);

   if(_this->commContext)
      __IBVSocket_cleanupCommContext(_this->commContext);

   if(_this->cm_id)
      rdma_destroy_id(_this->cm_id);
}


int __IBVSocket_cmaHandler(struct rdma_cm_id* cm_id, struct rdma_cm_event* event)
{
   IBVSocket* _this;

   int retVal = 0;

   _this = (IBVSocket*)cm_id->context;
   if(!_this)
   {
      printk_fhgfs_ir_debug(KERN_INFO, "cm_id is being torn down. Event: %d\n", event->event);
      return (event->event == RDMA_CM_EVENT_CONNECT_REQUEST) ? -EINVAL : 0;
   }

   //Mutex_lock(&_this->mutex); // L O C K (sleeping is not allowed in callbacks!)

   //printk_fhgfs_debug(KERN_INFO, "%s: event %d cm_id %p\n",
   //   __func__, event->event, cm_id); // debug in

   switch(event->event)
   {
      case RDMA_CM_EVENT_ADDR_RESOLVED:
      {
         printk_fhgfs_debug(KERN_INFO, "RDMA_CM_EVENT_ADDR_RESOLVED\n");

         // new connState, but don't notify connecting thread, because we go on automatically
         //    with resolve route

         retVal = rdma_resolve_route(_this->cm_id, IBVSOCKET_CONN_TIMEOUT_MS);
         if(retVal)
         {
            printk_fhgfs_ir_debug(KERN_INFO, "%d:%s: rdma_resolve_route failed. ErrCode: %d\n",
               __LINE__, __func__, retVal);

            __IBVSocket_setConnStateNotify(_this, IBVSOCKETCONNSTATE_FAILED);
         }

      } break;
      case RDMA_CM_EVENT_ADDR_ERROR:
      {
         printk_fhgfs_ir_debug(KERN_INFO, "RDMA_CM_EVENT_ADDR_ERROR\n");

         __IBVSocket_setConnStateNotify(_this, IBVSOCKETCONNSTATE_FAILED);
         retVal = -ENETUNREACH;
      } break;

      case RDMA_CM_EVENT_ROUTE_RESOLVED:
      {
         printk_fhgfs_ir_debug(KERN_INFO, "RDMA_CM_EVENT_ROUTE_RESOLVED\n");

         __IBVSocket_setConnStateNotify(_this, IBVSOCKETCONNSTATE_ROUTERESOLVED);
      } break;
      case RDMA_CM_EVENT_ROUTE_ERROR:
      {
         printk_fhgfs_ir_debug(KERN_INFO, "RDMA_CM_EVENT_ROUTE_ERROR : %p\n", cm_id);

         __IBVSocket_setConnStateNotify(_this, IBVSOCKETCONNSTATE_FAILED);
         retVal = -ETIMEDOUT;
      } break;

      case RDMA_CM_EVENT_CONNECT_REQUEST:
      {
         printk_fhgfs_ir_debug(KERN_INFO, "RDMA_CM_EVENT_CONNECT_REQUEST\n");

         // incoming connections not supported => reject all
         rdma_reject(cm_id, NULL, 0);
      } break;
      case RDMA_CM_EVENT_CONNECT_RESPONSE:
      {
         printk_fhgfs_ir_debug(KERN_INFO, "RDMA_CM_EVENT_CONNECT_RESPONSE\n");

         retVal = __IBVSocket_connectResponseHandler(_this, cm_id, event);
         if(retVal)
            rdma_reject(cm_id, NULL, 0);
         else
            retVal = rdma_accept(cm_id, NULL);

         if(retVal)
            __IBVSocket_setConnStateNotify(_this, IBVSOCKETCONNSTATE_FAILED);

         break;
      }
      case RDMA_CM_EVENT_CONNECT_ERROR:
      {
         printk_fhgfs_ir_debug(KERN_INFO, "RDMA_CM_EVENT_CONNECT_ERROR\n");

         __IBVSocket_setConnStateNotify(_this, IBVSOCKETCONNSTATE_FAILED);
         retVal = -ETIMEDOUT;
      } break;
      case RDMA_CM_EVENT_UNREACHABLE:
      {
         printk_fhgfs_ir_debug(KERN_INFO, "RDMA_CM_EVENT_UNREACHABLE\n");

         __IBVSocket_setConnStateNotify(_this, IBVSOCKETCONNSTATE_FAILED);
         retVal = -ENETUNREACH;
      } break;

      case RDMA_CM_EVENT_REJECTED:
      {
         printk_fhgfs_ir_debug(KERN_INFO, "RDMA_CM_EVENT_REJECTED\n");

         __IBVSocket_setConnStateNotify(_this, IBVSOCKETCONNSTATE_FAILED);
         retVal = -ECONNREFUSED;
      } break;
      case RDMA_CM_EVENT_ESTABLISHED:
      {
         printk_fhgfs_ir_debug(KERN_INFO, "RDMA_CM_EVENT_ESTABLISHED\n");

         retVal = __IBVSocket_connectedHandler(_this, event);
         if(retVal)
            __IBVSocket_setConnStateNotify(_this, IBVSOCKETCONNSTATE_FAILED);
         else
            __IBVSocket_setConnStateNotify(_this, IBVSOCKETCONNSTATE_ESTABLISHED);
      } break;

      case RDMA_CM_EVENT_DISCONNECTED:
      {
         printk_fhgfs_ir_debug(KERN_INFO, "RDMA_CM_EVENT_DISCONNECTED\n");

         rdma_disconnect(cm_id);
         retVal = __IBVSocket_disconnectedHandler(_this);
      } break;

      case RDMA_CM_EVENT_DEVICE_REMOVAL:
         printk_fhgfs_ir_debug(KERN_INFO, "RDMA_CM_EVENT_DEVICE_REMOVAL\n");
         retVal = -ENETRESET;
         break;

      default:
      {
         printk_fhgfs_ir_debug(KERN_INFO, "Ignoring RDMA_CMA event: %d\n", event->event);
         //retVal = -ECONNABORTED;
      } break;
   }

   //printk_fhgfs_debug(KERN_INFO, "%s: event %d handled\n", __func__, event->event); // debug in

   //Mutex_unlock(&_this->mutex); // U N L O C K


   // Note: rdma_cm automatically calls rdma_destroy_id() if retVal indicates an error
   //    (so we must make sure that the cm_id won't be destroyed again in IBVSocket_close() )

   if(retVal)
   {
      //_this->commContext->qp = NULL; // (doesn't seem to be needed here)

      _this->cm_id = NULL; // cm_id will automatically be destroyed because retVal!=0

      if(!_this->errState)
         _this->errState = -1;
   }


   //printk_fhgfs_ir_debug(KERN_INFO, "%s: event %d done. status %d\n",
   //   __func__, event->event, retVal); // debug in


   return retVal;
}

/**
 * Invoked when an asynchronous event not associated with a completion occurs on the CQ.
 */
void __IBVSocket_cqSendEventHandler(struct ib_event *event, void *data)
{
   // nothing to be done here
   printk_fhgfs_ir_debug(KERN_INFO, "%s:%d: called. event type: %d (not handled)\n",
      __func__, __LINE__, (int)event->event);
}

/**
 * Invoked when a completion event occurs on the CQ.
 *
 * @param cq_context IBVSocket* _this
 */
void __IBVSocket_sendCompletionHandler(struct ib_cq *cq, void *cq_context)
{
   IBVSocket* _this = (IBVSocket*)cq_context;
   IBVCommContext* commContext = _this->commContext;
   int reqNotifySendRes;

   atomic_inc(&commContext->sendCompEventCount);
   //printk_fhgfs_ir(KERN_INFO, "INC( SEND ); read: %2d -- %p\n",
   //   atomic_read(&commContext->sendCompEventCount), _this); // debug in

   reqNotifySendRes = ib_req_notify_cq(commContext->sendCQ, IB_CQ_NEXT_COMP);
   if(unlikely(reqNotifySendRes) )
      printk_fhgfs_ir(KERN_INFO, "%s: Couldn't request CQ notification\n", __func__);

   wake_up_interruptible(&commContext->sendCompWaitQ);
}

/**
 * Invoked when an asynchronous event not associated with a completion occurs on the CQ.
 */
void __IBVSocket_cqRecvEventHandler(struct ib_event *event, void *data)
{
   // nothing to be done here
   printk_fhgfs_ir_debug(KERN_INFO, "%s:%d: called. event type: %d (not handled)\n",
      __func__, __LINE__, (int)event->event);
}

/**
 * Invoked when a completion event occurs on the CQ.
 *
 * @param cq_context IBVSocket* _this
 */
void __IBVSocket_recvCompletionHandler(struct ib_cq *cq, void *cq_context)
{
   IBVSocket* _this = (IBVSocket*)cq_context;
   IBVCommContext* commContext = _this->commContext;
   int reqNotifyRecvRes;

   atomic_inc(&commContext->recvCompEventCount);
   //printk_fhgfs_ir(KERN_INFO, "INC( RECV ); read: %2d -- %p\n",
   //   atomic_read(&commContext->recvCompEventCount), _this); // debug in

   reqNotifyRecvRes = ib_req_notify_cq(commContext->recvCQ, IB_CQ_NEXT_COMP);
   if(unlikely(reqNotifyRecvRes) )
      printk_fhgfs_ir(KERN_INFO, "%s: Couldn't request CQ notification\n", __func__);

   wake_up_interruptible(&commContext->recvCompWaitQ);
}

void __IBVSocket_qpEventHandler(struct ib_event *event, void *data)
{
   // nothing to be done here
   printk_fhgfs_ir_debug(KERN_INFO, "%s:%d: called. event type: %d (not handled)\n",
      __func__, __LINE__, (int)event->event);
}

int __IBVSocket_routeResolvedHandler(IBVSocket* _this, struct rdma_cm_id* cm_id,
   IBVCommConfig* commCfg, IBVCommContext** outCommContext)
{
   fhgfs_bool createContextRes;
   struct rdma_conn_param conn_param;

   createContextRes = __IBVSocket_createCommContext(_this, _this->cm_id, commCfg,
      &_this->commContext);
   if(!createContextRes)
   {
      printk_fhgfs_ir(KERN_INFO, "%d:%s: creation of CommContext failed\n", __LINE__, __func__);
      _this->errState = -1;

      return -EPERM;
   }

   // establish connection...

   __IBVSocket_initCommDest(_this->commContext, &_this->localDest);

   memset(&conn_param, 0, sizeof(conn_param) );
   conn_param.responder_resources = 1;
   conn_param.initiator_depth = 1;
   conn_param.flow_control = 1;
   conn_param.retry_count = 7; // (3 bits)
   conn_param.rnr_retry_count = 7; // rnr = receiver not ready (3 bits, 7 means infinity)
   conn_param.private_data = &_this->localDest;
   conn_param.private_data_len = sizeof(_this->localDest);


   return rdma_connect(_this->cm_id, &conn_param);
}


int __IBVSocket_connectResponseHandler(IBVSocket* _this, struct rdma_cm_id* cm_id,
   struct rdma_cm_event* event)
{
   return 0;
}

int __IBVSocket_connectedHandler(IBVSocket* _this, struct rdma_cm_event* event)
{
   int retVal = 0;
   fhgfs_bool parseCommDestRes;

   const void* private_data;
   u8 private_data_len;

#if defined FHGFS_OFED_1_2_API && (FHGFS_OFED_1_2_API == 1)
   private_data = event->private_data;
   private_data_len = event->private_data_len;
#else // OFED 1.2.5 or higher API
   private_data = event->param.conn.private_data;
   private_data_len = event->param.conn.private_data_len;
#endif

   parseCommDestRes = __IBVSocket_parseCommDest(
      private_data, private_data_len, &_this->remoteDest);
   if(!parseCommDestRes)
   {
      printk_fhgfs_ir(KERN_INFO, "%d:%s: bad private data received. len: %d\n",
         __LINE__, __func__, private_data_len);

      retVal = -EOPNOTSUPP;
      goto err_invalidateSock;
   }


   return retVal;


err_invalidateSock:
   _this->errState = -1;

   return retVal;
}

int __IBVSocket_disconnectedHandler(IBVSocket* _this)
{
   return 0;
}


/**
 * Note: Doesn't notify (wake) waiting threads.
 */
void __IBVSocket_setConnState(IBVSocket* _this, IBVSocketConnState_t connState)
{
   _this->connState = connState;
}

/**
 * Note: Notifies (wakes) waiting threads.
 */
void __IBVSocket_setConnStateNotify(IBVSocket* _this, IBVSocketConnState_t connState)
{
   _this->connState = connState;

   wake_up_interruptible(&_this->eventWaitQ);
}

IBVSocketConnState_t __IBVSocket_getConnState(IBVSocket* _this)
{
   IBVSocketConnState_t retVal;

   retVal = _this->connState;

   return retVal;
}

///**
// * @return fhgfs_true if value changed, fhgfs_false on timeout
// */
//fhgfs_bool __IBVSocket_waitForConnStateChange(IBVSocket* _this, int timeoutMS)
//{
//   fhgfs_bool stateChanged = fhgfs_false;
//   IBVSocketConnState_t oldState;
//   cond_wait_res_t waitRes;
//
//
//   Mutex_lock(&_this->connStateMutex);
//
//   oldState = _this->connState;
//
//   for( ; ; )
//   {
//      waitRes = Condition_timedwait(&_this->connStateCond, &_this->connStateMutex, timeoutMS);
//
//      if(waitRes == COND_WAIT_TIMEOUT)
//         break;
//
//      if(_this->connState != oldState)
//      {
//         stateChanged = fhgfs_true;
//         break;
//      }
//   }
//
//   Mutex_unlock(&_this->connStateMutex);
//
//   return stateChanged;
//}

/**
 * @return fhgfs_true if value changed, fhgfs_false on timeout
 */
fhgfs_bool __IBVSocket_waitForConnStateChange(IBVSocket* _this, IBVSocketConnState_t oldState)
{
   wait_event_interruptible(_this->eventWaitQ, _this->connState != oldState);

   return fhgfs_true;
}

/**
 * Convert millisecs to jiffies and mind the max schedule timeout.
 */
long __IBVSocket_msToJiffiesSchedulable(unsigned ms)
{
   unsigned long resultJiffies = msecs_to_jiffies(ms);

   return (resultJiffies >= MAX_SCHEDULE_TIMEOUT) ? (MAX_SCHEDULE_TIMEOUT-1) : resultJiffies;
}

struct ib_cq* __IBVSocket_createCompletionQueue(struct ib_device* device,
   ib_comp_handler comp_handler, void (*event_handler)(struct ib_event *, void *),
   void* cq_context, int cqe)
{
   #if defined (FHGFS_OFED_1_2_API) && FHGFS_OFED_1_2_API >= 1
      return ib_create_cq(device, comp_handler, event_handler, cq_context, cqe);
   #else // OFED 1.2.5 or higher API
      return ib_create_cq(device, comp_handler, event_handler, cq_context, cqe, 0);
   #endif
}



#endif // FHGFS_OPENTK_IBVERBS

